(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: achain.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2007-10-23 14:46:54 $
 * Version: $Revision: 1.24 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. AChain/def_achain.ml pour les infos sur les types achain et acell
 *          
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> seq (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = seq
 *
 *=========================================================================
 *)

open Global ;;
open Valeur ;;
open Printer ;;
open Dispatch_sig ;;

open Def_achain ;;
open Type_achain ;;
open Core_achain ;;

(* Fonctions dispatches des acell *)

let _size_acell _ = VAL_int (-8)

let _intof_acell c = VAL_int(get_uid_from_acell(c))
let _acellof_int i = VAL_acell(get_acell_from_uid(i))

let _acellPredicate_acell _ = mgstrue
let _acellPredicate_any _  = mgsfalse

let _val2bool_acell c = VAL_bool(acelldim c = -1)

let _compareMgsVal_acell_acell (s1, s2) = VAL_int(compare_acell s1 s2)

and _dim_acell s = VAL_int(get_dim_from_acell s) ;;


let rec _faces_acell s = VAL_seq(Array.of_list (List.map (fun x -> VAL_acell(x)) (Core_achain.faces s)), Type_seq.seq_desc)

and _cofaces_acell s = VAL_seq(Array.of_list (List.map (fun x -> VAL_acell(x)) (Core_achain.cofaces s)), Type_seq.seq_desc)

and _icells_acell_int(s,i) = VAL_seq(Array.of_list (List.map (fun x -> VAL_acell(x)) (Core_achain.icells s i)), Type_seq.seq_desc)

and _ccells_acell_int(s,i) = VAL_seq(Array.of_list (List.map (fun x -> VAL_acell(x)) (Core_achain.ccells s i)), Type_seq.seq_desc)

and _transitivefaces_acell s = VAL_seq(Array.of_list (List.map (fun x -> VAL_acell(x)) (Core_achain.transitive_faces s)), Type_seq.seq_desc)

and _transitivecofaces_acell s = VAL_seq(Array.of_list (List.map (fun x -> VAL_acell x) (Core_achain.transitive_cofaces s)), Type_seq.seq_desc)

and _commonfaces_acell_acell (s1, s2) = VAL_seq(Array.of_list(List.map (fun x -> VAL_acell x) (Core_achain.common_faces s1 s2)), Type_seq.seq_desc)

and _commoncofaces_acell_acell (s1, s2) = VAL_seq(Array.of_list(List.map (fun x -> VAL_acell x) (Core_achain.common_cofaces s1 s2)), Type_seq.seq_desc)

and _commonicells_acell_acell_int (s1, s2,i) = VAL_seq(Array.of_list(List.map (fun x -> VAL_acell x) (Core_achain.common_icells s1 s2 i)), Type_seq.seq_desc)

and _commonccells_acell_acell_int (s1, s2,i) = VAL_seq(Array.of_list(List.map (fun x -> VAL_acell x) (Core_achain.common_ccells s1 s2 i)), Type_seq.seq_desc)

and _commontransitivefaces_acell_acell (s1, s2) = VAL_seq(Array.of_list(List.map (fun x -> VAL_acell x) (Core_achain.common_transitive_faces s1 s2)), Type_seq.seq_desc)

and _commontransitivecofaces_acell_acell (s1, s2) = VAL_seq(Array.of_list(List.map (fun x -> VAL_acell x) (Core_achain.common_transitive_cofaces s1 s2)), Type_seq.seq_desc)

;;

let _newEdge_acell_acell (v1, v2) = 
  if v1.dim_acell = 0 & v2.dim_acell = 0 then VAL_acell(new_edge v1 v2) 
  else begin
    let err = "Error: new_edge: faces of a vertex must be of dim 0"
    in prerr_endline err; raise (Erreur err)
  end
;;

let _newFace_acell_acell_acell (f1, f2, f3) = 
  if f1.dim_acell = 1 & f2.dim_acell = 1 & f3.dim_acell = 1 then VAL_acell(new_face f1 f2 f3) 
  else begin
    let err = "Error: new_face: faces of a 'face' (acell of dim. 2) must be of dim 1"
    in prerr_endline err; raise (Erreur err)
  end
;;

(* new_vertex() est directement dans kernel.ml  *)

let _newAcell_int_seq_seq (d, (fl, _), (cfl, _)) = 
  if d < 0 then raise (Erreur "Error: new_acell: invalid negative dimension");
  let proj = function 
    | VAL_acell(sp) -> sp
    | x -> let err = "Error: new_acell: invalid acell sequence in face/coface list"
           in Printf.printf "%s: %a\n" err Printer.prfvalue2 x; raise (Erreur err)
  in 
    begin
      let fl1 = List.map proj (Array.to_list fl)
      and cfl1 = List.map proj (Array.to_list cfl)
      in VAL_acell(new_acell d fl1 cfl1)
    end
;;

let _cartesianProd_funct_achain_achain (f, (ch1, t1), (ch2, t2)) = 
  VAL_achain(cartesian_prod (fun x y -> (to_funct (f x))y) ch1 ch2, Type.commontype t1 t2)
;;


(* Fonctions dispatches des achain *)

(*
 * 1 : Fonctions de construction
*)

let rec _sequify_achain (ch, _) = 
  VAL_seq(Array.of_list (Core_achain.fold (fun s v suite -> v::suite) ch []), Type_seq.seq_desc)

and _bagify_achain  (ch, _) = 
  VAL_bag(Core_achain.fold (fun s v suite -> Core_bag.add_b v suite) ch Core_bag.empty_b, Type_bag.bag_desc)

and _achainify_qmf (q,_) =
  VAL_achain(Core_achain.from_qmf q, Type_achain.achain_desc)

and _achainify_del (d,_) =
  VAL_achain(Core_achain.from_delaunay d, Type_achain.achain_desc)

and _achainify_seq (s,_) =
  VAL_achain(Core_achain.from_seq (fun x -> VAL_rec(x,Type_rec.record_desc)) s, Type_achain.achain_desc)

and _mult_any_acell (v,c) =
  VAL_achain(Core_achain.add_acell (Core_achain.new_achain ()) c v,Type_achain.achain_desc)

and _mult_int_achain (i, (ch,t)) =
  match i with
      0 -> VAL_achain(Core_achain.new_achain (), t)
    | 1 -> VAL_achain(ch, t)
    | _ -> VAL_achain(
	Core_achain.map
			(fun c -> function
			     VAL_gab(g,tg) as v -> VAL_gab(Core_gab.mult_g i g, tg)
			   | VAL_undef _ as v   -> v
			   | VAL_int i'         -> VAL_int(i*i')
			   | VAL_float f        -> VAL_float((float_of_int i) *. f)
			   | v                  -> VAL_gab(Core_gab.addn_g v i Core_gab.empty_g, Type_gab.gab_desc)
			)
			ch, t
      )

and _mult_float_achain (f, (ch,t)) =
  VAL_achain(
    Core_achain.map
	       (fun c -> function
		  | VAL_undef _ as v   -> v
		  | VAL_int i          -> VAL_float((float_of_int i) *. f)
		  | VAL_float f'       -> VAL_float(f' *. f)
		  | _                  -> raise (Erreur "Achain: _mult_float_achain: don't know how to multiply that a float")
	       )
	       ch, t
  )

and combine_mgsval v1 v2 = match (v1,v2) with
  | VAL_undef "", VAL_undef ""               -> VAL_undef ""
  | VAL_undef _, VAL_undef _                 -> VAL_undef "Recombination value"
  | _, VAL_undef _                           -> v1
  | VAL_undef _, _                           -> v2
  | VAL_int i1, VAL_int i2                   -> VAL_int(i1+i2)
  | VAL_float f1, VAL_float f2               -> VAL_float(f1+.f2)
  | VAL_achain(c1',t1'), VAL_achain(c2',t2') -> _add_achain_achain ((c1',t1'),(c2',t2'))
  | _                                        -> VAL_gab(Core_gab.add_g v1 (Core_gab.add_g v2 Core_gab.empty_g), Type_gab.gab_desc)
      
and _add_achain_achain ((c1,t1),(c2,t2)) =
  let t = Type.commontype t1 t2 in
  let h = Core_achain.copy c1 in
    Core_achain.iter (fun c v -> Core_achain.update_achain h combine_mgsval v c) c2 ;
    VAL_achain(h,t)

and _sub_achain_achain ((c1,t1),(c2,t2)) =
  let t = Type.commontype t1 t2 in
  let h = Core_achain.copy c1 in
  let combine_mgsval v1 v2 = match (v1,v2) with
    | VAL_undef "", VAL_undef "" -> VAL_undef ""
    | VAL_undef _, VAL_undef _   -> VAL_undef "Recombination value"
    | _, VAL_undef _             -> v1
    | VAL_undef _, _             -> VAL_gab(Core_gab.remove_g v2 Core_gab.empty_g, Type_gab.gab_desc)
    | _                          -> VAL_gab(Core_gab.add_g v1 (Core_gab.remove_g v2 Core_gab.empty_g), Type_gab.gab_desc)
  in
    Core_achain.iter (fun c v -> Core_achain.update_achain h combine_mgsval v c) c2 ;
    VAL_achain(h,t)

and _boundary_acell c =
  let h = Core_achain.new_achain () in
    List.iter (fun f -> ignore(Core_achain.add_acell h f (VAL_int 1))) (faces c) ;
    VAL_achain(h, Type_achain.achain_desc)

and _boundary_achain (ch, t) =
  let h = Core_achain.new_achain () in
    Core_achain.iter
      (fun c v ->
	 List.iter (fun f ->
		      try
			ignore(Core_achain.add_acell h f (
				 match v, Core_achain.find h f with
				   | (VAL_gab _ as g), v -> Dispatch_sig.join v g
				   | _, VAL_undef _    -> v
				   | _, VAL_gab(g, t)  -> VAL_gab(Core_gab.add_g v g, t)
				   | _, vv             -> VAL_gab(Core_gab.add_g vv (Core_gab.add_g v Core_gab.empty_g), Type_gab.gab_desc)))
		      with Not_found -> ignore(Core_achain.add_acell h f v)
		   )
	 (faces c)
      )
      ch ;
    VAL_achain(h, t)

and _coboundary_acell c =
  let h = Core_achain.new_achain () in
    List.iter (fun f -> ignore(Core_achain.add_acell h f (VAL_int 1))) (cofaces c) ;
    VAL_achain(h, Type_achain.achain_desc)

and _coboundary_achain (ch, t) =
  let h = Core_achain.new_achain () in
    Core_achain.iter
      (fun c v ->
	 List.iter (fun f ->
		      try
			ignore(Core_achain.add_acell h f (
				 match (v, Core_achain.find h f) with
				   | VAL_gab _ as g, v -> Dispatch_sig.join v g
				   | _, VAL_undef _    -> v
				   | _, VAL_gab(g, t)  -> VAL_gab(Core_gab.add_g v g, t)
				   | _, vv             -> VAL_gab(Core_gab.add_g vv (Core_gab.add_g v Core_gab.empty_g), Type_gab.gab_desc)))
		      with Not_found -> ignore(Core_achain.add_acell h f v)
		   )
	 (cofaces c)
      )
      ch ;
    VAL_achain(h, t)

and _modulo_achain_int((ch,t), i) =
  VAL_achain(Core_achain.map (fun c -> function
				  VAL_gab(_,_) as g -> (
				    match Dispatch_sig.modulo g (VAL_int i) with
					VAL_gab(g, t) when g=Core_gab.empty_g -> VAL_undef ""
				      | v                                     -> v
				  )
				| v -> v
			     ) ch, t)
;;



(** Extension --------------------------------- **)

let rec extension_achain desc a =
  Type.check_vplist desc.name (to_funct desc.pospredicate) a;
  VAL_achain(Array.fold_left add_elem (Core_achain.new_achain()) a, desc)

and  add_elem ch =  function
  | VAL_seq(aa, _) -> Core_achain.add_acell ch (to_acell aa.(0)) aa.(1)
  | _ -> raise (Erreur "Internal Error: extension: add_elem in extension_achain")
;;


(*
 * 2 : Prdicats, accs et observateur
*)

let _dim_achain (c,_) = VAL_int(Core_achain.get_dim_from_achain c)

let cmp_entry v1 v2 = compare_mgs_value v1 v2

let _compareMgsVal_achain_achain  ((x, tx), (y, ty)) =
  let tdiff = Valeur.compare_type tx ty in
    if tdiff <> 0 then VAL_int(tdiff)
    else VAL_int(Core_achain.compare cmp_entry x y)

let _val2bool_achain (c, _) = VAL_bool(Core_achain.size(c) = 0)

let _size_achain (c, _) = VAL_int(Core_achain.size(c))

let _empty_achain (c, _) = VAL_bool(Core_achain.size(c) = 0)

let _count_any_achain (a, (ch, _)) =
  let nb = ref 0
  in
    Core_achain.iter (fun s v -> if (0 = compare_mgs_value v a) then (incr nb)) ch;
    VAL_int(!nb)

let _member_acell_achain (a, (ch, _)) = VAL_bool(Core_achain.mem ch a)

let _hd_achain   (ch, td) = let (sp, v) = Core_achain.hd ch in v

let _tl_achain   (ch, td) = VAL_achain(Core_achain.tl ch, td)

let _last_achain (ch, td) = let (sp, v) = Core_achain.last ch in v

let _take_achain_acell ((ch, _), acell) =
  try
    (Core_achain.find ch acell)
  with Not_found ->
    let err = "Error: 'take: acell not in achain"
    in prerr_endline err; raise(Erreur err)
;;

(*
 * 3 : Fold et autre itrations
*)

let _map_funct_achain(fct, (ch, t)) = 
  VAL_achain(Core_achain.map (fun s v -> fct v) ch, t)
    
and _mapD_funct_achain_int(fct, (ch, t), d) = 
  VAL_achain(Core_achain.map_d d (fun s v -> fct v) ch, t)
    
and  _mapindexed_funct_achain(fct, (ch, t)) =
  VAL_achain(Core_achain.map (fun s v -> (app_2 fct (VAL_acell s) v)) ch, t)
    
and  _mapDindexed_funct_achain_int(fct, (ch, t), d) =
  VAL_achain(Core_achain.map_d d (fun s v -> (app_2 fct (VAL_acell s) v)) ch, t)
    
and _iter_funct_achain(fct, (ch, _)) =
  Core_achain.iter (fun s v -> fct v) ch;
  VAL_undef("iter")
  
and _iterindexed_funct_achain(fct, (ch, _)) =
  Core_achain.iter (fun s v -> (app_2 fct (VAL_acell s) v)) ch;
  VAL_undef("iter")
  
and _iterD_funct_achain_int(fct, (ch, _), d) =
  Core_achain.iter_d (VAL_undef("iterD: default value")) d (fun s v -> fct v) ch;
  VAL_undef("iterD")
  
and _iterDindexed_funct_achain_int(fct, (ch, _), d) =
  Core_achain.iter_d (VAL_undef("iterD: default value")) d (fun s v -> (app_2 fct (VAL_acell s) v)) ch;
  VAL_undef("iterD")
  
and _fold_funct_any_achain (fct, zero, (ch, _)) =
  Core_achain.fold (fun s v suite -> app_2 fct v suite) ch zero

and _foldindexed_funct_any_achain (fct, zero, (ch, _)) = 
  Core_achain.fold (fun s v suite -> app_3 fct (VAL_acell s) v suite) ch zero

and _foldD_funct_any_achain_int (fct, zero, (ch, _), d) =
  Core_achain.fold_d d (fun s v suite -> app_2 fct v suite) ch zero

and _foldDindexed_funct_any_achain_int (fct, zero, (ch, _), d) = 
  Core_achain.fold_d d (fun s v suite -> app_3 fct (VAL_acell s) v suite) ch zero

and _forall_funct_achain (f, (ch, _))  =
  VAL_bool(Core_achain.forall (fun _ v -> Auxfct.bool_from_val (f v)) ch)

and _exists_funct_achain (f, (ch, _))  =
  VAL_bool(Core_achain.exists (fun _ v -> Auxfct.bool_from_val (f v)) ch)

and _forallD_funct_achain_int (f, (ch, _), d)  =
  VAL_bool(Core_achain.forall_d d (fun _ v -> Auxfct.bool_from_val (f v)) ch)

and _existsD_funct_achain_int (f, (ch, _), d)  =
  VAL_bool(Core_achain.exists_d d (fun _ v -> Auxfct.bool_from_val (f v)) ch)

;;

(*
 * 4 : Neighbors, neighbor
 *)

let rec _faces_achain (ch, _)                             = VAL_funct(localDispatch1 ch faces_achain)
and     _cofaces_achain (ch, _)                           = VAL_funct(localDispatch1 ch cofaces_achain)
and     _transitivefaces_achain (ch, _)                   = VAL_funct(localDispatch1 ch transitive_faces_achain)
and     _transitivecofaces_achain (ch, _)                 = VAL_funct(localDispatch1 ch transitive_cofaces_achain)
and     _commonfaces_achain_acell ((ch, _),c)             = VAL_funct(localDispatch2 ch c common_faces_achain)
and     _commoncofaces_achain_acell ((ch, _),c)           = VAL_funct(localDispatch2 ch c common_cofaces_achain)
and     _commontransitivefaces_achain_acell ((ch, _),c)   = VAL_funct(localDispatch2 ch c common_transitive_faces_achain)
and     _commontransitivecofaces_achain_acell ((ch, _),c) = VAL_funct(localDispatch2 ch c common_transitive_cofaces_achain)
and     _icells_achain_acell ((ch, _),c)                  = VAL_funct(localDispatch3 ch c icells_achain)
and     _ccells_achain_acell ((ch, _),c)                  = VAL_funct(localDispatch3 ch c ccells_achain)

and localDispatch1 ch fct sp = match sp with
  | VAL_acell(c) ->
      let sl0 = List.map fst (fct ch c) in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_achain.find ch ac) with
			 | VAL_undef _ -> acc
			 | v           -> v::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Achain: localDispatch1: expect a acell for the second argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch2 ch c1 fct sp = match sp with
  | VAL_acell(c2) ->
      let sl0 = List.map fst (fct ch c1 c2) in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_achain.find ch ac) with
			 | VAL_undef _ -> acc
			 | v           -> v::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Achain: localDispatch2: expect a acell for the third argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch3 ch c fct sp = match sp with
  | VAL_int(i) ->
      let sl0 = List.map fst (fct ch c i) in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_achain.find ch ac) with
			 | VAL_undef _ -> acc
			 | v           -> v::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Achain: localDispatch3: expect a int for the third argument"
      in prerr_endline err; raise(Erreur err)
;;

let rec _facespos_achain (ch, _)                             = VAL_funct(localDispatch4 ch faces_achain)
and     _cofacespos_achain (ch, _)                           = VAL_funct(localDispatch4 ch cofaces_achain)
and     _transitivefacespos_achain (ch, _)                   = VAL_funct(localDispatch4 ch transitive_faces_achain)
and     _transitivecofacespos_achain (ch, _)                 = VAL_funct(localDispatch4 ch transitive_cofaces_achain)
and     _commonfacespos_achain_acell ((ch, _),c)             = VAL_funct(localDispatch5 ch c common_faces_achain)
and     _commoncofacespos_achain_acell ((ch, _),c)           = VAL_funct(localDispatch5 ch c common_cofaces_achain)
and     _commontransitivefacespos_achain_acell ((ch, _),c)   = VAL_funct(localDispatch5 ch c common_transitive_faces_achain)
and     _commontransitivecofacespos_achain_acell ((ch, _),c) = VAL_funct(localDispatch5 ch c common_transitive_cofaces_achain)
and     _icellspos_achain_acell ((ch, _),c)                  = VAL_funct(localDispatch6 ch c icells_achain)
and     _ccellspos_achain_acell ((ch, _),c)                  = VAL_funct(localDispatch6 ch c ccells_achain)

and localDispatch4 ch fct sp = match sp with
  | VAL_acell(c) ->
      let sl0 = List.map fst (fct ch c) in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_achain.find ch ac) with
			 | VAL_undef _ -> acc
			 | _           -> (VAL_acell ac)::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Achain: localDispatch4: expect a acell for the second argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch5 ch c1 fct sp = match sp with
  | VAL_acell(c2) ->
      let sl0 = List.map fst (fct ch c1 c2) in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_achain.find ch ac) with
			 | VAL_undef _ -> acc
			 | _           -> (VAL_acell ac)::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Achain: localDispatch5: expect a acell for the third argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch6 ch c fct sp = match sp with
  | VAL_int(i) ->
      let sl0 = List.map fst (fct ch c i) in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_achain.find ch ac) with
			 | VAL_undef _ -> acc
			 | _           -> (VAL_acell ac)::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Achain: localDispatch6: expect a int for the third argument"
      in prerr_endline err; raise(Erreur err)
;;


(*
 * 5 : Fonctions spcifiques
 *)

let _filtration_funct_achain (fct,(ch,_)) =
  VAL_seq(Array.map
	    (fun c -> VAL_acell c)
	    (Core_achain.filtration
	       ch
	       (fun d v v' -> to_int(app_3 fct (VAL_int d) v v'))
	    ),
	    Type_seq.seq_desc)
;;

let _purge_achain((ch,c)) =
  let ch' = Hash.copy ch in
    Core_achain.iter (fun p v -> match v with VAL_undef _ -> (Hash.remove ch' (acelluid p)) | _ -> () ) ch ;
    VAL_achain(ch',c)
;;

(*
 * 6 : mise  jour du descripteur du type
 *)

achain_desc.xxxify <- VAL_funct(achainify);;
achain_desc.empty  <- VAL_achain(Core_achain.new_achain (), achain_desc);;
achain_desc.make   <- extension_achain achain_desc;;
