(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_bag.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-11-14 09:45:41 $
 * Version: $Revision: 1.6 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *     La representation d'un multi-ensemble (ou bag)
 *     reprsentation adapte  partir des arbres binaires balances (cf. set)
 *     pour tenir compte de l'occurence de chauqe lments.
 *
 *=========================================================================
 *)


open Def_bag;;
open Valeur;;
open Dispatch_sig;;


Global.__register_version "$RCSfile: core_bag.ml,v $" "$Revision: 1.6 $" "$Date: 2007-11-14 09:45:41 $";;



let compare_elt = compare_mgs_value
;;

let empty = Def_bag.Empty_bag
;;


(********************************************************************************)
(********************************************************************************
  ######
  #     #    ##     ####    ####
  #     #   #  #   #    #  #
  ######   #    #  #        ####
  #     #  ######  #  ###       #
  #     #  #    #  #    #  #    #
  ######   #    #   ####    ####
********************************************************************************)
(********************************************************************************)

(* Bags are represented by balanced binary trees (the heights of the
   children differ by at most 2 *)

let height_b = function
    Empty_bag -> 0
  | Node_bag(_, _, _, _, h) -> h

(* Creates a new node with left son l, value x and right son r.
   l and r must be balanced and | height l - height r | <= 2.
   Inline expansion of height for better speed. *)

let create_b l x n r =                                           (* 
								    au lieu d'appeler height il reecrit le code *)
  let hl = match l with Empty_bag -> 0 | Node_bag(_,_,_,_,h) -> h in
  let hr = match r with Empty_bag -> 0 | Node_bag(_,_,_,_,h) -> h in
    Node_bag(l, x,n, r, (if hl >= hr then hl + 1 else hr + 1))

(* Same as create_b, but performs one step of rebalancing if necessary.
   Assumes l and r balanced.
   Inline expansion of create for better speed in the most frequent case
   where no rebalancing is required. *)

let bal_b l x n r =
  let hl = match l with Empty_bag -> 0 | Node_bag(_,_,_,_,h) -> h in
  let hr = match r with Empty_bag -> 0 | Node_bag(_,_,_,_,h) -> h in
    if hl > hr + 2 then begin
      match l with
          Empty_bag -> invalid_arg "Set.bal"
        | Node_bag(ll, lv,ln, lr, _) ->
	    if height_b ll >= height_b lr then
              create_b ll lv ln (create_b lr x n r)
            else begin
              match lr with
                  Empty_bag -> invalid_arg "Set.bal"
		| Node_bag(lrl, lrv,lrn, lrr, _)->
		    create_b (create_b ll lv ln lrl) lrv lrn (create_b lrr x n r)
            end
    end else if hr > hl + 2 then begin
      match r with
          Empty_bag -> invalid_arg "Set.bal"
        | Node_bag(rl, rv,rn , rr, _) ->
	    if height_b rr >= height_b rl then
              create_b (create_b l x n rl) rv rn rr
            else begin
              match rl with
                  Empty_bag -> invalid_arg "Set.bal"
		| Node_bag(rll, rlv,rln, rlr, _) ->
		    create_b (create_b l x n rll) rlv rln (create_b rlr rv rn rr)
            end
    end else
      Node_bag(l, x,n, r, (if hl >= hr then hl + 1 else hr + 1))

(* Same as bal_b, but repeat rebalancing until the final result
   is balanced. *)

let rec join_b l x n r =
  match bal_b l x n r with
      Empty_bag -> invalid_arg "Set.join"
    | Node_bag(l', x',n', r', _) as t' ->
	let d = height_b l' - height_b r' in
          if d < -2 or d > 2 then join_b l' x' n' r' else t'

(* Merge two trees l and r into one.
   All elements of l must precede the elements of r.
   Assumes | height_b l - height_b r | <= 2. *)

let rec merge_b t1 t2 =
  match (t1, t2) with
      (Empty_bag, t) -> t
    | (t, Empty_bag) -> t
    | (Node_bag(l1, v1,n1, r1, h1), Node_bag(l2, v2,n2, r2, h2)) ->
	bal_b l1 v1 n1 (bal_b (merge_b r1 l2) v2 n2 r2)

(* Same as merge_b, but does not assume anything about l and r. *)

let rec concat_b t1 t2 =
  match (t1, t2) with
      (Empty_bag, t) -> t
    | (t, Empty_bag) -> t
    | (Node_bag(l1, v1,n1, r1, h1), Node_bag(l2, v2,n2, r2, h2)) ->
	join_b l1 v1 n1(join_b (concat_b r1 l2) v2 n2 r2)

(* Splitting attention peut-etre probleme *)

let rec split_b x = function
    Empty_bag ->
      (Empty_bag, None,0, Empty_bag)
  | Node_bag(l, v,n, r, _) ->
      let c = compare_elt x v in
        if c = 0 then (l, Some v,n, r)
        else if c < 0 then
          let (ll, vl,nl, rl) = split_b x l in (ll, vl,nl, join_b rl v 
						  n r)
        else
          let (lr, vr,nr, rr) = split_b x r in (join_b l v n lr, 
						vr,nr, rr)

(* Implementation of the bag operations *)

let empty_b = Empty_bag
;;

let is_empty_b = function Empty_bag -> true | _ -> false
;;

let rec mem_b x = function
    Empty_bag -> false
  | Node_bag(l, v,n, r, _) ->
      let c = compare_elt x v in
        c = 0 || mem_b x (if c < 0 then l else r)
;;

let rec num_b x = function
    Empty_bag -> 0
  | Node_bag(l, v,n, r, _) ->
      let c = compare_elt x v in
        if  c = 0 then n else num_b x (if c < 0 then l else r)
;;

let rec add_b x = function
    Empty_bag -> Node_bag(Empty_bag, x,1, Empty_bag, 1)
  | Node_bag(l, v, n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then Node_bag(l, v,n+1, r, h) else
          if c < 0 then bal_b (add_b x l) v n r else bal_b l v n (add_b x r)
;;

let rec addn_b x m= function
    Empty_bag -> Node_bag(Empty_bag, x,m, Empty_bag, 1)
  | Node_bag(l, v,n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then Node_bag(l, v,n+m, r, h) else
          if c < 0 then bal_b (addn_b x m l) v n r else bal_b l v n 
	    (addn_b x m r)
;;

let rec multn_b m= function
    Empty_bag -> Empty_bag
  | Node_bag(l, v, n, r, h) -> Node_bag((multn_b m l), v, m*n, (multn_b m r), h)
;;


let singleton_b x = Node_bag(Empty_bag, x,1, Empty_bag, 1)
;;
let rec remove_b x = function
    Empty_bag -> Empty_bag
  | Node_bag(l, v,n, r, h) ->
      let c = compare_elt x v in
        if c = 0 then (if n=1 then merge_b l r else Node_bag(l, v,n-1, 
							     r, h) ) else
          if c < 0 then bal_b (remove_b x l) v n r else bal_b l v n 
	    (remove_b x r)
;;

let rec union_b s1 s2 =
  match (s1, s2) with
      (Empty_bag, t2) -> t2
    | (t1, Empty_bag) -> t1
    | (Node_bag(l1, v1,n1, r1, h1), Node_bag(l2, v2,n2, r2, h2)) ->
	if h1 >= h2 then
          if h2 = 1 then addn_b v2 n2 s1 else begin
            let (l2, _,nb, r2) = split_b v1 s2 in
              join_b (union_b l1 l2) v1 (n1+nb) (union_b r1 r2)
          end
        else
          if h1 = 1 then addn_b v1 n1 s2 else begin
            let (l1, _,nb, r1) = split_b v2 s1 in
              join_b (union_b l1 l2) v2 (n2+nb) (union_b r1 r2)
          end
;;

let rec inter_b s1 s2 =
  match (s1, s2) with
      (Empty_bag, t2) -> Empty_bag
    | (t1, Empty_bag) -> Empty_bag
    | (Node_bag(l1, v1,n1, r1, _), t2) ->
	match split_b v1 t2 with
            (l2, None, _, r2) ->
	      concat_b (inter_b l1 l2) (inter_b r1 r2)
          | (l2, Some _, n2, r2) ->
	      join_b (inter_b l1 l2) v1 (min n1 n2) (inter_b r1 r2)
;;

let rec diff_b s1 s2 =
  match (s1, s2) with
      (Empty_bag, t2) -> Empty_bag
    | (t1, Empty_bag) -> t1
    | (Node_bag(l1, v1,n1, r1, _), t2) ->
	match split_b v1 t2 with
            (l2, None,_, r2) ->
	      join_b (diff_b l1 l2) v1 n1 (diff_b r1 r2)
          | (l2, Some _,n2, r2) ->
	      if (n2 >= n1) then
		concat_b (diff_b l1 l2) (diff_b r1 r2)
	      else
		join_b (diff_b l1 l2) v1 (n1-n2) (diff_b r1 r2)
;;


let rec compare_b s1 s2 =  compare_aux_b [s1] [s2]
			   
and compare_aux_b l1 l2 =           
  match (l1, l2) with
      ([], []) -> 0
    | ([], _)  -> -1
    | (_, []) -> 1
    | (Empty_bag :: t1, Empty_bag :: t2) -> compare_aux_b t1 t2
    | (Node_bag(Empty_bag, v1,n1, r1, _) :: t1, Node_bag(Empty_bag, v2,n2, r2, _) :: t2) ->
	if n1 = n2 (** On fait le test numrique avant de tester les elts, c'est plus rapide **)
	then let c = compare_elt v1 v2 in if c <> 0 then c else compare_aux_b (r1::t1) (r2::t2)
	else n1-n2 
    | (Node_bag(l1, v1,n1, r1, _) :: t1, t2) ->
	compare_aux_b (l1 :: Node_bag(Empty_bag, v1,n1, r1, 0) :: t1) t2
    | (t1, Node_bag(l2, v2,n2, r2, _) :: t2) ->
	compare_aux_b t1 (l2 :: Node_bag(Empty_bag, v2,n2, r2, 0) :: t2)
	
let equal_b s1 s2 = (0 = compare_b s1 s2)


let rec subset_b s1 s2 =
  match (s1, s2) with
      Empty_bag, _ ->
	true
    | _, Empty_bag ->
	false
    | Node_bag (l1, v1,n1, r1, _), (Node_bag (l2, v2,n2, r2, _) as t2) ->
	let c = compare_elt v1 v2 in
          if c = 0 then
            (n1<=n2) && subset_b l1 l2 && subset_b r1 r2
          else if c < 0 then
            subset_b (Node_bag (l1, v1,n1, Empty_bag, 0)) l2 && subset_b r1 t2
          else
            subset_b (Node_bag (Empty_bag, v1,n1, r1, 0)) r2 && subset_b l1 t2
;;


let rec iter_b f =
  let rec repete n f v = (if n = 1
			  then (f v)
			  else (f v; repete (n-1) f v))
  in
    function
        Empty_bag               -> ()
      | Node_bag(l, v, n, r, _) -> iter_b f l; repete n f v; iter_b f r
;;

let rec iter_occurences_b f = function
    Empty_bag               -> ()
  | Node_bag(l, v, n, r, _) -> iter_occurences_b f l; f (v, n) ; iter_occurences_b f r
;;

let rec fold_b f s accu =
  let rec repete n f v acc = (if n=1 
			      then f v acc 
			      else repete (n-1) f v (f v acc))
  in
    match s with
        Empty_bag -> accu
      | Node_bag(l, v,n, r, _) -> fold_b f l (repete n f v (fold_b f r accu))
;;

let rec fold_occurences_b f s accu =
  match s with
      Empty_bag -> accu
    | Node_bag(l, v,n, r, _) -> fold_occurences_b f l (f (v, n) (fold_occurences_b f r accu))
;;
 


let rec for_all_b p = function
    Empty_bag -> true
  | Node_bag(l, v,n, r, _) -> p v && for_all_b p l && for_all_b p r
;;

let rec exists_b p = function
    Empty_bag -> false
  | Node_bag(l, v,n, r, _) -> p v || exists_b p l || exists_b p r
;;

let filter_b p s =
  let rec filt accu = function
    | Empty_bag -> accu
    | Node_bag(l, v,n, r, _) ->
	filt (filt (if p v then addn_b v n  accu else accu) l) r in
    filt Empty_bag s
;;

let partition_b p s =
  let rec part (t, f as accu) = function
    | Empty_bag -> accu
    | Node_bag(l, v,n, r, _) ->
	part (part (if p v then (addn_b v n t, f) else (t, addn_b v n f)) l) r in
    part (Empty_bag, Empty_bag) s
;;

let rec cardinal_b = function
    Empty_bag -> 0
  | Node_bag(l, v,n, r, _) -> cardinal_b l + n + cardinal_b r
;;

let rec set_cardinal_b = function
    Empty_bag -> 0
  | Node_bag(l, v,_, r, _) -> set_cardinal_b l + 1 + set_cardinal_b r
;;

let rec elements_aux_b accu =
  let rec do_n v n = if n=0 then [] else v::(do_n v (n-1))in
    function
        Empty_bag -> accu
      | Node_bag(l, v,n, r, _) -> elements_aux_b ((do_n v n) @ elements_aux_b accu r) l
;;

let elements_b s = elements_aux_b [] s
;;

let rec elements_occ_aux_b accu = function
    Empty_bag -> accu
  | Node_bag(l, v, n, r, _) -> elements_occ_aux_b ((n,v) :: (elements_occ_aux_b accu r)) l
;;

let elements_occ_b s = elements_occ_aux_b [] s
;;

let rec min_elt_b = function
    Empty_bag -> raise Not_found
  | Node_bag(Empty_bag, v,n, r, _) -> v
  | Node_bag(l, v,n, r, _) -> min_elt_b l
;;

let rec max_elt_b = function
    Empty_bag -> raise Not_found
  | Node_bag(l, v,n, Empty_bag, _) -> v
  | Node_bag(l, v,n, r, _) -> max_elt_b r
;;

let choose_b = min_elt_b
;;



let _compareMgsVal_bag_bag ((s1, t1), (s2, t2)) = 
  let tdiff = compare_type t1 t2 
  in
    if tdiff <> 0 
    then VAL_int(tdiff) 
    else  VAL_int(compare_b s1 s2)
;;


