(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: global.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-07-12 15:14:17 $
 * Version: $Revision: 1.4 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

open Unix;;

(** gestion de l'enregistrement des versions des fichiers **)

let (__register_version, __get_versions) =
  let version       = Hashtbl.create 3
  and cut deb fin l = String.sub l deb ((String.length l) - deb - fin)
  in
    (fun nom vers date -> let nom  = cut 10 4 nom
			  and vers = cut 11 2 vers
			  and date = cut 7 2 date
     in
       Hashtbl.add version nom (vers, date)),
    (function () -> version)
;;

let major_version_name = "2_MGS 2005 (ocaml, eval. by combinators, higher dimensional topologies)"

and more_commande = "/bin/more -d -c "
and rm_commande = "/bin/rm "

and without_gui = ref true
;;

__register_version "$RCSfile: global.ml,v $" "$Revision: 1.4 $" "$Date: 2005-07-12 15:14:17 $";;


(** --- Les exceptions globales ----------------------------------- **)

(** l'exception d'erreur lors de l'evaluation en mgs **)
exception Erreur of string;;
exception MgsAssertionFailed;;


(** sert de break dans les boucles *)
exception BreakLoop;;
exception BreakLoop_1;;
exception BreakLoop_2;;




(** --- Types et fonctions de services ----------------------------------- **)

let is_sys_var name = ('\'' = String.get name 0)

and is_symbol name = ('`' = String.get name 0)
;;


let make_symbol_name s = if is_symbol s then s else String.concat "" ["`"; s]

and pad_name n s =
  let lg = String.length s
  in if lg < n then s ^ (String.make (n-lg) ' ') else s

and new_name =
  let cpt = ref 0 in function name -> (incr cpt; name^"_"^(string_of_int !cpt))

and from_option = function
  | Some x -> x
  | None -> raise (Erreur "Error: internal: from_option: access None instead of something")
;;



(** l1@l2 mais sans rptition d'lment. L'ordre des lments n'a pas d'importance. **)
let union_list l1 l2 =
  List.fold_left (fun suite elem -> (if (List.mem elem suite) then suite else elem::suite)) l2 l1
;;
(** enlve les rptitions. Les lments restent dans le mme ordre. **)
let without_double l =  List.rev (union_list l [])
;;



(* Un fold qui s'applique jusqu'a rencontrer une exception (passee en 
   premier argument. Quand l'exception est levee, on finit le calcul
   avec le zero passe en argument
*)

let rec fold_until_exn exn f zero next =
  try
    let v = next()
    in
      f v (fold_until_exn exn f zero next)
  with x -> if (x = exn) then zero else raise x
;;

let id x = x 
;;

let remove_empty f s x y = 
  match (f x) with
    | "" -> y
    | g  -> g^s^y
;;


let rec clean_comma_string_of_general_list f l = 
  without_comma 1 (List.fold_right (remove_empty f ", ") l "")

and clean_comma_string_of_general_array f l = 
  without_comma 1 (Array.fold_right (remove_empty f ", ") l "")

and without_comma nb r = 
  if ((1+nb) <= String.length r) then (String.sub r 0 ((String.length r) - (1+nb))) else ""
;;

let clean_comma_string_of_string_list l  = clean_comma_string_of_general_list id l
and clean_comma_string_of_string_array l = clean_comma_string_of_general_array id l
;;

let string_of_STRING f = 
  let l = String.length f
  in
    if ((l < 2) || (f.[0] != '"') || (f.[l-1] != '"'))
    then f 
    else String.sub f 1 (l-2)
;;

let of_some = function
  | Some x -> x
  | _      -> raise Not_found;;

let pre_inc v    = (incr v; !v)
and post_inc v   = (let r = !v in incr v; r)
;;


let array_forall p s =
  try
    for i = 0 to Array.length(s) -1 do
      if not((p s.(i))) then raise BreakLoop
    done;
    true
  with BreakLoop -> false

and array_exists p s =
  try
    for i = 0 to Array.length(s) -1 do
      if (p s.(i)) then raise BreakLoop
    done;
    false
  with BreakLoop -> true
;;


(* j'ai besoin d'un nom "simple", i.e. sans un point, 
 * pour ecrire une fonction de dispatch 
 *)
type outBuffer  = Buffer.t;;
type outBufferF = Format.formatter;;
type intarray = int array;;
type intlist = int list;;

let array_in_bound tab i = (i >= 0) & (i < Array.length tab);;





(** --- Variables globales ------------------------------------------------------- **)

let trace_opt      = ref false  (* trace les optimisation *)
and trace_input    = ref false  (* trace la combinatorisation *)
and match_random   = ref 0      (* niveau de random dans le filtrage *)
and delaunay_opt   = ref ""     (* Option passee au calcul de Qhull *)
and abort_on_error = ref true   (* on arrete l'interprete sur une erreur *)
and silent_error   = ref false  (* affichage des erreurs ou pas *)
and exit_on_error  = ref false  (* une erreur interrompt le calcul et retourne au top-level *)
and lrandom_string = ref 11     (* longueur max d'une string gnre au hasard, cf.  _random_string *)
;;

let crandom n =  if n <= 0 then 0 else (if !match_random > 0 then (Random.int n) else 0)
;;


let (temps_start, temps_stop) =
  let debut = ref 0.0
  in ( (function () -> debut := (Unix.times()).tms_utime),
       (function () -> (Unix.times()).tms_utime -. !debut) )
;;

(* declaration forward de la fct d'analyse et d'evaluation: cf. eval.ml *)
let (analyse_et_exec : (bool -> Lexing.lexbuf -> bool) ref) = ref (fun _ _ -> failwith "internal error : forward on analyse_et_exec");;


(** Permet d'enregister des mots supplementaire dans le dico de completion.
 ** Par defaut a ne fait rien. Sera mis a la bonne valeur dans le fichier completion.ml
 **)
let (add_completion_word : (string -> unit)  ref) = ref (fun _ -> ())
;;


(** --- Balise de debut des initialisations ----------------------------------- 
 ** ce fichier est le premier dans la liste de linkage 
 **)
print_string "\n\n(* MGS: Start of initializations *)\n";
