(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: kernel_gbf.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2007-09-24 17:34:59 $
 * Version: $Revision: 1.6 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *          
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Environnement;;
open Dispatch;;

__register_version "$RCSfile: kernel_gbf.ml,v $" "$Revision: 1.6 $" "$Date: 2007-09-24 17:34:59 $";;

Printf.printf "(* loading kernel functions on GBF into user-space ... ";
flush Pervasives.stdout;;

(* === GBF =================================================== *)
(*!*)

(***
  @module GBF 
  The Group Based Field collection
*)

(** 
  @module GBF 
  @name 'smith
  @type gbf_seq
  
  Returns the sequence of integer coefficients of the smith normal 
  form of the GBF type of the argument. An [undef] value is returned 
  if the argument is not a [GBF].
  
  This sequence describes the canonical isomorphic {i Z}-module used to
  represent the elements of the group [G3]. A coefficient [1] means that 
  the corresponding generator is free (i.e. the corresponding subgroup
  is isomorphic to {i Z}). A coffecient {i n} means that the 
  corresponding generator is cyclic  (i.e. the corresponding subgroup
  is isomorphic to {i Z/nZ}).

  For example, given the GBF declaration
  {[ gbf G3 = < a, b, c; 10 b, 33 c > ]}
  the expression [smith(G3:())] returns the sequence {[(1, 330):'seq]}
  which means that the group structure of [G3] is isomorphic to 
  {i Z}[x]{i Z/330Z}. The interpretation is the following: 
  generator [a] is mapped to {i Z} and the subgroup spanned by [b] 
  and [c] is mapped to {i Z/330Z} (note that [330 = 10 * 33]).
*)
addenv_pure "'smith"         (fun_1 __smith);;


(** 
  @module GBF 
  @name 'smith_backpos
  @type posgbf_seq

  Inverse of the ['smith] function. 

  {b !!! Not yet correctly implemented !!!}

  @voir 'smith smith
*)
addenv_pure "'smith_backpos" (fun_1 __backpos);;




(** 
  @module GBF 
  @name 'following
  @type seq_seq_gbf
  @type seq_posgbf_gbf

  Build a GBF from a sequence of values and a sequence of directions.

  [l following d] is the infix form of ['following(l, d)] (note that
  the unquoted name is a keyword and cannot be used as an
  identifier). This expression builds a GBF by enumeration: the values
  are given in the nested sequence [l] and the direction of the
  enumeration are given in the sequence of direction [d].

  Suppose that there are two directions [d1] and [d2] in the
  directions list: [D == d1, d2]. Then, the sequence of values must be a
  sequence of sequences of elements [l == ((L00, L01, L02)::(L10, L11,
  L12, L13)::(L20, L21)::seq:())]. Then, the expression [l following d]
  builds a GBF [g] that is build in the following manner: the value
  [Lij] becomes the value of position [i*D1 + j*D2]. And more generally,
  [Lij...k] becomes the value of position [i*D1 + j*D2 + ... + k*Dn].

  Perhaps a better explanation is given considering a path: the
  directions list specifies a path starting from the position [0] and
  the values in the first argument of the following becomes the values
  of the positions crossed along this path. The direction [Di] are
  followed for each element in the list or each time we go from a
  sublist to another sublist. For our example,
  {ul
  {- we start from the position [0] (called the initial starting
  point) and we give the value [L00] to this position,}
  {- we move towards the direction [d1] and give to the current
  position the value [L01],}
  {- we continue this way, until we reach the end of the sublist.}
  {- Then, jump to our initial position [0] and we move following
  [d2]. This point becomes the new starting point.}
  {- We repeat the previous procedure again, that is, the current
  position takes the value [L10], we move following [d1], the current
  position takes the value [L11], etc.}
  {- At the end of the current sublist, we move to our current
  starting point and we move again following [d2] to obtain our new
  starting point. And we proceed to affect the values of sublist
  [l2x].}
  {- And so on, until the exhaustion of the sequences of values.}
  }
  More formally,


  @example 
  Suppose we have defined a GBF [gbf GG = <X, Y>], then 
  {[(a, a, a) :: (b, b) :: (c, c, c, c) :: seq:() following  |X>, |Y>]}
  returns the GBF
  {[a, a, a]}
  {[b, b]}
  {[c, c, c, c]} 
  (we assume that the [|X>] dimension is mapped horizontally and [|Y>]
  is mapped vertically).  
*)
addenv_impure "'following"    (fun_2 __following);;

addenv_impure "'neighbor" (fun_3 __neighbor);;

(*!*)
(* ============================================================ *)

(* --- fin de l'installation des fonctions --- *)

Printf.printf "done *)\n";

(* ======================================================= *)


(** 
  @module GBF 
  @name 'set_gbfpos
  @type gbf_posgbf_any

  Function with side effect that modify the definition of a gbf at a given, replacing the old value by the third argument.
*)
addenv_impure "'set_gbfpos" (fun_3 __setGBFPos);;
