(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: prox.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    sequences.
 *
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> seq (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = gbf
 *
 *=========================================================================
 *)


open Global;;
open Valeur;;
open Type;;
open Auxfct;;
open Dispatch_sig;;
open Type_prox ;;
open Core_prox ;;
open Def_prox ;;
open Printer;;

__register_version "$RCSfile: prox.ml,v $" "$Revision: 1.2 $" "$Date: 2005-11-23 11:10:57 $";;


let rec _join_prox_prox ((tab, c), (tab', c')) = (if (c.name <> c'.name)
						  then raise (Erreur (Printf.sprintf "cannot join proximal of different names %s and %s"
									c.name
									c'.name))
						  else (let f        = get_defs_from_proximal c.name in
							let sommets  = Array.map deco tab
							and sommets' = Array.map deco tab'
							in
							  proximal_from_all_sommets (Array.append sommets sommets') f c))


and _join_any_prox (s, (tab, c)) = (let f        = get_defs_from_proximal c.name in
				    let sommets  = array_of_mgs_value s in
				    let sommets' = Array.map deco tab
				    in
				      proximal_from_all_sommets (Array.append sommets sommets') f c)

and _join_prox_any (p, a)             = _join_any_prox (a, p)

and _proxify_prox_any (d, a)          = _join_any_prox (a, d)

and _cons_any_prox x                  = _join_any_prox x

and _size_prox (tab, _)               = VAL_int(Array.length tab)

and _hd_prox (tab, _ )                = Core_graph_generique.hd tab deco

and _tl_prox (tab, c)                 = (let ntab  = (let taille = Array.length tab
						      in
							if (taille = 0)
							then raise (Erreur "tl : cannot be applied on an empty proximal")
							else (Array.map deco (Array.sub tab 1 (taille-1)))) in
					 let fonc  = get_defs_from_proximal c.name
					 in
					   proximal_from_all_sommets ntab fonc c)

and _last_prox (tab, _)               = Core_graph_generique.last tab deco

and _exists_funct_prox (f, (tab, _))  = Core_graph_generique.exists tab deco f
and _forall_funct_prox (f, (tab, _))  = Core_graph_generique.forall tab deco f

and _count_any_prox (a, (tab, _))     = Core_graph_generique.count tab deco a

and _member_any_prox (a, (tab, _))    = Core_graph_generique.member  tab deco a

and _compareMgsVal_prox_prox (g1, g2) = VAL_int (Core_graph_generique.compare deco vois g1 g2)

and _isomorphism_prox_prox (g1, g2)   = VAL_bool (0 = Core_graph_generique.compare_iso vois g1 g2)

and _neighbors_prox_int ((t, _), i)   = VAL_seq ((Core_graph_generique.neighbors deco vois t i), Type_seq.seq_desc)

and _neighborspos_prox_int ((t, _), i) = (assert ((i >= 0) && (i < (Array.length t)));
					  let _, voisins = t.(i)
					  in
					    VAL_seq ((Array.of_list (List.map (fun x -> VAL_int x) voisins)), Type_seq.seq_desc))


and _map_funct_prox (fonc, (tab, c)) = (let fonction x = fonc (deco x) in
					let ntab       = Array.map fonction tab
					in
					  proximal_from_all_sommets ntab (get_defs_from_proximal c.name) c)
					 


and _mapindexed_funct_prox (fonc, (tab, d)) = (let ntab = Array.mapi (fun index  (v, _) -> (app_2 fonc (VAL_int index) v)) tab
					       in
						 proximal_from_all_sommets ntab (get_defs_from_proximal d.name) d)

and _iter_funct_prox (fonc, (tab, _))                  = Core_graph_generique.iter        fonc deco tab
and _iterindexed_funct_prox (fonc, (tab, _))           = Core_graph_generique.iterindexed fonc deco tab
and _fold_funct_any_prox (fonc, zero, (tab, _))        = Core_graph_generique.fold        fonc deco tab zero
and _foldindexed_funct_any_prox (fonc, zero, (tab, _)) = (let idx = ref ((Array.length tab) - 1)
							  in
							    Array.fold_right (fun v s -> (let local = VAL_int !idx
											  in
											    decr idx;
											    (app_3 fonc local (deco v) s))) tab zero)


and _setify_prox   (tab, _)         = VAL_set(Core_graph_generique.setify  tab deco, Type_set.set_desc)
and _bagify_prox   (tab, _)         = VAL_bag(Core_graph_generique.bagify  tab deco, Type_bag.bag_desc)
and _sequify_prox  (tab, _)         = VAL_seq(Core_graph_generique.sequify tab deco, Type_seq.seq_desc)
and _graphify_prox (x, deco)        = (let tbl = Hash.create 1
				       in
					 VAL_grf ((tbl, x), Type_graph.graph_desc))


and _empty_prox   (tab, _)          = VAL_bool (0 = Array.length tab)
;;

