(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Antopine Spicher    : aspicher@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: qmf.old.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2005-01-13 00:59:54 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. Qmf/def_qmf.ml pour les infos sur les types qmf et ncell
 *          
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> seq (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = seq
 *
 *=========================================================================
 *)

open Global ;;
open Valeur ;;
open Printer ;;
open Dispatch_sig ;;

open Def_qmf ;;
open Type_qmf ;;
open Core_qmf ;;

__register_version "$RCSfile: qmf.old.ml,v $" "$Revision: 1.1 $" "$Date: 2005-01-13 00:59:54 $";;


(** === fonctions utiles ======================================================= **)

(*****************************************************************************************************
  Etant donne :
  _ une mgs_value (VAL_int ou VAL_float)

  => retourne
               _ la valeur float associee (avec conversion int -> float si necessaire)

  Par exemple : (VAL_float 2.3) => 2.3
                (VAL_int 2)     => 2.0

  ATTENTION: existe deja. Cf. floatof dans kernel.ml. Et je crois que ca traite aussi les string. (JLG)
*****************************************************************************************************)

let float_of_mgs_value = function
  | VAL_int i   -> float_of_int i
  | VAL_float f -> f
  | _           -> raise (Erreur "illegal use of float_of_mgs_value")
;;

let ncell_of_mgs_value = function
  | VAL_ncell c -> c
  | _           -> raise (Erreur "illegal use of ncell_of_mgs_value")
;;

(** === Wrapping des gmap ======================================================= **)


let _size_gmap _ = VAL_int (-9) ;;
let _size_ncell _ = VAL_int (-10) ;;

let _ncellPredicate_ncell _ = mgstrue
let _ncellPredicate_any _   = mgsfalse

let _gmapPredicate_gmap _ = mgstrue
let _gmapPredicate_any _  = mgsfalse

let _val2bool_ncell c = VAL_bool(Core_qmf.ncelldim c = -1)
let _val2bool_gmap  g = mgstrue

let _compareMgsVal_gmap_gmap (g1, g2) = VAL_int(Pervasives.compare g1 g2) ;;
let _compareMgsVal_ncell_ncell (c1, c2) = VAL_int(Pervasives.compare c1 c2) ;;

let _dim_ncell c = VAL_int(Core_qmf.get_dim_from_ncell c) ;;



let rec _faces_ncell c =
  if ((ncelldim c) = 0) then
    VAL_seq([||], Type_seq.seq_desc)
  else
    VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.faces c), Type_seq.seq_desc)

and _cofaces_ncell c =
  if ((ncelldim c) = 3) then
    VAL_seq([||], Type_seq.seq_desc)
  else
    VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.cofaces c), Type_seq.seq_desc)

and _icells_ncell_int (c,i) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.icells  c i), Type_seq.seq_desc)

and _ccells_ncell_int (c,v) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.ccells c v), Type_seq.seq_desc)

and _transitivefaces_ncell c =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.transitivefaces c), Type_seq.seq_desc) 

and _transitivecofaces_ncell c =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.transitivecofaces c), Type_seq.seq_desc) 

and _commonfaces_ncell_ncell (c1, c2) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.commonfaces c1 c2), Type_seq.seq_desc)

and _commoncofaces_ncell_ncell (c1, c2) = 
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.commoncofaces c1 c2), Type_seq.seq_desc)

and _commontransitivefaces_ncell_ncell (c1, c2) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.commontransitivefaces c1 c2), Type_seq.seq_desc)

and _commontransitivecofaces_ncell_ncell (c1, c2) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.commontransitivecofaces c1 c2), Type_seq.seq_desc)



and _copyGmap_gmap g = VAL_gmap(Core_qmf.copy_gmap g)

and _addNcell_gmap_int (g, i) = VAL_ncell (new_ncell (g,i)) 

and _addNcellWithCopy_gmap_int (g, i) = VAL_ncell (new_ncell_with_copy (g,i)) 

and _addVertex_gmap g = VAL_ncell (new_vertex g)

and _addVertexWithCopy_gmap g = VAL_ncell (new_vertex_with_copy g) 

and _addEdge_gmap g = VAL_ncell (new_edge g)

and _addEdgeWithCopy_gmap g   = VAL_ncell (new_edge_with_copy g) 

and _addFace_gmap g = VAL_ncell (new_face g) 

and _addFaceWithCopy_gmap g   = VAL_ncell (new_face_with_copy g) 

and _addVolume_gmap g = VAL_ncell (new_volume g)

and _addVolumeWithCopy_gmap g = VAL_ncell (new_volume_with_copy g)



and _addPolygon_gmap_int (g, i) = VAL_ncell (new_ncell_polygon (g,i))

and _addPolygonWithCopy_gmap_int (g,i) = VAL_ncell (new_ncell_polygon_with_copy (g,i))

and _addSquare_gmap_int_int_int (g,ax,ay,dim) = VAL_ncell (new_ncell_square g ax ay dim)

and _addSquareWithCopy_gmap_int_int_int (g,ax,ay,dim) = VAL_ncell (new_ncell_square_with_copy g ax ay dim)

and _addSphere_gmap_int_int (g, m, p) = let (c1,c2) = new_ncell_sphere g m p in
  VAL_seq([|VAL_ncell(c1);VAL_ncell(c2)|],Type_seq.seq_desc)

and _addSphereWithCopy_gmap_int_int (g,m,p) = let (c1,c2) = new_ncell_sphere_with_copy g m p in
  VAL_seq([|VAL_ncell(c1);VAL_ncell(c2)|],Type_seq.seq_desc)

and _addPolyline_gmap_int (g, i) = VAL_ncell (new_ncell_polyline g i)

and _addPolylineWithCopy_gmap_int (g, i) = VAL_ncell (new_ncell_polyline_with_copy g i)




and _glue2FacesBy4Verteces_ncell_ncell_ncell_ncell_ncell_ncell (c1,c2,c3,c4,c5,c6) =
  glue2FacesBy4Verteces((c1,c2),(c3,c4),(c5,c6));
  VAL_undef "dummy return for function glue2FacesBy4Verteces"

and _getCoordFromVertex_ncell c =
  if (ncelldim c = 0) then
    let coord = Array.map (fun x -> VAL_float(x)) (Core_qmf.get_coord_from_vertex c) in
      VAL_seq(coord,Type_seq.seq_desc)
  else
    VAL_undef "arg must be a vertex (dimension0)"

and _setCoordOfVertex_ncell_seq (c,(s,_)) =
  if (ncelldim c = 0) then
    if (Array.length s = 3) then (
      Core_qmf.set_coord_of_vertex c (float_of_mgs_value s.(0)) (float_of_mgs_value s.(1)) (float_of_mgs_value s.(2));
      VAL_undef "dummy return for function set_coord_of_vertex"
    )
    else
      VAL_undef "the size of arg2 must be 3"
  else
    VAL_undef "arg1 must be a vertex (dimension0)"

and _getGmapFrom_ncell (g,_,_) = VAL_gmap g

and _normalise_funct_ncell (f,c) = VAL_ncell(Core_qmf.normal_form (c))

and _loadAse_gmap_string (g,s) =  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.load_ase g s),Type_seq.seq_desc)

and _saveAse_gmap_string (g,s) =  Core_qmf.save_ase g s ; VAL_undef "dummy return for function save_ase"

and _loadMoka_gmap_string (g,s) =  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.load_moka g s),Type_seq.seq_desc)

and _saveMoka_gmap_string (g,s) =  Core_qmf.save_moka g s ; VAL_undef "dummy return for function save_moka"






and _insertVertex_ncell c =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.insert_vertex c),Type_seq.seq_desc)

and _insertEdge_ncell_ncell (c1,c2) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.insert_edge c1 c2),Type_seq.seq_desc)

and _insertFace_ncell_seq (c,(s,_)) =
  let ss = Array.map ncell_of_mgs_value s in
  let ll = Array.length s in
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.insert_face c ss ll),Type_seq.seq_desc)

and _createEdge_ncell_ncell (c1,c2) =
  VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.create_edge c1 c2),Type_seq.seq_desc)

and _createFace_seq (s,_) =
  let ss = Array.map
	     (function VAL_seq(x,_) ->
		let l = Array.length x in
		  if (l=2) then
			(ncell_of_mgs_value (x.(0)),ncell_of_mgs_value (x.(1)))
		  else
		    raise (Erreur "Gmap: createFace: need a sequence of couple of vertices1")
		| _ -> raise (Erreur "Gmap: createFace: need a sequence of couple of vertices2")
	     )
	     s
  in
  let ll = Array.length ss in
    VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.create_face ss ll),Type_seq.seq_desc)

and _removeNcell_ncell c =
  VAL_ncell(Core_qmf.remove_ncell c)

and _deleteNcell_ncell c =
    VAL_seq(Array.map (fun x -> VAL_ncell(x)) (Core_qmf.delete_ncell c),Type_seq.seq_desc)

and _createVolume_seq_seq ((f,_),(s,_)) =
  let ss = Array.map
	     (function VAL_seq(x,_) ->
		let l = Array.length x in
		  if (l>0) then
			(Array.map ncell_of_mgs_value x)
		  else
		    raise (Erreur "Gmap: createVolume: need a sequence of sequence of vertices1")
		| _ -> raise (Erreur "Gmap: createFace: need a sequence of sequence of vertices2")
	     )
	     s
  in
  let llt = (Array.map Array.length ss) in
  let ll = Array.length ss in
    Core_qmf.create_volume (Array.map ncell_of_mgs_value f) (Array.length f) ss llt ll;
    VAL_undef("dummy return of the function create_volume")
;;


(* Fonctions dispatches des qmf *)

(*
 * 1 : Fonctions de construction
*)

let rec _sequify_qmf (q,_) =
  VAL_seq(Array.of_list (Core_qmf.fold (fun c v suite -> v::suite) q []),Type_seq.seq_desc)

and _qmfify_funct_achain a = VAL_funct(withfun a)
and withfun (f,c) a = match a with
  | VAL_qmf q -> qmfify_achain_qmf_funct (c,q,f)
  | _ -> 
      let err = "Error: Qmf.qmfify: type error"
      in prerr_endline err; raise(Erreur err)

and _qmfify_achain_qmf (c,q) = qmfify_achain_qmf_funct (c,q,(fun _ -> VAL_seq([|VAL_float 0.;VAL_float 0.;VAL_float 0.|],Type_seq.seq_desc)))

and qmfify_achain_qmf_funct ((c,_),(q,t),f) =
  let ff v = match (f v) with
    | VAL_seq(a,_) -> (float_of_mgs_value a.(0), float_of_mgs_value a.(1), float_of_mgs_value a.(2))
    | _ ->
	let err = "Error: Qmf.qmfify: waiting for a seq of 3 floats..."
	in prerr_endline err; raise(Erreur err)
  in
    VAL_qmf((Core_achain.to_qmf ff c q (VAL_undef "no initial val")),t)

and _shape_qmf (q,t) = 
  let g = Core_qmf.get_gmap_from_qmf q in
  let d = Core_qmf.get_dim_from_qmf q in
  let taille = Core_qmf.size q in
  let r   = Array.make taille (default_ncell g)
  and idx = ref (taille-1) in 
    begin
      (* on remplit r avec la suite des cls de la table. La fonction
       * sequify suit le mme parcourt et donc la cle en 2eme position
       * correspond a l'lment en 2me position dans le sequify.
       *)
      Core_qmf.iter (fun c _ -> r.(!idx) <- c; decr idx) q;
      (* On renvoie une fonction qui reconstruit la collection *)
      VAL_funct (function
		   | VAL_seq (s, _) -> 
		       if (taille <> Array.length s)
		       then raise (Erreur "shape (qmf) : argument is not of the same size of the original")
		       else (let nq = Core_qmf.create_qmf g d in
			       Array.iteri (fun idx v -> Core_qmf.add nq (r.(idx)) v) s;
			       VAL_qmf (nq, t)
			    )
		   | _  -> raise (Erreur "Error: shape (qmf): a sequence as argument is required"))
    end

and _bagify_qmf  (q,_) = 
  VAL_bag(Core_qmf.fold (fun c v suite -> Core_bag.add_b v suite) q Core_bag.empty_b, Type_bag.bag_desc)
;;

(** Extension --------------------------------- **)

let rec extension_qmf g desc a =
  Type.check_vplist desc.name (to_funct desc.pospredicate) a;
  VAL_qmf(Array.fold_left add_elem (Core_qmf.new_qmf g) a, desc)

and  add_elem q =  function
  | VAL_seq(aa, _) -> Core_qmf.add_ncell q (to_ncell aa.(0)) aa.(1)
  | _ -> raise (Erreur "Internal Error: extension: add_elem in extension_qmf")
;;

(*
 * 2 : Prdicats, accs et observateur
*)

let _dim_qmf (c,_) = VAL_int(Core_qmf.get_dim_from_qmf c)

let cmp_entry v1 v2 =  compare_mgs_value v1 v2

let _compareMgsVal_qmf_qmf  ((qx,tx),(qy,ty)) =  
  let tdiff = compare_type tx ty in
    if tdiff <> 0 then VAL_int(tdiff) 
    else
      VAL_int(Core_qmf.compare cmp_entry qx qy)

let _val2bool_qmf (q, _) = VAL_bool(Core_qmf.size(q) = 0)

let _size_qmf (q,_) = VAL_int(Core_qmf.size q)

let _empty_qmf (q, _) = VAL_bool(Core_qmf.size(q) = 0)

and _count_any_qmf  (a, (q,_)) = 
  let nb = ref 0
  in
    Core_qmf.iter (fun c v -> if (0 = compare_mgs_value v a) then (incr nb)) q;
    VAL_int(!nb)

let _member_ncell_qmf (a, (ch, _)) = VAL_bool(Core_qmf.mem ch a)

let _hd_qmf   (q, td) = let (c, v) = Core_qmf.hd q in v

let _tl_qmf   (q, td) = VAL_qmf(Core_qmf.tl q, td)

let _last_qmf (q, td) = let (c, v) = Core_qmf.last q in v

and _take_qmf_ncell ((q,_), ncell) = 
  try 
    (Core_qmf.find q ncell)
  with Not_found -> 
    let err = "Error: 'take: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)
;;

(*
 * 3 : Fold et autre itrations
*)

let _map_funct_qmf(fct, (q, t)) =
  VAL_qmf(Core_qmf.map (fun c v -> fct v) q, t)

and  _mapindexed_funct_qmf(fct, (q, t)) =
  VAL_qmf(Core_qmf.map (fun c v -> (app_2 fct (VAL_ncell c) v)) q,t)

and _iter_funct_qmf(fct, (q, _)) =
  Core_qmf.iter (fun c v -> fct v) q;
  VAL_undef("iter")

and _iterindexed_funct_qmf(fct, (q, _)) =
  Core_qmf.iter (fun c v -> (app_2 fct (VAL_ncell c) v) ) q;
  VAL_undef("iter")

and _fold_funct_any_qmf (fct, zero, (q, _)) =
  Core_qmf.fold (fun c v suite -> app_2 fct v suite) q zero

and _foldindexed_funct_any_qmf (fct, zero, (q,_)) =
  Core_qmf.fold (fun c v suite -> app_3 fct (VAL_ncell c) v suite) q zero

and _forall_funct_qmf (f, (q,_))  =
  VAL_bool(Core_qmf.forall (fun _ v -> Auxfct.bool_from_val (f v)) q)

and _exists_funct_qmf (f, (q,_))  =
  VAL_bool(Core_qmf.exists (fun _ v -> Auxfct.bool_from_val (f v)) q)

;;

(*
 * 4 : Neighbors, neighbor
 *)

let rec _neighbors_qmf_ncell ((q,_), c) = (* retourne les valeurs des faces et cofaces *)
(*  let lst = Array.fold_right
	      (fun nc suite-> try (Core_qmf.find q nc)::suite with Not_found -> suite)
	      (faces c)
	      (Array.fold_right
		 (fun nc suite-> try (Core_qmf.find q nc)::suite with Not_found -> suite)
		 (cofaces c)
		 [])*)
  let lst = Array.map snd (Array.append (faces_qmf q c)  (cofaces_qmf q c))
  in VAL_seq(lst, Type_seq.seq_desc)

and _neighbor_qmf_ncell_bool ((q,_), c, b) =
  let lst = Array.fold_right
	      (fun nc suite-> try (Core_qmf.find q nc)::suite with Not_found -> suite)
	      (if b then (cofaces c) else (faces c) )
	      []
  in VAL_seq(Array.of_list lst, Type_seq.seq_desc)

and _neighbor_qmf_ncell_int (q, c, n) =  _neighbor_qmf_ncell_bool(q, c, (n > 0))


and _faces_qmf (q, _)                             = VAL_funct(localDispatch1 q faces_qmf)
and _cofaces_qmf (q, _)                           = VAL_funct(localDispatch1 q cofaces_qmf)
and _transitivefaces_qmf (q, _)                   = VAL_funct(localDispatch1 q transitive_faces_qmf)
and _transitivecofaces_qmf (q, _)                 = VAL_funct(localDispatch1 q transitive_cofaces_qmf)
and _commonfaces_qmf_ncell ((q, _),c)             = VAL_funct(localDispatch2 q c common_faces_qmf)
and _commoncofaces_qmf_ncell ((q, _),c)           = VAL_funct(localDispatch2 q c common_cofaces_qmf)
and _commontransitivefaces_qmf_ncell ((q, _),c)   = VAL_funct(localDispatch2 q c common_transitive_faces_qmf)
and _commontransitivecofaces_qmf_ncell ((q, _),c) = VAL_funct(localDispatch2 q c common_transitive_cofaces_qmf)
and _icells_qmf_ncell ((q, _),c)                  = VAL_funct(localDispatch3 q c icells_qmf)
and _ccells_qmf_ncell ((q, _),c)                  = VAL_funct(localDispatch3 q c ccells_qmf)

and localDispatch1 q fct sp = match sp with
  | VAL_ncell(c) ->
      let l = Array.to_list (fct q c) in
      let sl0 = List.map fst l in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_qmf.find q ac) with
			 | VAL_undef _ -> acc
			 | v           -> v::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Qmf: localDispatch1: expect a ncell for the second argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch2 q c1 fct sp = match sp with
  | VAL_ncell(c2) ->
      let l = Array.to_list (fct q c1 c2) in
      let sl0 = List.map fst l in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_qmf.find q ac) with
			 | VAL_undef _ -> acc
			 | v           -> v::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Qmf: localDispatch2: expect a ncell for the third argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch3 q c fct sp = match sp with
  | VAL_int(i) ->
      let l = Array.to_list (fct q c i) in
      let sl0 = List.map fst l in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_qmf.find q ac) with
			 | VAL_undef _ -> acc
			 | v           -> v::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Qmf: localDispatch3: expect a int for the third argument"
      in prerr_endline err; raise(Erreur err)

and _facespos_qmf (q, _)                             = VAL_funct(localDispatch4 q faces_qmf)
and _cofacespos_qmf (q, _)                           = VAL_funct(localDispatch4 q cofaces_qmf)
and _transitivefacespos_qmf (q, _)                   = VAL_funct(localDispatch4 q transitive_faces_qmf)
and _transitivecofacespos_qmf (q, _)                 = VAL_funct(localDispatch4 q transitive_cofaces_qmf)
and _commonfacespos_qmf_ncell ((q, _),c)             = VAL_funct(localDispatch5 q c common_faces_qmf)
and _commoncofacespos_qmf_ncell ((q, _),c)           = VAL_funct(localDispatch5 q c common_cofaces_qmf)
and _commontransitivefacespos_qmf_ncell ((q, _),c)   = VAL_funct(localDispatch5 q c common_transitive_faces_qmf)
and _commontransitivecofacespos_qmf_ncell ((q, _),c) = VAL_funct(localDispatch5 q c common_transitive_cofaces_qmf)
and _icellspos_qmf_ncell ((q, _),c)                  = VAL_funct(localDispatch6 q c icells_qmf)
and _ccellspos_qmf_ncell ((q, _),c)                  = VAL_funct(localDispatch6 q c ccells_qmf)

and localDispatch4 q fct sp = match sp with
  | VAL_ncell(c) ->
      let l = Array.to_list (fct q c) in
      let sl0 = List.map fst l in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_qmf.find q ac) with
			 | VAL_undef _ -> acc
			 | _           -> (VAL_ncell ac)::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Qmf: localDispatch4: expect a ncell for the second argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch5 q c1 fct sp = match sp with
  | VAL_ncell(c2) ->
      let l = Array.to_list (fct q c1 c2) in
      let sl0 = List.map fst l in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_qmf.find q ac) with
			 | VAL_undef _ -> acc
			 | _           -> (VAL_ncell ac)::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Qmf: localDispatch5: expect a ncell for the third argument"
      in prerr_endline err; raise(Erreur err)

and localDispatch6 q c fct sp = match sp with
  | VAL_int(i) ->
      let l = Array.to_list (fct q c i) in
      let sl0 = List.map fst l in
      let sl1 = List.fold_right
		  (fun ac acc ->
		     try (
		       match (Core_qmf.find q ac) with
			 | VAL_undef _ -> acc
			 | _           -> (VAL_ncell ac)::acc
		     ) with Not_found -> acc
		  )
		  sl0
		  []
      in VAL_seq(Array.of_list(sl1), Type_seq.seq_desc)
  | _ ->
      let err = "Error: Qmf: localDispatch6: expect a int for the third argument"
      in prerr_endline err; raise(Erreur err)
;;

(*
 * 5 : Fonctions spcifiques
 *)

let _getGmapFrom_qmf (q,_) = VAL_gmap(Core_qmf.get_gmap_from_qmf q)
;;

(*
 * 6 : mise  jour du descripteur du type
 *)

qmf_desc.xxxify <- (fun_2 qmfify);;
(*qmf_desc.empty  <- VAL_qmf(Core_qmf.new_qmf (), qmf_desc);;*)
(*qmf_desc.make   <- extension_qmf qmf_desc;;*)


















(***********************************************************************************)
(***********************************************************************************)
(***********************************************************************************)
(***********************************************************************************)
(*
and _setify_qmf  (q,_) =
  VAL_set(Core_qmf.fold (fun c v suite -> Core_set.add v suite) q Core_set.empty, Type_seq.set_desc)

and _dim_qmf (q,_) = VAL_int(dimension q)


;;


(** === Wrapping des qmf =========================================================== **)





(* --- Extension de fonctions gnriques --------------------------------------------- *)

;;


*)
