(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_rec.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-08-16 20:52:45 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *          
 *  A partir du module map.ml de la bibliotheque standard OCAML
 *  modifie pour rendre le type utilisable dans un type recursif.
 *
 *     La representation d'un record: reprsentation par map.
 *     Code  partir du module map de la librairie CAML, augment de plusieurs 
 *     fonctions.
 *
 *     Un record MGS est un dictionnaire qui associe une valeur MGS  une cl.
 *     La cl est n'importe quelle valeur MGS, mme si une syntaxe spciale
 *     est reserve au cls qui sont des symboles.
 *     L'acces se fait par:  r.(key)  dans la syntaxe MGS et 
 *        r.(`key) == r.key  
 *     on retouve donc l'acces par "nom du champ" des records habituels.
 *
 *     La topologie d'un record est celle d'un graphe sans artes: 
 *     tous les sommets sont isols, sans voisins. La position est la cl.
 *
 *=========================================================================
 *)

open Global;;
open Def_rec;;
open Valeur;;
open Dispatch_sig;;

__register_version "$RCSfile: core_rec.ml,v $" "$Revision: 1.1 $" "$Date: 2004-08-16 20:52:45 $";;


(** Les map sont des dictionnaires qui associent une cl  une valeur.
 ** Ici on gnralise par rapport  1_MGS. Dans 1_MGS la cl est une chaines de caractre 
 ** et donc, la fonction de comparaison systme convient bien  la comparaison des cls.
 **
 ** Dans 2_MGS la cl est une valeur mgs quelconque.
 **
 **)
let compare_key = Dispatch_sig.compare_mgs_value
;;


let empty = Empty_map

let height = function
    Empty_map -> 0
  | Node_map(_,_,_,_,h) -> h

let create l x d r =
  let hl = height l and hr = height r in
    Node_map(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))
      
let bal l x d r =
  let hl = match l with Empty_map -> 0 | Node_map(_,_,_,_,h) -> h in
  let hr = match r with Empty_map -> 0 | Node_map(_,_,_,_,h) -> h in
    if hl > hr + 2 then begin
      match l with
          Empty_map -> invalid_arg "Map.bal"
        | Node_map(ll, lv, ld, lr, _) ->
            if height ll >= height lr then
              create ll lv ld (create lr x d r)
            else begin
              match lr with
                  Empty_map -> invalid_arg "Map.bal"
		| Node_map(lrl, lrv, lrd, lrr, _)->
                    create (create ll lv ld lrl) lrv lrd (create lrr x d r)
            end
    end else if hr > hl + 2 then begin
      match r with
          Empty_map -> invalid_arg "Map.bal"
        | Node_map(rl, rv, rd, rr, _) ->
            if height rr >= height rl then
              create (create l x d rl) rv rd rr
            else begin
              match rl with
                  Empty_map -> invalid_arg "Map.bal"
		| Node_map(rll, rlv, rld, rlr, _) ->
                    create (create l x d rll) rlv rld (create rlr rv rd rr)
            end
    end else
      Node_map(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

let rec add x data = function
    Empty_map ->
      Node_map(Empty_map, x, data, Empty_map, 1)
  | Node_map(l, v, d, r, h) as t ->
      let c = compare_key x v in
        if c = 0 then
          Node_map(l, x, data, r, h)
        else if c < 0 then
          bal (add x data l) v d r
        else
          bal l v d (add x data r)

let rec find x = function
    Empty_map ->
      raise Not_found
  | Node_map(l, v, d, r, _) ->
      let c = compare_key x v in
        if c = 0 then d
        else find x (if c < 0 then l else r)

let rec mem x = function
    Empty_map ->
      false
  | Node_map(l, v, d, r, _) ->
      let c = compare_key x v in
        c = 0 || mem x (if c < 0 then l else r)

let rec merge t1 t2 =
  match (t1, t2) with
      (Empty_map, t) -> t
    | (t, Empty_map) -> t
    | (Node_map(l1, v1, d1, r1, h1), Node_map(l2, v2, d2, r2, h2)) ->
        bal l1 v1 d1 (bal (merge r1 l2) v2 d2 r2)

let rec remove x = function
    Empty_map ->
      Empty_map
  | Node_map(l, v, d, r, h) as t ->
      let c = compare_key x v in
        if c = 0 then
          merge l r
        else if c < 0 then
          bal (remove x l) v d r
        else
          bal l v d (remove x r)

let rec iter f = function
    Empty_map -> ()
  | Node_map(l, v, d, r, _) ->
      iter f l; f v d; iter f r

let rec map f = function
    Empty_map               -> Empty_map
  | Node_map(l, v, d, r, h) -> Node_map(map f l, v, f d, map f r, h)

let rec mapi f = function
    Empty_map               -> Empty_map
  | Node_map(l, v, d, r, h) -> Node_map(mapi f l, v, f v d, mapi f r, h)

let rec fold f m accu =
  match m with
      Empty_map -> accu
    | Node_map(l, v, d, r, _) ->
        fold f l (f v d (fold f r accu))


(** --- Fct supplementaire ----------------------------------- **)

let size m = (let nb = ref 0 in iter (fun _ _ -> incr nb) m ; !nb) (* fold (fun _ _ n -> succ n) m 0 *)
;;

let singleton k v = Node_map(Empty_map, k, v, Empty_map, 1)
;;

let equal egal m1 m2 = 
  let rec equal_2 egal m1 m2 = try ignore(mapi (equal_3 egal m2) m1); true with Not_found -> false
  and equal_3 egal m key v = if not(egal v (find key m)) then raise Not_found
  in  (let l1 = size m1
       and l2 = size m2
       in if l1 <> l2 then false
	 else (equal_2 egal m1 m2))
;;


(** copier sur core_set en adaptant **)

let rec compare_map_aux l1 l2 =
  match (l1, l2) with
      ([], []) -> 0
    | ([], _)  -> -1
    | (_, []) -> 1

    | (Empty_map :: t1, Empty_map :: t2) -> compare_map_aux t1 t2

    | (Node_map(Empty_map, k1, v1, r1, _) :: t1, Node_map(Empty_map, k2, v2, r2, _) :: t2) ->
	let diffk = compare_key k1 k2 
	in (if diffk <> 0 then diffk
	    else (let c = compare_mgs_value v1 v2
		  in if c <> 0 then c else compare_map_aux (r1::t1) (r2::t2)))

    | (Node_map(l1, k1, v1, r1, _) :: t1, t2) ->
        compare_map_aux (l1 :: Node_map(Empty_map, k1, v1, r1, 0) :: t1) t2

    | (t1, Node_map(l2, k2, v2, r2, _) :: t2) ->
        compare_map_aux t1 (l2 :: Node_map(Empty_map, k2, v2, r2, 0) :: t2)
;;

let compare s1 s2 = compare_map_aux [s1] [s2]
;;

let hd h = 
  let ret = ref None
  in
    try
      iter (fun k v -> ret := Some (k, v) ; raise BreakLoop) h;
      failwith "hd"
    with BreakLoop -> (match !ret
		       with Some x -> x
			 | _       -> failwith "hd")



(* ******************************************************************************** *)
(* Pour jouer un peu ;-) Le tl est code en faisant un dispatch sur la fonction      *)
(* du fold. Il faut simplement modifier le fold (pour que le code soit plus elegant *)
(* afin de laisser la fonction "gelee")                                             *)
(* ******************************************************************************** *)

(*
let tl h = 
  let first = ref true
  in
    fold (fun k v s -> if !first then (first := false; s) else (add k v s)) h empty
*)


let rec my_fold f m accu =
  match m with
      Empty_map               -> accu
    | Node_map(l, v, d, r, _) -> my_fold f l (!f v d (my_fold f r accu))


let tl h = 
  let f     = ref (fun k v s -> empty) in
  let first = (fun k v s -> f:= add ; s)
  in
    f := first;
    my_fold f h empty
  
(* ******************************************************************************** *)
(* ******************************************************************************** *)


let last h = 
  let ret = ref None
  in
    iter (fun k v -> ret := Some (k, v)) h;
    (match !ret
     with
	 Some x -> x
       | None   -> failwith "last")
;;

