(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: type_rec.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Voir le fichier Common/type.ml
 * 
 *=========================================================================
 *)



open Global;;
open Valeur;;
open Type;;
open Dispatch_sig;;

__register_version "$RCSfile: type_rec.ml,v $" "$Revision: 1.2 $" "$Date: 2005-11-23 11:10:57 $";;




(* --- Record ---------------------------------------------------- *)

(*!*)(** 
       @module collections
       @name 'record
       @type any_bool
       Record type and predicate.
       
       A record is a collection of elements munished with a void
       topology (an element has no neighbors). The elements of a
       record are the values associated to the fields of the records.

       Record's are constructed types. The corresponding declaration
       begins with the keyword [record]. This implies that the
       corresponding predicate is denoted only by its quoted name
       ['record]. 
       
       @voir Record the chapter record in the documentation.
*)(*!*)

let record_desc = create_type
		    { name = "'record";
		      concrete = true;
		      leibniz = false;
		      predicate = wrap_pred(is_newton);
		      empty = VAL_undef "dummy for empty field in type descriptor 'record";
		      pospredicate = pospredicate_string;
		      make = make_dummy;
		      xxxify = xxxify_dummy "'rec";
		      derive_from = [newton_desc];
		      father_of = [];
		      uid = get_new_uid();
		    }
;;


(** !!! Attention !!!
 *** Ne pas oublier de mettre  jour les champs
 ***     _ make 
 ***     _ xxxify
 ***     _ empty
 *** du descripteur de type dans le fichier rec.ml
 ***)



(* === Gestion des specification de records =================================================== *)


(** --- record avec spec. de champs ------------------------------------------------- **)

(** On ne rajoute que les champs qui sont prsent dans le type de
 ** record, en partant de la valeur par dfaut du type. Mais on ne 
 ** vrifie pas que les valeurs donnes sont correctement types.
 **)

let rec extension_record desc defaultval a = 
  check_vplist desc.name (to_funct desc.pospredicate) a;
  VAL_rec(Array.fold_left add_elem defaultval a, desc)

and add_elem map =  function
  | VAL_seq(aa, _) -> if Core_rec.mem aa.(0) map then Core_rec.add aa.(0) aa.(1) map else map
  | _ -> failwith "Internal Error: extension: add_elem in extension_record"
;;



(* === Les records ====================================== *)

let create_new_record nom pred defaultval =
  try
    let t = Hash.find type_tbl nom
    in let err = "Error: type  "^nom^"  already defined."
    in prerr_endline err; prftype stderr t; raise (Erreur err)
  with Not_found -> 
    begin
      (if Hash.mem Environnement.var_tbl nom
       then prerr_endline ("Warning: type predicate  "^nom^"  hiddes a function with same name"));
      let nt = create_type { name = nom;
		 	     concrete = true;
			     leibniz = false;
			     predicate =  VAL_funct(function x ->  VAL_bool(pred x));
			     empty = VAL_rec(defaultval, record_desc);
			     pospredicate = pospredicate_string;
			     make = make_dummy;
			     xxxify = xxxify_dummy nom;
			     derive_from = [record_desc];
			     father_of = [];
			     uid = get_new_uid();
			   }
      in 
	begin 
	  nt.empty <- change_type nt nt.empty; 
	  nt.make <- extension_record nt defaultval;
	  ignore(make_xxxify nt);
	  ignore(Environnement.addfct nom nt.predicate);
	  nt
	end
    end
;;

let create_dummy_record id = create_type
  { name = id;
    concrete = true;
    leibniz = false;
    predicate = record_desc.predicate;
    empty = record_desc.empty;
    pospredicate = VAL_funct(function x -> VAL_bool(true));
    make = make_dummy;
    xxxify = xxxify_dummy id;
    derive_from = [record_desc];
    father_of = [];
    uid = get_new_uid();
  }
;;

(* --- Les CHAMPS dans un record (typage) --------------------------------- *)


exception MergeError of string;;


let fieldspec_to_string s = function
  | Present        -> s
  | Absent         -> "~"^s
  | WithType t     -> s^" : "^(to_string (Printer.uprsvalue t))
  | WithCstrt c    -> raise (Erreur "Type_rec: fieldspec_to_string: not yet implemented on constraint")
  | WithValue v    -> s^" = "^(to_string (Printer.uprsvalue v))
  | WithoutValue v -> s^" ~= "^(to_string (Printer.uprsvalue v))
;;

let merge_error s ft ft' = raise (MergeError (s^" has two incompatible field types: "
					      ^(fieldspec_to_string s ft)^" and "
					      ^(fieldspec_to_string s ft')))
;;

let rec merge_field_spec s ft1 ft2 = match ft1, ft2 with
  | Absent, Absent -> Absent

  | Present, Present -> ft2
  | Present, (WithType t) -> ft2
  | Present, (WithValue v) -> ft2
  | Present, (WithoutValue v) -> ft2

  | (WithType t), Present -> ft1
  | (WithType t), (WithType t') -> if equal_mgs_value t t' then ft1 else merge_error s ft1 ft2
  | (WithType t), (WithValue v) -> if to_bool(val2bool((to_funct t)v)) then ft2 else merge_error s ft1 ft2
  | (WithType t), (WithoutValue v) -> (** on cree une fct qui verifie le type et la non-valeur **)
      WithType (VAL_funct(function x -> if equal_mgs_value v x then mgsfalse else val2bool((to_funct t)x)))


  | (WithValue _), Present -> ft1
  | (WithValue v), (WithType t) ->  merge_field_spec s ft2 ft1
  | (WithValue v), (WithValue v') -> if equal_mgs_value v v' then ft1 else merge_error s ft1 ft2
  | (WithValue v), (WithoutValue v') -> if not(equal_mgs_value v v') then ft1 else merge_error s ft1 ft2


  | (WithoutValue v), (WithoutValue v') -> 
      WithType (VAL_funct(function x -> 
			   if (equal_mgs_value v x) or (equal_mgs_value v' x) then mgsfalse else mgstrue))

  | (WithoutValue _), _ -> merge_field_spec s ft2 ft1

  | _, _ ->  merge_error s ft1 ft2
;;




(** Test qu'un record r a un champ s correspondant  un field_type donn **)

type reponse_get_field = OK of mgs_value | Good | RecFailure
;;
let rec check_field s r ft = 
  let field = VAL_symbol (make_symbol_name s) in
    match get_field field r ft with
    | OK _ -> true
    | Good -> true
    | RecFailure -> false

and get_field s r ft = try get_field_aux s r ft with Not_found -> RecFailure

and  get_field_aux s r = function
  | Present -> OK(Core_rec.find s r)
  | Absent -> if not(Core_rec.mem s r) then Good else RecFailure
  | WithType t -> let v = Core_rec.find s r in if to_bool(val2bool((to_funct t)v)) then OK(v) else RecFailure
  | WithCstrt c -> raise (Erreur "Type_rec: get_field_aux: internal error: this case shouldn't happen")
  | WithValue v -> let v' = Core_rec.find s r in if equal_mgs_value v v' then OK(v) else RecFailure
  | WithoutValue v -> let v' = Core_rec.find s r in if not(equal_mgs_value v v') then OK(v) else RecFailure
;;


(** Test qu'un record a le type dfini par une liste fs de field_specification **)

let rec check_record fs = function
  | VAL_rec (r, _) -> List.for_all (function (id, spec) -> check_field id r spec) fs
  | _ -> false
;;



(** Table de tous les types records **)

let recordtype_tbl =  ((Hash.create 101) : ((string, field_specification list) Hash.t))
;;


(** cration de la liste des champs d'un type record,  partir de l'expression dfinissant le type **)

let rec get_record_spec_from_name id = 
  try
    Hash.find  recordtype_tbl id
  with Not_found ->
    let err = "Error: identifier "^id^" does not refer to a record type"
    in prerr_endline err; raise (Erreur err)


and field_spec_from_record_type = function
  | Fields l -> l
  | RecordType id -> get_record_spec_from_name id 


and merge_rec_spec l = function
  | [] -> l
  | (s, ftype)::l' ->
      try
	let ft' = List.assoc s l in
	  begin
	    try
	      let nft = merge_field_spec s ftype ft'
	      and ll = List.remove_assoc s l
	      in merge_rec_spec ((s, nft)::ll) l'
	    with MergeError msg ->
	      let err = "Error: field "^msg
	      in prerr_endline err; raise (Erreur err)
	  end
      with Not_found -> merge_rec_spec ((s, ftype)::l) l'

and normalise_rec_spec l = List.fold_left merge_rec_spec [] (List.map field_spec_from_record_type l)
;;


let default_field_value = VAL_undef "default field value"
and warning_default_field_value() = 
  let err1 = "Type_rec: warning_default_field_value: can't create a default value for this record specification."
  and err2 = "                                       If xxx is the name of the new record specification, the"
  and err3 = "                                       behavior of xxx(xxx:()) is unknown."
  in
    Printf.fprintf stderr "WARNING: %s\n         %s\n         %s\n\n\n" err1 err2 err3
;;

exception Found of mgs_value;;

let s2field s = VAL_symbol (make_symbol_name s)
;;

let get_all_dependances recspec = List.fold_left (fun acc -> (function RecordType id -> id::acc | _ -> acc )) [] recspec ;;

(** Construction de la valeur par dfaut d'un record.
 ** Cette valeur est celle retourne quand on fait ():R o R est le type d'une record.
 ** - la valeur par defaut d'un champ simplement prsent est une valeur undef.
 ** - il n'y a pas de problme si un champ est absent :-)
 ** - si le champ est prsent et avec une valeur, cette valeur est la valeur par dfaut.
 ** - la valeur par defaut d'un champ typ est la valeur par defaut du type.
 **   Cette valeur est calcule par la fonction  make_default_value_from_pred   laquelle
 **   on passe un predicat qui teste si les valeurs proposes conviennent. Les valeurs
 **   proposes sont prises dans une liste prdfinie pour les scalaires et on teste
 **   la valeurs empty de chaque type de collections. Si rien ne va, on met une valeur
 **   undef qui signale qu'on a pas su faire.
 ** - si un champ ne doit pas avoir une valeur donne, on utilise l'approche prcdente avec 
 **   un prdicat les valeur de mme type que la valeur interdite, et qui teste que la valeur 
 **   par defaut n'est pas la valeur interdite.
 **)
let rec default_value = function
  | [] -> Core_rec.empty
  | (s, Present)::l -> Core_rec.add (s2field s) default_field_value (default_value l)
  | (_, Absent)::l -> default_value l
  | (s, WithType pred)::l -> Core_rec.add (s2field s) (make_default_value_from_pred pred) (default_value l)
  | (s, WithCstrt c)::l -> raise (Erreur "Type_rec: default value: internal error: this case shouldn't happen")
  | (s, WithValue v)::l -> Core_rec.add (s2field s) v (default_value l)
  | (s, WithoutValue v)::l ->  
      let pred = VAL_funct(fun x -> VAL_bool(is_same_kind v x & not(equal_mgs_value x v)))
      in Core_rec.add (s2field s) (make_default_value_from_pred pred) (default_value l)

and make_default_value_from_pred  =
  function pred ->
    let predfct = to_funct pred in
    let check   = function v -> to_bool(val2bool(predfct(v))) in 
      begin (** on check les couleurs **)
	try
	  Hash.iter
	    (fun _ desc -> if check desc.empty then raise (Found desc.empty))
	    Type.type_tbl;
	  warning_default_field_value();
	  VAL_undef "Type_rec: make_default_value_from_pred: default value"
	with Found v -> v
      end
;;


let process name spec =
  let nspec = normalise_rec_spec spec in
  let rtype = create_new_record name (check_record nspec) (default_value nspec) in
    begin
      Hash.add recordtype_tbl name nspec;
      rtype
    end
;;



let compare_record (id1, _) (id2, _) = Pervasives.compare id1 id2
;;
let rec listing out =   
  let lst = List.sort compare_record  (Hash.fold  (fun k v l -> (k, v)::l) recordtype_tbl [])
  in List.iter (fun (k, v) -> Printf.fprintf out "record %s = { %a }\n" k recspec v) lst

and recspec out = function
  | [] -> ()
  | [(s, ft)] -> Printf.fprintf out "%s" (fieldspec_to_string s ft)
  | (s, ft)::l -> Printf.fprintf out "%s, %a" (fieldspec_to_string s ft) recspec l 
;;


let number_defined () = Hash.size recordtype_tbl
;;






(* ******************************************************************* *)
(* Fonctions pour la gestion du "and" entre les dclarations de types. *)
(* ******************************************************************* *)

let rec check_unknown_names_record_type_specification llhs = function
  | Fields fspecl -> List.for_all (check_unknown_names_field_specification llhs) fspecl
  | RecordType id -> (Hash.mem type_tbl id) or (Hash.mem type_tbl ("'"^id)) or (List.mem id llhs)

and check_unknown_names_field_specification llhs = function
  | (_, Present          ) -> true
  | (_, Absent           ) -> true
  | (_, WithType     (mv)) -> false
  | (_, WithCstrt    c   ) -> Type.check_unknown_names_cstrt_tree llhs c
  | (_, WithValue    (mv)) -> true
  | (_, WithoutValue (mv)) -> true
;;

let rec replaceDico_unknown_type_specification dico = function
  | Fields fspecl -> Fields (List.map (replaceDico_unknown_type_field_specification dico) fspecl)
  | RecordType id -> RecordType id

and replaceDico_unknown_type_field_specification dico = function
  | (a, WithCstrt    c   ) ->
      let pred = (VAL_funct((function v -> VAL_bool(build_constraint_pred_tree v (replaceDico_unknown_type_cstrt_tree dico c))))) in
      (a, WithType pred)
  | x -> x
;;



let update_new_rec_node dico n id recspec =
  let recspec = List.map (replaceDico_unknown_type_specification dico) recspec in
  let nrecspec = normalise_rec_spec recspec in
  let default = default_value nrecspec in
  let chk = check_record nrecspec in
    Hash.add recordtype_tbl id nrecspec;
    n.empty <- VAL_rec(default, n);
    n.make <- extension_record n default;
    n.predicate <- VAL_funct(function x ->  VAL_bool(chk x));
    ignore(make_xxxify n);
    ignore(Environnement.addfct n.name n.predicate);
    n
;;
