(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_ring.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-07-12 15:32:18 $
 * Version: $Revision: 1.4 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *          
 * Ici on developpe les instance de generic_topo qui correspondent 
 * aux collections monoidales.
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Topology;;

__register_version "$RCSfile: topo_ring.ml,v $" "$Revision: 1.4 $" "$Date: 2005-07-12 15:32:18 $";;



(* =================================================================
 * === Ring ====================================================
 * =================================================================
 * Voir Seq/topo_seq.ml
 * ================================================================-
 *)





exception FoundIntPos of int;;
exception FoundMgsValue of mgs_value;;
exception FoundMgsValue2 of mgs_value;;

let rec last = function
  | a::[] -> a
  | _::l -> last l
  | _ -> raise (Erreur "Internal error: last (file topo_ring.ml)\n");;


(** --- print d'un batch de Ring ----------------------- **)

let rec print_ring_batch chan valpos batch = 
  Printf.fprintf chan "-- batchRING : begin\n";
  List.iter (print_one_occurence chan valpos) batch;
  Printf.fprintf chan "-- batchRING : end\n";
  flush chan

and print_one_occurence chan valpos  (pt, pl, v) =
  Printf.fprintf chan "\t%s %a --> %a\n"
    (if pt.a_flat = FlatPath then "flat" else "noFlat")
    (fun o l -> List.iter (function (p) -> Printf.fprintf chan " (pos=%d, val=%a)" p prfvalue2 (valpos p)) l)
    pl
    prfvalue2 v;
  flush chan
;;



(** --- Trie d'un batch de Ring par position croissante ----------------------- **)

let rec compare_ring_occurence posval q1 q2 = cso posval (q1, q2)

and cso posval = 
  let err = "Error: internal error: compare_ring_occurence: bad batch\n"
  in (function (_, p1::_, _),  (_, p2::_, _) -> p1 - p2 
	| (b1, b2) -> 
	    Printf.fprintf stderr "%s" err;
	    print_one_occurence stderr posval b1; print_one_occurence stderr posval b2; 
	    raise (Erreur err))
;;


(* divise : *)
(* prend une liste d'entiers compose de deux sous-listes non-vides d'entiers *
 * chacune rangee dans l'ordre croissant, et telles que 
 * tous les entiers de la premier sont plus grands que ceux de la seconde *)
(* renvoit le split de ces deux sous listes *)
 
let rec divise = function 
  | i1 :: i2 :: r -> 
      if i1 < i2 
      then match divise (i2::r) with l1,l2 -> (i1::l1,l2)
      else [i1],(i2::r) 
  | _ -> raise (Erreur "Internal error: batch file for ring: imossible split") ;;
	


let rec split_all = function 
    [] -> []
  | (p,l,v)::r -> 
      if last l < List.hd l 
      then  
	match divise l with 
	    l1,l2 -> (p,l1,VAL_undef "empty") :: (p,l2,v) :: (split_all r)
	      (* on aurrait pu mettre v et undef au lieu de undef et v *)
      else (p,l,v) :: (split_all r )
;;
    

(** --- Construction d'un ring a partir d'un batch et d'un ring ----------------------- **)

(* c est un VALUE array *)

let rec merge_ring couleur c = function
  | [] -> c   (* batch vide: rien a faire *)
  | b -> Array.concat (mergering couleur c 0 b)

and mergering couleur c i = function 
  (** batch vide **)
  | [] -> let last_index = Array.length c 
          in if i > last_index then [] else [Array.sub c i (last_index - i)]

  (** traitement d'un chemin vide et cas non-flat: ca arrive avec x* **)
  | (rp, [], v)::b
      when (rp.a_flat = NoFlatPath) 
	or not(is_collection v)
	or not(Type.is_subtype (collection_type v) couleur) ->
      (Array.make 1 v)::(mergering couleur c i b)

  (** traitement d'un chemin vide et cas flat: ca arrive avec x* **)
  | (_, [], v)::b ->
      assert(is_collection v);
      assert(Type.is_subtype (collection_type v) couleur);
      (to_ring v)::(mergering couleur c i b)
      
  (** traitement de la valeur undef : a s'limine (ring est un leibnizien) **)
  | (_, p, VAL_undef _)::b  ->
      let i0 = List.hd p and i1 = last p in
	if i < i0 then (Array.sub c i (i0 - i))::(mergering couleur c (i1+1) b)
	else if i0 <= i then (mergering couleur c (i1+1) b) 
	else raise (Erreur "Internal error: batch file for ring: impossible case (c)")

  (** traitement du cas non-flat: on insere v comme 1 seul element **)
  | (rp, p, v)::b 
      when (rp.a_flat = NoFlatPath) 
	or not(is_collection v)
	or not(Type.is_subtype (collection_type v) couleur) ->
      let i0 = List.hd p and i1 = last p in
	if i < i0 then (Array.sub c i (i0 - i))::(Array.make 1 v)::(mergering couleur c (i1+1) b)
	else if i0 <= i then (Array.make 1 v)::(mergering couleur c (i1+1) b) 
	else raise (Erreur "Internal error: batch file for ring: impossible case (a)")

(* reconstruction pas plus difficile que sur les seq                  *
 * car le  cas ou le chemin a remlacer                                *
 * va de la fin au debut de l'anneau  a ete separe en deux sous-batch *)
  (** traitement du cas flat: v est une collection compatible et il faut l'inserer **)
(* pathtype : Terme.path_type : pas utilise *)
(* p : int list : liste des positions a remplacer *)
(* v : Valeur.mgs_value : valeur de remplacement (en general, un scalaire ou une sequence) *)

  |  (_, p, v)::b ->
       
       assert(is_collection v);
       assert(Type.is_subtype (collection_type v) couleur);
       
       let i0 = List.hd p and i1 = last p in

	
	   
	 (* i : position dans le ring initial que l'on considere *)
	 if i < i0 
	 then (Array.sub c i (i0 - i))::(to_ring v)::(mergering couleur c (i1+1) b)
	 else 
	   if i0 <= i 
	   then (to_ring v)::(mergering couleur c (i1+1) b)
	   else raise (Erreur "Internal error: batch file for ring: impossible case (b)")

	

;;





(** --- La topologie d'un ring  ------------------------------ **)

class topo_ring (c, couleur) mgsval =
object(self)
  inherit [int] generic_topo 
    string_of_int                     (** afficheur de position    **)
    (c, couleur)                      (** collection a transformer **)
    mgsval                            (** la valeur mgs complete sur laquelle on travaille **)
    (Global.crandom (Array.length c)) (** position de depart       **)

  val _lg = Array.length c
  val _used = Array.make (Array.length c) PosFree
  val mutable _batch = ([] : (rule_property * 'position path * mgs_value) list)

  method state p          = _used.(p)
  method update_state f p = _used.(p) <- f(_used.(p))

  method unvisit_all = for i = 0 to _lg-1 do (if self#visited(i) then self#unvisit(i)) done

			 (* p : position concerne *)
  method next_usable p =
    let delta = if !Global.match_random >= 2 then (Global.crandom _lg) else 0 in
      try 
	for i = 0 to _lg - 1 do (* parcours les positions suivant p (en bouclant) *)
	  assert begin if !trace_opt then 
	    (Printf.printf "---- topoRING: next_usable: try %d\n" ((p+i) mod _lg); flush stdout);
	    true
	  end;
	  let pp = (p+i+delta) mod _lg 
	  in if not(self#used pp) & not(self#visited pp) then raise (FoundIntPos pp)
	done;
	raise NoMoreUsable
      with FoundIntPos p -> 
	assert begin if !trace_opt then 
	  (Printf.printf "---- topoRING: next_usable: found %d\n" p); 
	  true 
	end;
	p

  method add_match pt p v = _batch <- (pt, p, v)::_batch

  method get_value_from_pos p = c.(p)
  method mgs_value_from_pos p = VAL_int(p)

(* ici difference par rapport a seq *) 
  method some_neighbors p = function
    | AllDirections ->
	if p > _lg - 1 then (0,[||]) (* n'arrive pas *)
	else let next_p = (p+1) mod _lg 
	in 
	  if self#used(next_p) 
	  then (0, [||])
	  else (1, [| next_p |])
	
	

    | SomeDirections(d) -> (let err = "Error: cannot use directions in ring\n"
				      ^(Array.fold_left (fun s x -> s^"\t"^(to_string (prsvalue x))^"\n") "" d)
			    in 
			      prerr_endline err; 
			      raise (Erreur err))

  method direction_from_to p1 p2 = VAL_int(p2 - p1) (** Ou VAL_undef, je ne sais pas ce qui est le mieux **)

  method build_result = 
    let sorted_batch = List.sort (compare_ring_occurence (self#get_value_from_pos)) (split_all _batch)
    in 
      VAL_ring(merge_ring couleur c sorted_batch, couleur)

end;;

