(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: kernel_scalar.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2008-06-03 12:59:19 $
 * Version: $Revision: 1.20 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *   voir aussi Common/kernel_collection.ml
 *   qui contient les fcts gnriques sur les collections et les monoidaux.
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Type;;
open Dispatch_sig;;
open Dispatch;;
open Environnement;;
open Scalar;;
open Big_int;;

open Auxfct;;
open Kombinator;;


__register_version "$RCSfile: kernel_scalar.ml,v $" "$Revision: 1.20 $" "$Date: 2008-06-03 12:59:19 $";;

print_string "(* loading kernel functions on SCALAR into user-space ... ";
flush Pervasives.stdout;;


(*!*)

(*** 
  @module scalar      
  Functions acting on {i scalar values}: int, string, float, undef, splx, funct, posgbf, gmap, bint etc. 

  The values in MGS are typed. The universe of types is splitted in
  two: {i scalar} or {i atomic} values that cannot be decomposed and
  {i collections} that are aggregate of simpler values. Scalar types
  includes
  {ul
  {- {i undef} this type contains only one value representing an
  undefinition in various circumstances
  }
  {- {i bool} the type of the two boolean values [true] and [false] 
  }
  {- {i int} the type of signed integers}
  {- {i bint} the type of signed arbitrary-precision integers}
  {- {i float} the type of floating point numbers}
  {- {i string} the type of strings of characters}
  {- {i symbol} the type of symbols, Lisp-like constants definied by
  a back quoted identifiant (e.g. `Constant)
  }
  {- {i funct} the type of functions}
  {- {i posgbf} the type of position in a GBF}
  {- {i splx} the type of simplices}
  {- {i ncell} n-dimensionnal cells that are the type of position in a qmf}
  {- {i gmap} the underlying space of a qmf}
  }

  @voir Function For functionnal values refer also to the chapter [Functional].
  @voir Chain    For functions related to simplex, cf. the chapter [Chain].
  @voir GBF      For functions related to posgbf, cf. the chapter [GBF].
  @voir Math     For various mathematical constants and arithmetics operators, cf. the chapter [Math].
*)







(* === SCALAR  ================================================= *) 



(* --- Undef ------------------------------------------------------- *)


(** 
  @module scalar
  @name string2undef
  @type string_undef
  Returns an undef value with its string argument as a comment

  @example
  ['string2undef("toto")] return [<undef: toto>]
*)
addenv_pure "'string2undef" (fun_1 __string2undef);;


(** 
  @module scalar
  @name undef2string
  @type undef_string
  Returns the comment of an undef value as a string

  @example
  ['undef2string(<undef: toto>)] return ["toto"]
*)
addenv_pure "'undef2string" (fun_1 __undef2string);;


(* --- Bool ------------------------------------------------------- *)

(**
  @module scalar  
  @name 'val2bool
  @type any_bool
  Convert a value to a boolean.

  Any MGS value can be converted into a boolean. This
  conversion occurs {i implicitly} for instance in the condition of
  a [if ... then ... else ... fi] expression. This conversion can be
  done {i explicitly} using the function [val2bool]. The conversion
  rules are the following:
  {ul
  {- a [undef] is converted into [false].}
  {- a [funct] is converted into [false].}
  {- a [symbol] is converted into [false].}
  {- an [int] is converted into [true] if it is non-zero.
  }
  {- an [bint] is converted into [true] if it is non-zero.
  }
  {- a [float] is converted into [true] if it is different from [0.0].
  }
  {- a [string] is converted into [true] if its length is different
  from 0.
  }
  {- a [posgbf] is converted into [true] unless it represents the
  {i identity element} of the group law}
  {- a [splx] is converted into [true] if its dimension is different
  from [0].}
  {- a [ncell] is converted into [true] if its dimension is different
  from [0].}
  {- a [gmap] XXXXXXX  is converted into [????].}
  {- a collection is converted into [true] only if it is
  non-empty. Alternatively, a en empty collection corresponds to a
  [false] value. 
  }
  }
*)
(addenv_pure "'val2bool"   (fun_1 __val2bool));;



(**
  @module scalar    
  @name 'not any_bool
  @type any_bool
  returns the negation of the boolean value associated to its
  argument. For example [not(3) == false]. 
*)
(addenv_pure "'not"        (fun_1 not_val));;


(**
  @module scalar  
  @name and_and
  @type any_any_bool

  returns the logical conjunction of the two arguments.

  The arguments can be any MGS value. See the function ['val2bool] for the
  boolean interpretation of a value. [&&] can be aslo written [&].

  The evaluation is lazy: 
  {[false && (?("Hello"); true)]}
  does not print [Hello] to the terminal.

  Warning: [a && b] is not an ordinary function call but a special form that is expansed 
  during the parsing to [if a then b else false fi]. The quoted name does not exist.
*)

(**
  @module scalar  
  @name ||
  @type any_any_bool

  returns the logical disjunction of the two arguments.

  The arguments can be any MGS value. See the function ['val2bool] for the
  boolean interpretation of a value.

  The evaluation is lazy: 
  {[true || (?("Hello"); true)]}
  does not print [Hello] to the terminal.

  Warning [a || b] is not an ordinary function call but a special form that is expansed 
  during the parsing to [if a then true else b fi]. The quoted name does not exist.
*)


(* --- Int ------------------------------------------------------- *)

(** 
  @module math 
  @type int
  The greatest representable integer.  
*)
addenv_pure "'max_int"       (VAL_int   Pervasives.max_int);;


(**
  @module math 
  @type int
  The smallest representable integer. 
*)
addenv_pure "'min_int"       (VAL_int   Pervasives.min_int);;


(* --- Bint ------------------------------------------------------- *)

(** 
  @module math 
  @type bint
  The big integer 0.
*)
addenv_pure "'zero_bint"       (VAL_bint Big_int.zero_big_int);;


(** 
  @module math 
  @type bint
  The big integer 1.
*)
addenv_pure "'unit_bint"       (VAL_bint Big_int.unit_big_int);;



(* --- Symbol ---------------------------------------------------- *)




(* --- String ---------------------------------------------------- *)

(**
  @module scalar
  @name 'string_length
  @type string_int

  Returns the length of it's string argument
*)
(addenv_pure "'string_length"   (fun_1 __stringlength));;  



(**
  @module scalar
  @name 'string_get
  @type string_int_string

  [string_get(s, n)] returns character number [n] in string [s]. The
  first character is character number 0. The last character is
  character number [string_length(s) - 1]. 
  Returns [<undef>] if [n] is outside the range 0 to 
  [string_length(s) - 1].
*)
(addenv_pure "'string_get"   (fun_2 __stringget));;  



(**
  @module scalar
  @name 'string_make
  @type int_string_string

  [string_make(n, c)] returns a fresh string of length [n],
  filled with the first character of string [c]. Returns [<undef>] if
  [n] is negative or too big (currently, too big means greater than
  16777211 characters). 
*)
(addenv_impure "'string_make"   (fun_2 __stringmake));;  



(**
  @module scalar
  @name 'string_sub
  @type string_int_int_string

  [string_sub(s, start, len)] returns a fresh string of length [len],
  containing the characters number [start] to [start + len - 1]
  of string [s].
  Returns [<undef>] if [start] and [len] do not
  designate a valid substring of [s]; that is, if [start < 0],
  or [len < 0], or [start + len > string_length s].
*)
(addenv_pure "'string_sub"   (fun_3 __stringsub));;  




(**
  @module scalar
  @name 'string_index
  @type string_string_int

  [string_index(s, c)] returns the position of the leftmost
  occurrence in string [s] of the first character of string [c].
  Returns [<undef>] if [c] does not occur in [s].
*)
(addenv_pure "'string_index"   (fun_2 __stringindex));;  



(**
  @module scalar
  @name 'string_rindex
  @type string_string_int

  [string_rindex(s, c)] returns the position of the rightmost
  occurrence in string [s] of the first character of string [c].
  Returns [<undef>] if [c] does not occur in [s].
*)
(addenv_pure "'string_rindex"   (fun_2 __stringrindex));;  



(**
  @module scalar
  @name 'string_index_from
  @type string_int_string_int

  Same as [string_index], but start
  searching at the character position given as second argument.
  [string_index(s, c)] is equivalent to [string_index_from(s, 0, c)].
*)
(addenv_pure "'string_index_from"   (fun_3 __stringindexfrom));;  




(**
  @module scalar
  @name 'string_rindex_from
  @type string_int_string_int

  Same as [string_rindex], but start
  searching at the character position given as second argument.
  [string_rindex(s, c)] is equivalent to
  [string_rindex_from(s, string_length(s) - 1, c)].
*)
(addenv_pure "'string_rindex_from"   (fun_3 __stringrindexfrom));;  





(**
  @module scalar
  @name 'string_contains
  @type string_string_bool

  [string_contains(s, c)] tests if the first character of [c]
  appears in the string [s]. 
*)
(addenv_pure "'string_contains"   (fun_2 __stringcontains));;  



(**
  @module scalar
  @name 'string_contains_from
  @type string_int_string_bool

  [string_contains_from(s, start, c)] tests if the first character of [c]
  appears in the substring of [s] starting from [start] to the end
  of [s].
   Returns [<undef>] if [start] is not a valid index of [s].
*)
(addenv_pure "'string_contains_from"   (fun_3 __stringcontainsfrom));;  


(**
  @module scalar
  @name 'string_rcontains_from
  @type string_int_string_bool

  [string_rcontains_from(s, start, c)] tests if the first character of [c]
  appears in the substring of [s] starting from the beginning
  of [s] to index [stop]. Returns [<undef>] if [start] is not a valid
  index of [s]. 
*)
(addenv_pure "'string_rcontains_from"   (fun_3 __stringrcontainsfrom));;  

(**
  @module scalar
  @name 'gensym
  @type string_symbol
  @type symbol_symbol

  [gensym(s)] generates a symbol from the string or the symbol [s].
  When [s] is a string, unless the first character of [s] is a back quote,
  [gensym(s)] adds a back quote at the beginning of [s]:

  [gensym("foo")]

  returns

  [`foo1]

  and 

  [gensym("`foo")]

  returns

  [`foo2]

  Obviously, 

  [gensym(`foo)]

  returns

  [`foo3]

  The uniqueness is given by an integer that is concatenated at the end of
  the argument and incremented. [gensym] dosen't check if this integer is
  incremented out of range.
*)
(addenv_impure "'gensym"   (fun_1 __gensym));;  

(**
  @module scalar
  @name 'split
  @type symbol_symbol

  Split a symbol using underscore as separator. [split(`Split_This_Symbol)] returns the sequence [\[`Split,`This,`Symbol\]].

*)
(addenv_pure "'split"   (fun_1 __split));;  







(* === FUNCTIONAL  =============================================== *) 


(*** 
  @module functional   
  Functional and Transformational programing: lambda and transformations, etc. 
*)



(** 
  @module functional 
  @type any_any
  The [I] combinator: [I == \x.x].
*)
(addenv_pure "'I"    mgs_combI);;


(** 
  @module functional 
  @type any_funct
  The [K] combinator: [K(a) == \x.a].
*)
(addenv_pure "'K"    mgs_combK);;


(** 
  @module functional 
  @type funct_funct_funct
  The [B] combinator: [B(f, a) == \x.f(a(x))].
*)
(addenv_pure "'B"    mgs_combB);;


(** 
  @module functional 
  @type funct_any_funct
  The [C] combinator: [C(f, a) == \x.(f(x))(a)].
*)
(addenv_pure "'C"    mgs_combC);;


(** 
  @module functional 
  @type funct_funct_funct
  The [S] combinator: [S(f, a) == \x.(f(x))(a(x))].
*)
(addenv_pure "'S"    mgs_combS);;


(** 
  @module functional 
  @type funct_funct_funct_funct
  The [S4] combinator: [S4(f, a, b) == \x.f(a(x), b(x))].
*)
(addenv_pure "'S4"   mgs_combS4);;


(** 
  @module functional 
  @type funct_funct_funct_funct
  The [IF] combinator: [IF(c, t, f) == \x.if c(x) then t(x) else f(x) fi].
  Beware that this combinator is strict, that is: [c], [t] and [f] are evaluated 
  (but [f(x)] is not evaluated if [c(x)] is true and [t(x)] is not evaluated if [c(x)] is false).
*)
(addenv_pure "'IF"   mgs_combIF);;


(** 
  @module functional 
  @type funct_funct_funct
  The [S4_S] combinator: [S4_S == S4(S)]. In other words:
  [S4_S(a, b, x, y) = S(a(x), b(x), y) = (a(x, y))(b(x, y))].
*)
(addenv_pure "'S4_S"  mgs_combS4_S);;


(** 
  @module functional 
  @type funct_funct
  The [B_K] combinator: [B_K == B(K)]. In other words: [B_K(f, x, y) == f(x)].
*)
(addenv_pure "'B_K"  mgs_combB_K);;


(**
  @module functional 
  @type any
  @name 'fixpoint

  The ['fixpoint] keyword specifies fixed-point iteration in function application.
  ['fixpoint] is a predefined identifier used as an optional argument to indicate 
  that the function applied must be iterated until a fixpoint is reached. This can 
  be used with several variations :
  {ul
  {-  [f\['fixpoint\](x0)] compute the fixed point of [f] starting from [x0], that is, it computes 
  [f(x0), f(f(x0)), ..., f(...f(x0)...), ...] until two consecutive elements in this 
  series become equal (and this last value is returned).
  }
  {- [f\[*\](x0)] is an abbreviation for [f\['fixpoint\](x0)]. This abbreviation can be used only
  if there is no other optionnal argument.
  }
  {- [f\['fixpoint=eq\](x0)] computes the fixed point of [f] starting from [x0] but [eq] is
  used as an equality predicate. [eq] must be a function that takes two arguments. These two 
  arguments are consecutive values in the previous series and they are provided to [eq] in 
  the according order.
  }
  }
  If the argument of ['fixpoint] does not evaluates to a function, an error is raised.
  Recall that a transformation is a function and then ['fixpoint] specifications apply 
  equally to transformation applications.

  The index of the current iteration is available within the body of the iterated function 
  as the value of the global variable ['iteration].

  <i>Optionnal wrapping functions</i> 

  The fixed point iterations of [f] can be wrapped by the optionnal wrapping functions ['prelude], 
  ['interlude] and ['postlude]. Such functions are unary functions and:
  {ul
  {- ['prelude] is applied prior any iteration and its results constitute the starting point 
  for the iteration.
  }
  {- ['interlude] is applied between two iterations. If the result of the [n]th iteration is [xn] 
  then the [(n+1)] iteration is computed as [f(interlude(xn))].
  }
  {- ['postlude] is applied finally to get the final result of the fixed point computation.}
  }
  When one (or several) of these function is not explicitly specified as optionnal arguments, 
  then the fixed-point evaluation proceed as this functions is the identity [\x.x].

  Usually these functions simply return the argument and are used only to produce a side-effect, 
  for instance to trace a function iteration or to cumulate some output in a file. However,
  they can be used to return another value. This is not a recommended practice but see the last 
  example in this section where this feature is used to avoid infinite loops. To summarize the use of 
  the optionnal wrapping function, expression
  {[    f['fixpoint=eq, 'prelude=pre, 'interlude=inter, 'postlude=pos](x)  ]}
  computes the series
  {[    x0 = pre(x)       ]}
  {[    x1 = f(x0)        ]}
  {[    x2 = f(inter(x1)) ]}
  {[    ...               ]}
  {[    xp = ...          ]}     
  {[    xn = f(inter(xp)) ]}     
  where [p == n-1]. The last element [xn] computed in this series is such that [eq(xp, xn)] holds.
  And the returned results is [post(xn)]. Note that if the pre/inter/post-lude functions are not 
  provided, the computation is exactly the fixed point computation described above.
  {[  ]}

  @example

  With the definition
  {[fun g(x) = x/2.0;; ]}
  the expression
  {[0.0 == g['fixpoint](2.0);;]}
  evaluates to [true]. Beware that the default comparaison function is the MGS ['equal] 
  predicate which is not necessarily adapted for numerical convergence. The predicate used 
  to test that the fixpoint is reached can be explicitly provided :
  {[ g['fixpoint=\x,y.abs(x-y) < 0.07](2.0);;]}
  returns [0.062500]. Indeed, [g(0.062500)] = [0.031250] and [abs(0.062500 - 0.031250)] = [0.031250]
  which is less than [0.07].

  The successive values during the iteration of [g] can be traced using the ['interlude] 
  wrapping function:
  {[g['fixpoint=(\x,y.abs(x-y) < 0.07), 'interlude=\x.(?(x); x)](2.0);; ]}
  write the following output:
  {[    1.000000 ]}
  {[    0.500000 ]}
  {[    0.250000 ]}
  {[    0.125000 ]}
  before returning the value [0.062500].
  {[   ]}
  @example

  Here is an example involving a collection.
  {[trans P = { x => x, (0-x) };; ]}
  {[P[*]((0, 1, -2, 3, 4, -5, set:()));;]}
  returns
  {[(-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5):'set]}

  The predicate used to test that a fixed-point is reached is arbitrary:
  {[trans T = { x => x, x };; ]}
  {[T\['fixpoint=\x,y.size(y)>20\]((0,1));;]}
  returns
  {[(0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1):'seq]}

  @example

  Here is another example involving the use of the optionnal wrapping functions : 
  {[fun pp(message, value)= (?(message+": iteration="+'iteration+": "+value); value);;  ]}
  {[fun k(x) = x+1;;]}
  {[k['prelude = pp("pre: "), postlude = pp("post: "), 'interlude=pp("inter: "), 'fixpoint=(\x,y. y >= 237)](234);;]}
  write the following output:
  {[    pre: : iteration=0: 234]}
  {[    inter: : iteration=1: 235]}
  {[    inter: : iteration=2: 236]}
  {[    post: : iteration=3: 237]}
  before returning the value [237]. Note the partial application of function [pp] to give 
  a value to the optional wrapping function ['prelude], ['interlude] and ['postlude].
  {[   ]}

  @example

  The index of the iteration, the fixpoint predicate and the postlude function can be used together
  to stop a fixed-point computation when a 
  maximal number of steps has been reached (to avoid infinite loops):
  {[fun m(x) = x+1;; ]}
  {[m['fixpoint = (\x, y. if 'iteration>100 then true else x == y fi),]}
  {[  'postlude = \x. if 'iteration>100 then <undef: max number of iterations in fixed point reached>]}
  {[                  else x fi](3)]}
  returns 
  {[<undef:  max number of iterations in fixed point reached>]}
  
  
  @voir 'iter 'iter
  @voir 'fixrule 'fixrule
*)






(* ==================================================================================================== *)

(**
  @module functional 
  @type symbol
  @name 'strategy

  The ['strategy] keyword allows to choose a rule application strategy different
  to the default MGS {i maximal parallel} strategy. For example, one can use a
  asynchronous gillespie-like stochastic rule application strategy by using the
  [`gillespie] strategy. The different strategies are:
  {ul
  {- [`default]: maximal parallel strategy (synchronous) with priority of the
  first rules over the last ones,
  }
  {- [`asynchronous]: asynchronous strategy with priority of the first rules
  over the last ones,
  }
  {- [`singleStochastic]: maximal parallel strategy (synchronous) with a randomly
  choosen priority between rules,
  }
  {- [`multiStochastic]: maximal parallel strategy (synchronous) without
  priority between rules,
  }
  {- [`stochastic]: true stochastic (asynchronous) with the explicit probability
  given on each rule,
  }
  {- [`gillespie]: a asynchronous gillespie-like stochastic strategy (see below).
  }
  }
  The application of a transformation [T] with a strategy [`s] on a topological
  collection [c] is written

  [ T[ 'strategy = `s ](c) ;; ]

  or

  [ T[ strategy = `s ](c) ;; ]

  The following details the different strategies:

  {[['strategy = `default]]}

  This strategy consits in a maximal parallel application of the rules with priority
  to the first rules. Applying a transformation [ T = trans { r1; r2; ... } ] is
  applying [r1] until no sub-collection matches with the pattern of [r1], applying
  [r2] until no sub-collection matches with the pattern of [r2], and so on.

  A maximal parallel application strategy corresponds to a {i synchronous} simulation
  mode. Several instancies of the pattern are matched and rewritten at the same time.
  Of course these instancies can't overlap.

  As an example, let [t] and [l] be respectively the transformation
  
  [ trans t = { x => x + 1 ; x => x + 2 } ;; ]

  and the topological collection

  [ l := (1,2,3,4,5,6,seq:()) ;; ]

  The [`default] application of [t] on [l] produces the sequence [ (2,3,4,5,6,7,seq:()) ].
  Indeed, the first rule (with a greater priority) is applied until no element
  matches. Therefore, the second rule is never applied because the first one has
  consumed all the elements of [l].

  {[['strategy = `asynchronous]]}

  The priority is given to the first rules, and only one rule is applied one time, if such
  a rule exists. In other word, only one path is rewritten if it is possible.

  {[['strategy = `singleStochastic]]}

  This strategy is same as [`default] but the priority between rules is choosen randomly.
  For the same example as previously, the [`singleStochastic] strategy can produces two
  results [ (2,3,4,5,6,7,seq:()) ] and [ (3,4,5,6,7,8,seq:()) ] with a probability of 0.5.

  {[['strategy = `multiStochastic]]}

  This strategy is same as [`default] but there is no priority between rules.

  {[['strategy = `stochastic]]}

  This strategy is a true asynchronous stochastic strategy. One can specify the probabiliy of each rule
  using the following [P= ] construction. The probability given as argument of [P] can be an integer 
  or a float value or even  a function (that will receive the collection,  onto which the transformation
  is appliedsformation as argument and that is supposed to return either an integer or a float value:

  [trans T = {]

  [	 x ={ P = (\x.(0.25)) }=> x+1;]

  [      x ={ P = 0.25        }=> x+10;]

  [      x ={ P = 0.5         }=> x+100]

  [ }]

  {[['strategy = `gillespie]]}

  The [`gillespie] rule application strategy allows the use of many options in
  the rules specification, according to Gillespie's original {i stochastic
  simulation algorithm}. Indeed, Gillespie's standart algorithm requires to
  compute the smallest {i tau} of a set of rules {i R} to choose which rule {i
  Ri} should be selected and the amount of time resulting from the selection of
  that rule. For example, let's look at the set of coupled, autocatalytic
  reactions due to {i Lotka} ({i Volterra} investigated later these equations to
  mathematically model a prey-predator ecosystem). The rules are stated as:

  [X  + Y1 -> X + 2 Y1] with {i C = 0.001}

  [Y1 + Y2 -> 2 Y2] with {i C = 0.01}

  [Y2      -> Z] with {i C = 10}

  The first rule states how a certain prey species [Y1] reproduces by feeding
  on a certain foodstuff [X]; the second rule expresses how a certain species
  of predator [Y2] reproduces by feeding of a species [Y1]; finally, the last
  rules expresses the disparition of species [Y2] by natural causes.

  For each rule, following Gillespie's words, the {i stochastic reaction
  constant} [C] corresponding to the average probability that a particular
  combination of reactant molecules will react accordingly in the next
  infinitesimal time interval {i dt}, is given. Its expression in MGS is
  straightforward (species are represented by symbols, [C] constants are given
  as rule options using the [C] constructions):

    [trans lotka_volterra = {]

    [  `X, `Y1  ={ C = 0.001 }=> #2 `Y1, `X;]

    [  `Y1, `Y2 ={ C = 0.01  }=> #2 `Y2;]

    [  `Y2      ={ C = 10    }=> `Z]

    [}]

  The application of the transformation to an initial condition requires the use
  of the [`gillespie] option:

    [lotka_volterra['fixpoint=(\x1,x2.('tau == 5.0)), 'strategy = `gillespie](bagify((#1000 `X, #100 `Y1, #100 `Y2)));;]

  Each rule applications increments the system variable ['tau] by the computed
  amount of time. During the transformation application, one is able to use the
  value of ['tau] to perform any kind of computation. The application of the
  example given above stops once ['tau] as reached a simulation time of
  [5.0]. The use of ['iter = x], ['fixpoint] enables a fine control on the
  number of transformatio applications.

  The [C] constants given in the MGS program above is used by the language to
  compute the new value of [tau] after the rule applications. It follows this
  scheme:
  {ul
    {- For every rule, according to the left-hand side of each rule, the [H]
    function is computed (for example, for the first rule, [H = #`X * #`Y1]
    where [#`X] represents the number of [`X] in the collection).
    }
    {- For every rule, the value [A = H * C] is computed.
    }
    {- For every rule, the value [tau = (1/A) * ln(1 / random())] is computed.
    }
  }
  The rule which has the smallest [tau] is selected and applied {i only
  once}. ['tau] is incremented by the [tau] associated with the selected rule.

  One that would like to explicitly compute the [A] value (for complex reactions involving
  not only reacting molecules but also enzymes, inhibitors, facilitators, ...) may use the
  following construction:

    [trans lotka_volterra = {]

    [  `X, `Y1  ={ A = \coll.(coll#`X * coll#`Y1 * 0.001) }=> #2 `Y1, `X;]

    [  ... ]

    [}]

*)

(* ==================================================================================================== *)

(**
  @module functional 
  @type int
  @name 'iter

  The ['iter] keyword specifies bounded iteration in function application.
  ['iter] is a predefined identifier used as an optional argument to indicate 
  the iterated application of a function.
  {[f['iter=nexp](x0)]}
  where [nexp] evaluates to the integer [n], computes the [n]-fold application 
  of [f] to [x0], that is:
  {[f(...f(0)...)]}
  where [f] is applied [n] times. By convention, [f['iter=0](x0)] returns [x0].
  If the argument of ['iter] is strictly negative, then [<undef>] is returned.
  If the argument of ['iter] does not evaluates to an integer, an error is raised.

  The expression [f[nexp](x0)] can be used to abbreviate [f['iter=nexp](x0)] if there is
  no other option specification.

  The index of the current iteration is available within the body of the iterated function 
  as the value of the global variable ['iteration].

  The optionnal wrapping functions ['prelude], ['interlude] and ['postlude] can be provided: 
  cf. ['fixpoint] for an explanation of this feature.  Note that [f['iter=0, ...](...)] does not
  apply any wrapping function and  [f['iter=1, ...](...)] apply only the ['prelude] and ['postlude] 
  functions (if they are present) since there is no iteration. In any case, for [n>=0],
  {[f['iter=n+1](...)  == f(f['iter=n](...)) ]}
  

  @example 

  {[fun f(x) = x+1;; f['iter=3](0)]}
  evaluates to [3].
  With the definition
  {[fun g(x) = x/2.0;; ]}
  the expression
  {[1.0 == g['iter=5](2*2*2*2*2);;]}
  evaluates to [true].

  @example

  Here is an example involving the use of the ['iteration] variable:
  {[fun h(x) = (?("iteration: "+'iteration); x+1);;]}
  {[h['iter=3](0)]}
  write the following output:
  {[    iteration: 1]}
  {[    iteration: 2]}
  {[    iteration: 3]}
  before returning the value [3].

  @example

  Finally, here is an example involving the use of the optionnal wrapping functions : 
  {[fun pp(message, value)= (?(message+": iteration="+'iteration+": "+value); value);;  ]}
  {[fun k(x) = x+1;;]}
  {[k['prelude = pp("pre: "), postlude = pp("post: "), 'interlude=pp("inter: "), 'iter=3](11);;]}
  write the following output:
  {[    pre: : iteration=0: 11]}
  {[    inter: : iteration=1: 12]}
  {[    inter: : iteration=2: 13]}
  {[    post: : iteration=3: 14]}
  before returning the value [14].




  @voir 'fixpoint  'fixpoint
  @voir 'iteration 'iteration
  @voir 'prelude   'prelude
  @voir 'interlude 'interlude
  @voir 'postlude  'postlude
*)

(**
  @module functional 
  @type funct
  @name 'prelude
  
  Optionnal wrapper function for iterated application.

  @voir 'fixpoint 'fixpoint
  @voir 'iter 'iter
*)

(**
  @module functional 
  @type funct
  @name 'interlude
  
  Optionnal wrapper function for iterated application.

  @voir 'fixpoint 'fixpoint
  @voir 'iter 'iter
*)

(**
  @module functional 
  @type funct
  @name 'postlude
  
  Optionnal wrapper function for iterated application.

  @voir 'fixpoint 'fixpoint
*)


(**
  @module functional 
  @type any
  @name 'fixrule
  
  ['fixrule] is equivalent to ['fixpoint].
  Reserved for future use.

  @voir 'fixpoint 'fixpoint
  @voir 'iter 'iter
*)



(* ======================================================= *)






(* === MATH ==================================================== *) 


(*** 
  @module math         
  Aritmetics and Mathematical operators. 

  @voir Scalar.'int int
  @voir Scalar.'bint bint
  @voir Scalar.'float float
  @voir Scalar the [Scalar] chapter in the documentation
*)



(* --- Arithmetique -------------------------------------------- *)

(** 
  @module math

  @type posgbf_posgbf_posgbf
  @type rec_rec_rec
  @type string_any_string
  @type any_string_string
  @type float_int_float
  @type int_float_float
  @type float_float_float 
  @type int_int_int
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float

  Numerical sum, record merging and string concatenation.

  The operator [+] is highly overloaded:
  {ul
  {- [int -> int -> int] is the numerical addition of integers.
  }
  {- [bint -> bint -> bint] is the numerical addition of big integers.
  }
  {- [float -> float -> float] is the numerical addition of floats
  (IEEE 754 compliant).
  }
  {- [int -> bint -> bint] and }
  {- [bint -> int -> bint] computes the numerical addition of big integers
  after the conversion of the integer argument into a big interger.
  }
  {- [int -> float -> float] and }
  {- [float -> int -> float] computes the numerical addition of floats
  after the conversion of the integer argument into a float.
  }
  {- [bint -> float -> float] and }
  {- [float -> bint -> float] computes the numerical addition of floats
  after the conversion of the big integer argument into a float.
  }
  {- [string -> string -> string] is the concatenation of
  strings. Similarly, }
  {- [any -> string -> string] and}
  {- [string -> any -> string] computes the concatenation of the
  string and of the string representation of the other argument.  For
  instance,
  {[ "pi/2 = " + ('PI/2.0)]}
  returns the string ["pi/2 = 1.570796"]. Another example: the string
  corresponding to the ascii representation written on the standard
  output when a value [v] is displayed, can be simply computed as
  [""+v].
  }
  {- [posgbf -> posgbf -> posgbf] 
  computes the result of the sum of the elements of the abelian group
  underlying the GBF. The result is given as an element of the
  isomorphic Smith normal form of the group. For example, let a GBF
  [G] be defined by 
  {[ gbf G = <a; 33*a = 0> ]}
  then 
  {[ (13*|a> + 20*|a>) ]}
  returns
  {[G:|0>]}
  }
  {- [rec -> rec -> rec]
  represents the asymetric merge of two records. That is, the result
  combines the fields of the two record arguments. If a field [f]
  appears in the two arguments, the value of [f] in the result is the
  value in the right argument. For example:
  {[ {a=1, b=true} + {b=false, c="xyz"}  ]}
  evaluates to
  {[ {a=1, b=false, c="xyz"} ]}
  }
  }
*)
(addenv_pure "'+"       (fun_2 __add));;


(**
  @type posgbf_posgbf_posgbf
  @type rec_rec_rec
  @type string_any_string
  @type any_string_string
  @type float_int_float
  @type int_float_float
  @type float_float_float 
  @type int_int_int
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float
  @alias math.'add 
  Prefix form of operator [+]
  @voir '+ +
*)
(addenv_pure "'add"     (fun_2 __add));;


(** 
  @type int_int_int
  @type float_float_float
  @type int_float_float
  @type float_int_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float
  @type posgbf_posgbf_posgbf

  Computes the substraction of its arguments. In the substraction
  between:
  {ul
  {- a float and an integer, the integer is first converted into
  a float.
  }
  {- a float and a big integer, the big integer is first converted into
  a float.
  }
  {- a big integer and an integer, the integer is first converted into
  a big integer.
  }
  }
  Between GBF elements [|x> - |y>] computes [|x> + (-1*|y>)].
*)
(addenv_pure "'-"       (fun_2 __sub));;

(**
  @type int_int_int
  @type float_float_float
  @type int_float_float
  @type float_int_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float
  @type posgbf_posgbf_posgbf
  @alias math.'sub 
  Prefix form of operator [-]
  @voir '- -
*)
(addenv_pure "'sub"     (fun_2 __sub));;

(** 
  @type int_posgbf_posgbf
  @type posgbf_int_posgbf
  @type int_float_float
  @type float_int_float
  @type float_float_float
  @type int_int_int
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float

  Computes the multiplication of its arguments. In the multiplication
  between:
  {ul
  {- a float and an integer, the integer is first converted into
  a float.
  }
  {- a float and a big integer, the big integer is first converted into
  a float.
  }
  {- a big integer and an integer, the integer is first converted into
  a big integer.
  }
  }
  
  The multiplication of a GBF element by an integer [n] computes the
  [n]-fold sum, i.e.: [n*|x>] computes [|x> + |x> + ... + |x>] (the sum has
  [n] terms). If [n] is negative, i.e. [n=-m] where [m] is positive,
  then [n*|x> = -1*(m*|x>)]. And [-1*|x>] equal [<x|] the inverse of
  the element [|x>].
*)
(addenv_pure "'*"       (fun_2 __mult));;

(** 
  @type int_posgbf_posgbf
  @type posgbf_int_posgbf
  @type int_float_float
  @type float_int_float
  @type float_float_float
  @type int_int_int
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float
  @alias math.'mult
  Prefix form of operator [*]
  @voir '* *
*)
(addenv_pure "'mult"    (fun_2 __mult));;

(** 
  @type int_int_int
  @type float_float_float
  @type float_int_float
  @type int_float_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float

  Computes the division of the first argument by the second. In the division
  between:
  {ul
  {- a float and an integer, the integer is first converted into
  a float.
  }
  {- a float and a big integer, the big integer is first converted into
  a float.
  }
  {- a big integer and an integer, the integer is first converted into
  a big integer.
  }
  }
 *)
(addenv_pure "'/"       (fun_2 __div));;

(** 
  @type int_int_int
  @type float_float_float
  @type float_int_float
  @type int_float_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float
  @alias math.'div
  Prefix form of operator [/]
  @voir '/ /
*)
(addenv_pure "'div"     (fun_2 __div));;


(** 
  @type int_float_float
  @type float_int_float
  @type float_float_float
  @type int_int_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float

  Computes the exponentiation of its first argument to the second argument.
*)

(addenv_pure "'**"       (fun_2 __pow));;

(** 
  @type int_float_float
  @type float_int_float
  @type float_float_float
  @type int_int_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  @type float_bint_float
  @type bint_float_float

  Computes the exponentiation of its first argument to the second argument.
*)

(addenv_pure "'pow"       (fun_2 __pow));;


(**
  @type any_any_bool
  The equality predicate.
  To be equal, the two arguments must have the same type. Provided
  that the two arguments have the same type, then
  {ul
  {- two [undef] values are always equal, whatever are the associated
  comments: [<undef> == <undef: a comment>] returns [true].
  }
  {- two [funct] are always differents. Even if the arguments are
  physically the same. For example:
  {[ let f = \x.x+1 in (f==f), (f==\y.1+y) ]}
  returns [(false, false):'seq]
  }
  {- two collections are equals if they have the same (equal) elements
  and the same organization.
  }
  }
*)
(addenv_pure "'=="      (fun_2 equalMgsVal));;

(**  
  @type any_any_bool
  @alias math.'equal
  Prefix form of operator [==]
  @voir '== ==
*)
(addenv_pure "'equal"   (fun_2 equalMgsVal));;



(** 
  @type any_any_bool
  The not-equal predicate. [x != y] evaluates to [not(x == y)].
  This operator can also be written as [<>], [~=] and [=!=].
*)
(addenv_pure "'!="      (fun_2 notequalMgsVal));;

(** 
  @type any_any_bool
  @alias math.'notequal
  Prefix form of operator [!=]. Other infix forms are [<>], [~=] and [=!=].
  @voir '!= !=
*)
(addenv_pure "'notequal"  (fun_2 notequalMgsVal));;



(**  
  @type any_any_undef
  @type any_any_bool
  The physical equality predicate.
  To be physically equal, the arguments must have the same
  representation (but not necessarilly at the same memory
  adresse). For example:
  {[  let s1 = iota(10, set:()) 
  and s2 = (iota(5, set:()), (9, set:())), (5,  6, 7, 8, 9, set:()) 
  in (s1 == s2), (s1 === s2), ("abc" === "a"+"b"+"c")]}
  returns [(true, false, true):'seq]: the set [s1] and [s2] are equal
  (they have the same elements) but do not have the same physical
  representation and the the value computed by expression
  ["a"+"b"+"c"] has the same representation as the string ["abc"].

  Beware that the physical comparaison of two [funct] generally fails
  and returns an [undef] value. For instance:
  [(\x.x) === (\y.y)] returns [true] but [(\x.x+1) === (\y.1+y)]
  returns [<undef: funct comparaison 'equal: functional value'>].
*)
(addenv_pure "'==="     (fun_2 pequal));;

(**
  @type any_any_bool
  @alias math.'pequal
  Prefix form of operator [===].
  @voir '=== ===
*)
(addenv_pure "'pequal"  (fun_2 pequal));;

(**  
  @type any_any_bool
  The negation of the [===] operator: [x !=== y] is the same as [not(x
  === y)].
  @voir math.'=== ===
*)
(addenv_pure "'!==="    (fun_2 pnotequal));;

(** 
  @type any_any_bool
  @alias math.'notpequal
  Prefix form of operator [!===]. 
  @voir '!=== !===
*)
(addenv_pure "'notpequal"    (fun_2 pnotequal));;



(**
  @type any_any_int
  The generic comparison.  

  This functions coincides with
  the usual ordering over integers, characters, strings
  and floating-point numbers, and extend them to a
  total ordering over all types except [funct].
  
  [compare(x, y)] returns [0] if [x == y], a negative integer if
  [x<y], and a positive integer if [x>y]. The same restrictions
  as for [==] apply: i.e., the comparison of two [funct] is
  meaningless:
  {[let a = \x.x and b = \y.y*y in compare(a, b), compare(b, a)]}
  returns [(1, 1):'seq].
*)
(addenv_pure "'compare" (fun_2 __compareMgsVal));;

(**
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  Less than predicate.
  Numerical comparaison between integers, big integers and floats.
  Lexicographic comparison between strings.
*)
(addenv_pure "'<"       (fun_2 __lt));;

(** 
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  @alias math.'lt
  Prefix form of operator [<]. 
  @voir '< <
*)
(addenv_pure "'lt"      (fun_2 __lt));;


(**
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  Less or equal predicate.
  Numerical comparaison between integers, big integers and floats.
  Lexicographic comparison between strings.
*)
(addenv_pure "'<="      (fun_2 __leq));;

(** 
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  @alias math.'leq
  Prefix form of operator [<=]. 
  @voir '<= <=
*)
(addenv_pure "'leq"     (fun_2 __leq));;



(**
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  Greather than predicate.
  Numerical comparaison between integers, big integers and floats.
  Lexicographic comparison between strings.
*)
(addenv_pure "'>"       (fun_2 __gt));;

(** 
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  @alias math.'gt
  Prefix form of operator [>]. 
  @voir '> >
*)
(addenv_pure "'gt"      (fun_2 __gt));;



(**
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  Greather or equal predicate.
  Numerical comparaison between integers, big integers and floats.
  Lexicographic comparison between strings.
*)
(addenv_pure "'>="      (fun_2 __geq));;

(** 
  @type string_string_bool
  @type int_float_bool
  @type float_int_bool
  @type float_float_bool
  @type int_int_bool
  @type bint_bint_bool
  @type int_bint_bool
  @type bint_int_bool
  @type float_bint_bool
  @type bint_float_bool
  @alias math.'geq
  Prefix form of operator [>=]. 
  @voir '>= >=
*)
(addenv_pure "'geq"     (fun_2 __geq));;

(**
  @type float_int_float
  @type int_float_float
  @type float_float_float
  @type int_int_int
  @type float_bint_float
  @type bint_float_float
  @type bint_int_bint
  @type int_bint_bint
  @type bint_bint_bint

  Exponentiation. 
*)
(addenv_pure "'power"   (fun_2 __power));;


(**
  @type float_int_float
  @type int_float_float
  @type float_float_float
  @type int_int_float
  @type float_bint_float
  @type bint_float_float
  @type bint_bint_float
  @type bint_int_float
  @type int_bint_float

  The [atan2] function calculates the arc tangent of the two
  arguments [x] and [y]. It is similar to calculating the arc tangent
  of [y / x], except that the signs of both arguments are used to
  determine the quadrant of the result. 
  This function  returns  the result in radians, which is between [-Pi] and [PI]
  (inclusive).
*)
(addenv_pure "'atan2"   (fun_2 __atan2));;


(**
  @type float_int_float
  @type int_float_float
  @type float_float_float
  @type int_int_int
  @type float_bint_float
  @type bint_float_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint

  Integer and float remainder.
  {ul
  {- Between integers, if [y] is not zero, the result of [x mod y]
  satisfies the following properties: [x == (x / y) * y + x mod y] and
  [abs(x mod y) < abs(y)]. If [y == 0], [x mod y] raises
  [Division_by_zero]. Notice that [x mod y] is negative if [x < 0].
  }
  {- Between floats, [mod(a, b)] the returned value is [a -. n *. b], where [n]
  is the quotient [a / b] rounded towards zero to an integer. If [b ==
  0] then ['nan] is returned.
  }
  {- If an arument is an integer and the other is a float, the integer
  is first converted into a float.
  }
  {- If an arument is an integer and the other is a big integer, the integer
  is first converted into a big integer.
  }
  {- If an arument is a big integer and the other is a float, the big integer
  is first converted into a float.
  }
  }
*)
(addenv_pure "'mod"     (fun_2 __modulo));;


(**
  @type any_any_any
  @type string_string_string
  @type int_float_float
  @type float_int_float
  @type float_float_float
  @type int_int_int
  @type bint_float_float
  @type float_bint_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  [max(a,b)] returns [a] if [compare(a, b) > 0]. 

  Gives the usual meaning when the arguments are numerical or strings.
*)
(addenv_pure "'max"     (fun_2 __maximum));;

(**
  @type any_any_any
  @type string_string_string
  @type int_float_float
  @type float_int_float
  @type float_float_float
  @type int_int_int
  @type bint_float_float
  @type float_bint_float
  @type bint_bint_bint
  @type int_bint_bint
  @type bint_int_bint
  [min(a,b)] returns [a] if [compare(a, b) < 0].

  Gives the usual meaning when the arguments are numerical or strings.
*)
(addenv_pure "'min"     (fun_2 __minimum));;


(**
  @type float_float
  @type int_float
  @type bint_float

  Exponential. 
*)
(addenv_pure "'exp" (fun_1 __exp));;

(**
  @type float_float
  @type int_float
  @type bint_float

  The [acos] trigonometric function. [acos(x)] calculates the arc
  cosine of [x]; that is the value whose cosine is [x].  If [x] falls
  outside the range -1 to 1, [acos] return [nan].
*)
(addenv_pure "'acos"       (fun_1 __acos));;

(**
  @type float_float
  @type int_float
  @type bint_float

  The [asin] trigonometric function. [asin(x)] calculates the arc
  sine of [x]; that is the value whose sine is [x].  If [x] falls
  outside the range -1 to 1, [acos] return [nan].
*)
(addenv_pure "'asin"       (fun_1 __asin));;

(**
  @type float_float
  @type int_float
  @type bint_float

  The [atan] trigonometric function. [atan(x)] calculates the arc
  tangent of [x]; that is the value whose tangent is [x].
*)
(addenv_pure "'atan"       (fun_1 __atan));;

(**
  @type float_float
  @type int_float
  @type bint_float

  The [cos] trigonometric function. [cos(x)] calculates the 
  cosine of [x], where [x] is given in radians.
*)
(addenv_pure "'cos"        (fun_1 __cos));;


(**
  @type float_float
  @type int_float
  @type bint_float

  The [sin] trigonometric function. [sin(x)] calculates the 
  sine of [x], where [x] is given in radians.
*)
(addenv_pure "'sin"        (fun_1 __sin));;

(**
  @type float_float
  @type int_float
  @type bint_float

  The [tan] trigonometric function. [tan(x)] calculates the 
  tangent of [x], where [x] is given in radians.
*)
(addenv_pure "'tan"        (fun_1 __tan));;


(**
  @type float_float
  @type int_float
  @type bint_float

  The [cosh] hyperbolic function. [cosh(x)] calculates the 
  hyperbolic cosine of [x], which is defined mathematically as
  [(exp(x) + exp(-x)) / 2].
*)
(addenv_pure "'cosh"       (fun_1 __cosh));;

(**
  @type float_float
  @type int_float
  @type bint_float

  The [sinh] hyperbolic function. [sinh(x)] calculates the 
  hyperbolic sine of [x], which is defined mathematically as
  [(exp(x) - exp(-x)) / 2].
*)
(addenv_pure "'sinh"       (fun_1 __sinh));;


(**
  @type float_float
  @type int_float
  @type bint_float

  The [tanh] hyperbolic function. [tanh(x)] calculates the 
  hyperbolic tangent of [x], which is defined mathematically as
  [sinh(x) / cosh(x)].
*)
(addenv_pure "'tanh"       (fun_1 __tanh));;



(**
  @type float_float
  @type int_float
  @type bint_float

  The natural logarithm. [log(0)] return [-inf] and the [log] of a
  negative argument returns [nan].
*)
(addenv_pure "'log"        (fun_1 __log));;


(**
  @type float_float
  @type int_float
  @type bint_float

  The base-10 logarithm. [log10(0)] return [-inf] and the [log10] of a
  negative argument returns [nan].
*)
(addenv_pure "'log10"      (fun_1 __log10));;

(**
  @type float_float
  @type int_float
  @type bint_float
  The non-negative square root. Returns [nan] if its argument is
  strictly negative.
*)
(addenv_pure "'sqrt"       (fun_1 __sqrt));;

(**
  @type float_int
  Round the given float to an integer value.
  [ceil f] returns the least integer value greater than or
  equal to [f].
  [floor f] returns the greatest integer value less than or
  equal to [f].
*)
(addenv_pure "'ceil"       (fun_1 __ceil));;


(**
  @type float_int
  Round the given float to an integer value.
  [floor f] returns the greatest integer value less than or
  equal to [f].
  [ceil f] returns the least integer value greater than or
  equal to [f].
*)
(addenv_pure "'floor"      (fun_1 __floor));;

(**
  @type int_int
  @type bint_bint
  @type float_float
  Return the absolute value of the argument.
*)
(addenv_pure "'abs"        (fun_1 __abs));;

(** 
  @type int_int 
  @type bint_int 
  @type string_int 
  @type float_int

  Conversion into an integer.

  If the argument is a [float], the given floating-point number is
  truncated into an integer. The result is unspecified if it falls
  outside the range of representable integers.

  If the argument is a [bint] that is not representable by a standard
  integer, the result of the conversion is
  [VAL_undef("Scalar: _intof_bint: too big to be cast as an int")].

  If the argument is a [string], the function converts the given string
  to an integer.  The string is read in decimal (by default) or in
  hexadecimal, octal or binary if the string begins with [0x], [0o] or
  [0b] respectively.  Returns [<undef: int_of_string failed>] if the given
  string is not a valid representation of an integer.

  If the argument is a [int], this integer is the returned value.
*)
(addenv_pure "'intof"      (fun_1 __intof));;

(** 
  @type int_int
  @type bint_int
  @type string_int 
  @type float_int 
  
  Conversion into an integer.

  @alias math.'int_of
  @voir 'intof intof
**)
(addenv_pure "'int_of"     (fun_1 __intof));;


(** 
  @type int_bint 
  @type bint_bint 
  @type string_bint 
  @type float_bint

  Conversion into a big integer.

  If the argument is a [float], the given floating-point number is
  truncated into an integer.

  If the argument is a [bint], this big integer is the returned value.

  If the argument is a [string], the function converts the given string
  to a big integer. The string is read in decimal.

*)
(addenv_pure "'bintof"      (fun_1 __bintof));;

(** 
  @type int_bint
  @type bint_bint
  @type string_bint 
  @type float_bint 
  
  Conversion into a big integer.

  @alias math.'bint_of
  @voir 'bintof bintof
**)
(addenv_pure "'bint_of"     (fun_1 __bintof));;


(** 
  @type int_float 
  @type bint_float 
  @type string_float
  @type float_float

  Conversion into a float.

  If the argument is an [int], it is converted into a [float].

  If the argument is an [bint], it is converted into a [float].

  If the argument is a string, the function converts the given string
  to a float. Returns [<undef: float_of_string failed>] if the given
  string is not a valid representation of a float.

  If the argument is a [float], this float is the returned value.
*)
(addenv_pure "'floatof"    (fun_1 __floatof));;

(** 
  @type int_float
  @type bint_float
  @type string_float
  @type float_float

  Conversion into a float.

  @alias math.'float_of
  @voir 'floatof floatof
*)
(addenv_pure "'float_of"   (fun_1 __floatof));;



(**
  @type float_int
  Classify IEEE-754 float numbers.

  The returned integer is:
  {ul
  {- [`ieee_zero]: number is 0.0 or -0.0}
  {- [`ieee_subnormal]: number very close to 0.0, has reduced precision}
  {- [`ieee_infinity]: number is positive or negative infinity}
  {- [`ieee_nan]: not a number (result of an undefined operation)}
  {- [`ieee_normal]: normal number, none of the other case}
  }
*)
(addenv_pure "'what_float" (fun_1 __whatfloat));;


(**
  @type int_int
  @type bint_bint
  @type float_float

  The (numeric) successor function. [succ(x) == x+1]
*)
(addenv_pure "'succ"       (fun_1 __succ));;

(**
  @type int_int_bint
  @type bint_int_bint
  @type int_bint_bint
  @type bint_bint_bint
  Binomial coefficient and number of different combinations.

  [nCp(n,p)] returns the binomial coefficient "[n] choose [p]" or the number of
  [p]-combinations of a set with [n] elements. The result is always a [bint].
*)
(addenv_pure "'nCp"       (fun_2 __cnp));;


(**
  @type int_int_bint
  @type bint_int_bint
  @type int_bint_bint
  @type bint_bint_bint
  Number of different permutations.

  [nAp(n,p)] returns the number of different permutations of [p] objects,
  taken from a pool of [n] objects. The result is always a [bint].
*)
(addenv_pure "'nAp"       (fun_2 __anp));;


(* --- math constante --------- *)


(* voir Ocaml Pervasives.mli *)

(** 
  @module math 
  @type float 
  ieee positive infinite value. 
*)
addenv_pure "'infinity"      (VAL_float Pervasives.infinity);;

(** 
  @type float 
  ieee negative infinite value  
*)
addenv_pure "'neg_infinity"  (VAL_float Pervasives.neg_infinity);;

(**
  @type float 
  ieee {i not a number} value.  
*)
addenv_pure "'nan"           (VAL_float Pervasives.nan);;


(** 
  @type float  
  The largest positive finite value of type [float].  
*)
addenv_pure "'max_float"     (VAL_float Pervasives.max_float);;

(**
  @type float
  The smallest positive, non-zero, non-denormalized value of type
  [float].  
*)
addenv_pure "'min_float"     (VAL_float Pervasives.min_float);;

(** 
  @type float 
  The smallest positive float [x] such that [1.0 +. x <> 1.0].  
*)
addenv_pure "'epsilon_float" (VAL_float Pervasives.epsilon_float);;



(* voir /usr/include/math.h *)

(** 
  @type float
  The float constant {i e}. 
*)
addenv_pure "'E"         (VAL_float 2.7182818284590452354);;   (* e          *)

(** 
  @type float
  The float constant {i log}{_ 2}{i (e)} (the logarithm is taken in
  base 2). 
*)
addenv_pure "'LOG2E"     (VAL_float 1.4426950408889634074);;   (* log_2 e    *)

(** 
  @type float
  The float constant {i log}{_ 10}{i (e)} (the logarithm is taken in
  base 10). 
*)
addenv_pure "'LOG10E"    (VAL_float 0.43429448190325182765);;  (* log_10 e   *)

(** 
  @type float
  The float constant {i log}{_ e}{i (2)} (the logarithm is taken in
  base {i e}). 
*)
addenv_pure "'LN2"       (VAL_float 0.69314718055994530942);;  (* log_e 2    *)

(** 
  @type float
  The float constant {i log}{_ e}{i (10)}  (the logarithm is taken in
  base {i e}). 
*)
addenv_pure "'LN10"      (VAL_float 2.30258509299404568402);;  (* log_e 10   *)

(** 
  @type float
  The float constant {i pi}. 
*)
addenv_pure "'PI"        (VAL_float 3.14159265358979323846);;  (* pi         *)

(** 
  @type float
  The float constant {i pi/2}. 
*)
addenv_pure "'PI_2"      (VAL_float 1.57079632679489661923);;  (* pi/2       *)

(** 
  @type float
  The float constant {i pi/4}. 
*)
addenv_pure "'PI_4"      (VAL_float 0.78539816339744830962);;  (* pi/4       *)

(** 
  @type float
  The float constant {i 1/pi}. 
*)
addenv_pure "'1_PI"      (VAL_float 0.31830988618379067154);;  (* 1/pi       *)

(** 
  @type float
  The float constant {i 2/pi}. 
*)
addenv_pure "'2_PI"      (VAL_float 0.63661977236758134308);;  (* 2/pi       *)

(** 
  @type float
  The float constant {i 2/sqrt(pi)}. 
*)
addenv_pure "'2_SQRT_PI" (VAL_float 1.12837916709551257390);;  (* 2/sqrt(pi) *)

(** 
  @type float
  The float constant {i  sqrt(2)}. 
*)
addenv_pure "'SQRT2"     (VAL_float 1.41421356237309504880);;  (* sqrt(2)    *)

(** 
  @type float
  The float constant {i 1/sqrt(2)}. 
*)
addenv_pure "'SQRT1_2"   (VAL_float 0.70710678118654752440);;  (* 1/sqrt(2)  *)



(* ================================================================= *)

(addenv_impure "'rng_name"        (fun_1 __rngname));;


 (addenv_impure "'rng_defaultseed" (fun_1 __rngdefaultseed));;
 (addenv_impure "'rng_setseed"     (fun_1 __rngsetseed));;
 (addenv_impure "'rng_min"         (fun_1 __rngmin));;
 (addenv_impure "'rng_max"         (fun_1 __rngmax));;
 (addenv_impure "'rng_get"         (fun_1 __rngget));;


(addenv_impure "'rng_uniform"     (fun_1 __rnguniform));;
(addenv_impure "'rng_uniformpos"  (fun_1 __rnguniformpos));;
(addenv_impure "'rng_uniformint"  (fun_1 __rnguniformint));;


(**
  @type monoidal_any
  @type bool_bool
  @type string_string
  @type float_float
  @type int_int

  returns a random value.

  If [n] is an integer,  [random(n)] returns a random integer between
  0 (inclusive) and [n] (exclusive).  [n] must be more than 0 and less
  than 2{^30}.

  If [f] is an float, [random(f)] returns a random floating-point number
   between 0 (inclusive) and [f] (exclusive).  If [f] is negative, the
  result is negative.  If [f] is 0, the result is 0.
  
  If [s] is a string, [random(s)] returns a random string made with the letters of [s].
  The length of the returned string is between 0 and the value of the system variable
  [lrandom_string] (default 11).

  If [b] is a boolean, [random(b)] returns a random boolean.

  If [c] is a collection, [random(c)] returns a randomly choosen element in the collection.
*)
(addenv_impure "'random" (fun_1 __random));; (* !!! impur car ce n'est pas une fonction !!! *)

print_string "done *)";
print_newline();;


print_string "(* loading kernel functions on QUASI RANDOM values into user-space ... ";
flush Pervasives.stdout;;

(**
  @type string_int

  returns a handle (of type [int]) for producing quasi-random values.

  The first argument may be equal to [NIEDERREITER2] or to [SOBOL]. The second argument
  is the [dimension] of the generated quasi-random values.
  {ul
  {- If the first argument is  equal to [NIEDERREITER2] then the dimension can be up to 12 and the sequences of floating values
     will be produced following the algorithm described in 
     [Bratley, Fox, Niederreiter, ACM Trans. Model. Comp. Sim. 2, 195 (1992)].}
  {- If the first arg is equal to [SOBOL] then the dimension can be up to 40 and the sequences
     of floating values will be produced following the algorithm described in
     [Antonov, Saleev, USSR Comput. Maths. Math. Phys. 19, 252 (1980)].}
  }
*)
(addenv_impure "'qrngmake"      (fun_2 __qrngmake));;

(**
  @type int

  initializes the generator of quasi-random values.
*)
(addenv_impure "'qrnginit"      (fun_1 __qrnginit));;

(**
  @type int

  returns a sequence of quasi-random floating values using always
  the same storage array.
*)
(addenv_impure "'qrngget"       (fun_1 __qrngget));;


(**
  @type int

  returns a sequence of quasi-random floating values using each time
  a new storage array.
*)
(addenv_impure "'qrngsample"    (fun_1 __qrngsample));;

(**
  @type string

  returns the name of the generator used.
*)
(addenv_impure "'qrngname"      (fun_1 __qrngname));;


(**
  @type int

  returns the dimension of the space that is filled by the quasi-random generator.

*)
(addenv_impure "'qrngdimension" (fun_1 __qrngdimension));;

(* --- fin de l'installation des fonctions de base --- *)

print_string "done *)";
print_newline();;

(* ======================================================= *)
