(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: mset.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2008-05-21 13:30:27 $
 * Version: $Revision: 1.6 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    set. Voir commentaire dans Seq/seq.ml
 *
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> seq (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = set
 *
 *=========================================================================
 *)


open Global;;
open Def_set;;
open Valeur;;
open Core_set;;
open Type;;
open Dispatch_sig;;



__register_version "$RCSfile: mset.ml,v $" "$Revision: 1.6 $" "$Date: 2008-05-21 13:30:27 $";;

(** alias **)
let set_desc = Type_set.set_desc
;;



(** === Constructeurs ======================================================- **)


let _cons_any_set (e, s) = VAL_set(Core_set.add e (fst s), snd s)

and _cons_undef_set (_, s) = VAL_set s

and _append_set_set (s1, s2) = VAL_set(union (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _join_set_set (s1, s2) = VAL_set(union (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _join_any_set (e, s) = if true_coll e then Seq._join_any_any (e, VAL_set(s)) else VAL_set(Core_set.add e (fst s), snd s)
			       
and _join_set_any (s, e) = if true_coll e then Seq._join_any_any (VAL_set(s), e) else VAL_set(Core_set.add e (fst s), snd s)

and _join_set_undef (s, _) = VAL_set s

and _join_undef_set (_, s) = VAL_set s
;;


let  _shape_set (b, deco) = 
  let taille = Core_set.cardinal b
  in VAL_funct 
       (function 
	  | VAL_seq (s, _) -> 
	      if (taille <> Array.length s)
	      then raise_error "shape (set) : argument is not of the same size of the original"
	      else VAL_set ((Array.fold_right (fun x y -> Core_set.add x y) s Core_set.empty), deco)
	  | _  -> raise_error "shape (set) : a sequence as argument is required")
;;


let _sequify_set s = VAL_seq(Array.of_list (elements (fst s)), Type_seq.seq_desc) ;;

let _setify_set s = VAL_set(fst s, set_desc)
and _setify_seq s = VAL_set(Array.fold_left (fun ss x -> Core_set.add x ss) Core_set.empty (fst s), set_desc)
and _setify_any x = 
  if not(true_coll x) then VAL_set(singleton(x), set_desc)
  else 
    Dispatch_sig.fold 
      (VAL_funct(fun x -> VAL_funct(function
				      | VAL_set(y, t) -> VAL_set((Core_set.add x y), t)
				      | _ -> failwith "internal error: setify_any")))
      set_desc.empty 
      x
;;


let _iota_int_set (n, (v, col)) = 
  if ( n < 0)
  then raise (Erreur "iota_set with a negative argument")
  else (let res = ref v
	in
	  for i = 0 to (n-1)
	  do
	    res := Core_set.add (VAL_int i) !res
	  done;
	  VAL_set (!res, col))
;;
 
let _inter_set_set ((s1,t1),(s2,t2)) =
  let t = Type.commontype t1 t2 in
    VAL_set(Core_set.inter s1 s2, t)
;;

let _diff_set_set ((s1,t1),(s2,t2)) =
  let t = Type.commontype t1 t2 in
    VAL_set(Core_set.diff s1 s2, t)
;;


(** Extension --------------------------------- **)


let rec extension_set desc a = 
  VAL_set(Array.fold_left add_elem Core_set.empty a, desc)

and add_elem set = function
  | VAL_seq(aa, _) -> Core_set.add aa.(1) set
  | _ -> raise (Erreur "Internal Error: extension: add_elem in extension_set")
;;



(** === Prdicats, accs et observateurs =========================================== **)

(** _compareMgsVal_set_set est dfini dans core_set *)

let _val2bool_set (b, _) = VAL_bool(not(Core_set.is_empty b))
;;


let _size_set (b, _) = VAL_int(cardinal b)
and _empty_set (b, _) = VAL_bool(b = Core_set.empty)
;;



let _count_any_set (a, (b, _)) = VAL_int (if (Core_set.mem a b) then 1 else 0)
;;

let _member_any_set (a, (b, _)) = VAL_bool (Core_set.mem a b) 



exception Ok of mgs_value;;

let _random_set(s, _) = 
  let n = Core_set.cardinal s in
    if n == 0 then VAL_undef "random: bad argument: empty set" 
    else 
      let nn = ref(Random.int n) in
	try
	  Core_set.iter (function x -> if !nn == 0 then raise (Ok x) else decr(nn)) s;
	  assert(false);
	  VAL_undef("internal error in _random_set")
	with Ok x -> x
;;



let _hd_set s = 
  let ss = fst s in if (0 = cardinal ss) then VAL_undef("hd on empty set") else choose ss

and _tl_set s = 
  let ss = fst s 
  in 
    if (0 = cardinal ss)
    then VAL_undef("tl on empty set") 
    else VAL_set(remove (choose ss) ss, snd s)
;;

(** take_set_int est pris en charge par le _take_any_int dfini dans Seq/seq.ml **)



(* === Fold et autre itrations ================================ *)

let _fold_funct_any_set (fonc, zero, (set, _)) = 
  Core_set.fold  (fun v suite -> (app_2 fonc v suite)) set zero

and _map_funct_set (fonc, (set, c)) = 
  VAL_set ((Core_set.fold (fun v suite -> Core_set.add (fonc v) suite) set Core_set.empty), c)

and _iter_funct_set (fonc, (set, _)) = 
  (Core_set.iter (fun x   -> fonc x; ()) set);  VAL_undef "iter set"

and _mapindexed_funct_set (fct, (set, t)) = 
  let elts = Core_set.elements set
  and cpt = ref (-1)
  in VAL_set((List.fold_left
		(fun acc v -> incr(cpt); Core_set.add (app_2 fct (VAL_int !cpt) v) acc) 
		Core_set.empty 
		elts),
	     t)

and _iterindexed_funct_set (fonc, (set, _)) = 
  let idx = ref 0 in 
    Core_set.iter (fun x -> ignore(app_2 fonc (VAL_int !idx) x); incr idx) set;  
    VAL_undef "iterindexed set"

and _foldindexed_funct_any_set (fct, zero, (set, _)) = 
  let cpt = ref (-1)
  in Core_set.fold (fun v acc -> incr(cpt); app_3 fct (VAL_int !cpt) v acc) set zero

and _forall_funct_set (p, (b, _)) = VAL_bool(Core_set.for_all (fun x -> Auxfct.bool_from_val(p x)) b)

and _exists_funct_set (p, (b, _)) = VAL_bool(Core_set.exists (fun x -> Auxfct.bool_from_val(p x)) b)

and _filter_funct_set (p, (b, t)) = VAL_set(Core_set.filter (fun x -> Auxfct.bool_from_val(p x)) b, t)
;;





(* === Neighbors, neighbor ================================ *)

let _neighbors_set_int ((t, _), i) = 
  let sz         = Core_set.cardinal t
  and retourne v = VAL_seq (v, Type_seq.seq_desc)
  in
    if (i < 0) or (i >= sz)
    then VAL_undef "neighbors (set): undefined neighbor"
    else
      let ret = Array.make (sz-1) (VAL_undef "")
      and idx = ref 0
      and ptr = ref 0
      in
	Core_set.iter (fun v -> if ((post_inc idx) <> i) then (ret.(post_inc ptr) <- v)) t;
	retourne ret
;;





(* === Fonctions spcifiques ======================================================- *)


(* Flatten ----------------------------------------------------------- *)


let _flatten_set (s, t) =
  let ss = Core_set.fold
	     (fun v suite -> match v with 
		| VAL_set (a, _) -> Core_set.union a suite
		| x -> Core_set.add x suite)
	     s 
	     Core_set.empty
  in VAL_set(ss, t)
;;

let _subset_set_set ((a,_),(b,_)) = VAL_bool(Core_set.subset a b) ;;
  

(** Une fonction partition sur les monoides ----------------------------------- *)



let rec _partition_int_set (n, (s, t)) =
  let ls = split_set n s in
  let ss = List.fold_left (fun acc x -> Core_set.add (VAL_set(x, t)) acc) Core_set.empty ls
  in VAL_set(ss, t)

(* On utilise le predicat de Set.partition en faisant un effet de
 * bord. En esperant qu'on appele ce predicat qu'une seule fois pour
 * chaque element, ce qui permet de correler le nombre de fois ou il a
 * ete appele et le nombre d'lment selectionn 
 *)

and split_set n s = 
  if n >= Core_set.cardinal s then [s] else
    begin
      let i = ref 0 in 
      let fct x = (incr i; !i <= n) in
      let (x, suite) = Core_set.partition fct s 
      in x::(split_set n suite)
    end
;;




(** Partition en 2 classes suivant un predicat ----------------------------------------------- *)


let   _partition_funct_set (f, (s, t)) = 
  let (s1, s2) = Core_set.partition (function x -> Auxfct.bool_from_val (f x)) s
  in VAL_set(Core_set.add (VAL_set(s1, t)) (Core_set.singleton (VAL_set(s2, t))), 
	     t)
;;



(** Une fonction partition par quivalence sur les monoide ----------------------------------- *)

let rec _partitioneq_funct_set (f, (s, t)) =
  let eql = ref [] 
  and fill acc (_, lx) = Core_set.add (VAL_set(make_set_from_list !lx, t)) acc
  in 
    begin
      Core_set.iter (function e -> update2 eql f e) s;
      VAL_set(List.fold_left fill Core_set.empty !eql, t)
    end
    
and make_set_from_list l = 
  List.fold_left (fun acc x -> Core_set.add x acc) Core_set.empty l

and update2 eql f x = eql := Auxfct.update_equivalence f x !eql
;;



(* ================================================================ *)

(*** !!! Ne pas oublier de mettre  jour les champs
 ***     _ make 
 ***     _ xxxify
 ***     _ empty
 *** du descripteur de type 
 ***)

set_desc.empty <- VAL_set(Core_set.empty, set_desc);;
set_desc.xxxify <- VAL_funct(setify);;
set_desc.make <- extension_set set_desc;;
