(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: lexer.mll,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-09-17 15:43:04 $
 * Version: $Revision: 1.7 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

{
open Global;;
open Parser;;

exception Eof              ;;
exception Unknown_character;;

__register_version "$RCSfile: lexer.mll,v $" "$Revision: 1.7 $" "$Date: 2007-09-17 15:43:04 $";;

let comment = ref 0;;



(*---------------------------------------------------------------------*)
(* Taken from  /usr/lib/ocaml/source/parsing/lexer.mll                 *)
(* from X. Leroy & al                                                  *)
(*---------------------------------------------------------------------*)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(*---------------------------------------------------------------------*)


(* To buffer string literals *)

let initial_string_buffer = String.create 256
let string_buff = ref initial_string_buffer
let string_index = ref 0

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0

let store_string_char c =
  if !string_index >= String.length (!string_buff) then begin
    let new_buff = String.create (String.length (!string_buff) * 2) in
      String.blit (!string_buff) 0 new_buff 0 (String.length (!string_buff));
      string_buff := new_buff
  end;
  String.unsafe_set (!string_buff) (!string_index) c;
  incr string_index

let get_stored_string () =
  let s = String.sub (!string_buff) 0 (!string_index) in
  string_buff := initial_string_buffer;
  s

(* To translate escape sequences *)

let char_for_backslash =
  match Sys.os_type with
  | "Unix" | "Win32" | "Cygwin" ->
      begin function
      | 'n' -> '\010'
      | 'r' -> '\013'
      | 'b' -> '\008'
      | 't' -> '\009'
      | c   -> c
      end
  | "MacOS" ->
      begin function
      | 'n' -> '\013'
      | 'r' -> '\010'
      | 'b' -> '\008'
      | 't' -> '\009'
      | c   -> c
      end
  | x -> (raise (Erreur "unknown system type"))

let char_for_decimal_code lexbuf i =
  let c = 100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
           10 * (Char.code(Lexing.lexeme_char lexbuf (i+1)) - 48) +
                (Char.code(Lexing.lexeme_char lexbuf (i+2)) - 48) in  
  if ((c < 0) || (c > 255 ))
  then raise (Erreur ("Illegal escape char"))
  else Char.chr c

(* To store the position of the beginning of a string and comment *)
let string_start_pos = ref 0

(*---------------------------------------------------------------------*)
(* END of Leroy's code (to be continued in the lexer itself !)         *)
(*---------------------------------------------------------------------*)
}


let digit       = ['0'-'9']
let hex_literal = '0' ['x' 'X'] ['0'-'9' 'A'-'F' 'a'-'f']+
let oct_literal = '0' ['o' 'O'] ['0'-'7']+
let bin_literal = '0' ['b' 'B'] ['0'-'1']+
let ident       = ['A'-'Z''a'-'z''\'''?']+['a'-'z''A'-'Z''0'-'9''_''\'''?''!']* 
                | '_'['a'-'z''A'-'Z''0'-'9''_''\'''?''!']+
let textundef   = [^'\n''>']+
let parrow      = ['*''+''''@''^''|''%''#'':''!''?''$''''&''('')']
let tige        = ['-''=''_']


rule token = 
parse
  | "\\/"          { DOWNARROW }
  | "/\\"          { UPARROW   }
  | "(*"           { incr comment; parse_comment lexbuf            }
  | "/*"           { incr comment; parse_comment lexbuf            }
  | "\\\\"         { ignore_until_eol lexbuf                       }
  | "//"           { ignore_until_eol lexbuf                       }
  | [' ' '\t']+    { token lexbuf      (* skip blanks *)           }
  | ['\n' '\r']+   { token lexbuf      (* skip carriage returns *) }


  | "begin"        { BEGIN       }
  | "end"          { END         }

  | "<undef>"               { UNDEF("")   }
  | "<undef:"(textundef)">" { let ss = Lexing.lexeme lexbuf
			      in UNDEF(String.sub ss 7 ((String.length ss)-8)) }

  | '|'(ident)'>'  { GENID(Lexing.lexeme lexbuf)     }
  | '<'(ident)'|'  { GENID_INV(Lexing.lexeme lexbuf) }
  | '''(ident)     { QID(Lexing.lexeme lexbuf)       }
  | '`'(ident)     { BQID(Lexing.lexeme lexbuf)      }
  | '<'(ident)'>'  { GENDIR(Lexing.lexeme lexbuf)    }

  | "/"(digit+)"/" {  let ss = Lexing.lexeme lexbuf   (* notation des simplex: /uid/ *)
		      in ACELL(int_of_string(String.sub ss 1 ((String.length ss) -2))) }

  | "{|"           { LENUMER }
  | "|}"           { RENUMER }

  | digit+          { INT(int_of_string(Lexing.lexeme lexbuf)) }
  | hex_literal     { INT(int_of_string(Lexing.lexeme lexbuf)) }
  | oct_literal     { INT(int_of_string(Lexing.lexeme lexbuf)) }
  | bin_literal     { INT(int_of_string(Lexing.lexeme lexbuf)) }

  | digit+ ('.' digit*)? (['e' 'E'] ['+' '-']? digit+)? { REAL(float_of_string (Lexing.lexeme lexbuf))  }

  | "\""            { reset_string_buffer();
		      let string_start = Lexing.lexeme_start lexbuf 
		      in
			string_start_pos := string_start;
			string lexbuf;
			lexbuf.Lexing.lex_start_pos <-	string_start - lexbuf.Lexing.lex_abs_pos;
			STRING (get_stored_string())
		    }

  | "'" '\\' _ "'"  { raise (Erreur ("Illegal escape char")) }

  | "=>"            { ARROW               }
  | "+=>"           { PLUS_ARROW          }
  | "->"            { ABSTRACT_ARROW      }
  | "+->"           { PLUS_ABSTRACT_ARROW }


  | parrow+tige+">"+              { IDENT_ARROW(Lexing.lexeme lexbuf) }
  | parrow*tige+(ident)tige+">"+  { IDENT_ARROW(Lexing.lexeme lexbuf) }
  | tige+"{"                      { ARROW_BEGIN }
  | "}"tige*">"+                  { ARROW_END   }

  | "_"            { UNDERSCORE }
  | "^"            { HAT      }

  | "="            { EQUAL    }
  | ":="           { UPDATE   }
  | "@="           { ATEQUAL  }
  | "&"            { AND      }
  | "||"           { PIPEPIPE }
  | "|"            { PIPE     }
		     
  | "<>"           { NEQ      }
  | "~="           { NEQ      }
  | "!="           { NEQ      }
  | "=!="          { NEQ      }

  | "<"            { LT       }
  | "<="           { LE       }
  | ">"            { GT       }
  | ">="           { GE       }
  | "=="           { EQ       }
  | "==="          { EQ3      }
  | "<|"           { LTB      }
  | "|>"           { GTB      }

  | "++"           { DPLUS    }
  | "+"            { PLUS     }
  | "-"            { MINUS    }
  | "**"           { DTIME    }
  | "*"            { TIME     }
  | "/"            { DIV      }
  | "%"            { MOD      }
  | "<<"           { INF2     }
  | "<<<"          { INF3     }
  | ">>"           { SUP2     }
  | ">>>"          { SUP3     }

  | "::"           { CONS     }
  | "@"            { AT       }
    			     
  | "~"            { NOT      }

  | "("            { LPARENT  }
  | ")"            { RPARENT  }
  | "["            { LCROCHET }
  | "]"            { RCROCHET }
  | "{"            { LBRACKET }
  | "}"            { RBRACKET }
	             
  | "#"            { SHARP       }
  | ":"            { COLUMN      }
  | ";"            { SEMI_COLUMN }
  | ";;"           { TERMINATOR  }
  | ","            { COMMA       }
  | ",,"           { COMMA2      }
  | ",,,"          { COMMA3      }

  | ",,"           { LIST_COMMA  }
  | "."            { DOT         }
  | "\\"           { LAMBDA      }
  | "!!"           { ASSERTION_MARK   }

  | "!"            { commande lexbuf  }

  | ident          { let result = Lexing.lexeme lexbuf
                     in try
                         Keyword.get_keyword result
                       with Not_found -> ID result
                   }
  | eof            { raise Eof }
  | _              { raise Unknown_character }



and commande = parse 

    [' ']+               { commande lexbuf      (* skip blanks *)           }

  | (['A'-'Z''a'-'z']+)  { let result = Lexing.lexeme lexbuf
			   in 
			     try
                               Keyword.get_commande(result)
                             with Not_found -> (print_string "\nNon existing command.\n\n";
						flush stdout;
						token lexbuf)
                         }
  |  _   { print_string "\nUnrecognized command\n\n"; flush stdout; ignore_until_eol lexbuf }


and parse_comment = parse
    '('+'*'        { incr comment; parse_comment lexbuf }
  | '*'+')'        { decr comment; (if (!comment = 0) then token lexbuf else parse_comment lexbuf)}
  | '/'+'*'        { incr comment; parse_comment lexbuf }
  | '*'+'/'        { decr comment; (if (!comment = 0) then token lexbuf else parse_comment lexbuf)}

  | '*'               { parse_comment lexbuf }
  | '('               { parse_comment lexbuf }
  | [^'(''*''/''\n']* { parse_comment lexbuf }
  | '\n'              { parse_comment lexbuf }
  | '/'               { parse_comment lexbuf }



and ignore_until_eol = parse
    "\n"      { token lexbuf }
  | [^'\n']*  { ignore_until_eol lexbuf }


and string = parse
    '"'
      { () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *
      { string lexbuf }
  | '\\' ['\\' '"' 'n' 't' 'b' 'r']
      { store_string_char(char_for_backslash(Lexing.lexeme_char lexbuf 1));
        string lexbuf }
  | '\\' ['0'-'9'] ['0'-'9'] ['0'-'9']
      { store_string_char(char_for_decimal_code lexbuf 1);
         string lexbuf }
  | eof
      { raise (Erreur "Unterminated string")}
  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        string lexbuf }

(* --- End ------------------------------------------------------- *)
