(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: pattern.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-16 16:04:03 $
 * Version: $Revision: 1.9 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *          
 * Une topo (defini dans topology et topo_XXX pour la collection XXX)
 * est une structure qui embarque :
 *
 * - une collection
 * - la notion de position dans cette collection
 * - les deplacements dans cette collection
 * - les elements deja utilise dans le pattern-matching (marquage)
 * - la position de l'element courant
 * - la notion de batch qui est l'enregistrement des occurences filtres
 *   de regles et qui permet de construire la collection finale.
 *
 * Un pattern accepte ou non la position et met a jour en accord les
 * elements utilise dans topo. On peut lui demander aussi d'annuler la
 * prise des elements filtres.
 * 
 * A l'issue d'un appel a la methode filtre :
 * _ on renvoie None si aucun chemin n'a ete trouve en partant de topo#cpos,
 *   rien n'est dit sur topo#cpos et rien ne doit etre marque
 *
 * _ on renvoie (Some c) si un chemin c a ete trouve ; de plus, les elements
 *   du chemin sont marques Used et topo#cpos vaut le dernier element du chemin
 *
 *=========================================================================
 *)

open Printf;;


open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Printer_term;;
open Dispatch_sig;;
open Topology;;

__register_version "$RCSfile: pattern.ml,v $" "$Revision: 1.9 $" "$Date: 2006-04-16 16:04:03 $";;

exception FoundPath;;

let is_elementary_path = 
  let blankpat _            = true
  and ctepat _ _ _ _        = true
  and commapat _ d dn p1 p2 = false
  and downpat _ p1 p2       = false
  and guardedpat _ p _      = p
  and orpat _ p1 p2         = (assert(p1 = p2); p1)
  and recordpat _ l         = true
  and iterpat _ _ p         = false
  and notpat _ p            = p
  in
    fold_named_pattern (blankpat, ctepat, commapat, downpat, guardedpat, orpat, recordpat, iterpat, notpat)
;;

    

class ['position] end_pattern =
object(self)
  inherit ['position] pattern [] true

  method minimal_path_length = 0
  method to_string = "<end pattern>"
  method to_full_string = "<end pattern>"
  method is_end_pattern = true
  method filtre topo = Some []
  method required_species = Hashtbl.create 37
end;;


class ['position] blank_pattern patvar next =
object(self)
  inherit ['position] linked_pattern patvar true next

  method minimal_path_length = 1 + next#minimal_path_length

  method to_string = "_"^(next#to_string)
  method to_full_string = "_"^(self#string_var)^" "^(next#to_full_string)

  method filtre topo = 
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ blank_pattern: filtre pos: %s\n" (topo#prpos (topo#cpos)); flush stdout); 
	true 
      end;
      let cpos = topo#cpos in
	begin
	  assert begin if !trace_opt then 
	    (printf "    ~~ blank_pattern: use OK\n"; flush stdout); 
	    true
	  end;
	  topo#use cpos;
	  self#update_elem_variables topo [cpos];
	  match next#filtre topo with
	    | None      -> topo#unuse cpos; None
	    | Some path -> Some (cpos::path)
	end
    end
end;;


class ['position] cte_pattern1 patvar (cte : mgs_value) next =
object(self)
  inherit ['position] linked_pattern patvar true next

  method minimal_path_length = 1 + next#minimal_path_length

  method to_string = (to_string(prsvalue cte))^(next#to_string)
  method to_full_string = (to_string(prsvalue cte))^(self#string_var)^" "^(next#to_full_string)

  method filtre topo = 
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ cte_pattern: filtre pos: %s\n" (topo#prpos (topo#cpos)); flush stdout); 
	true 
      end;
      let cpos = topo#cpos in
	begin
	  assert begin if !trace_opt then 
	    (printf "    ~~ cte_pattern: use OK\n"; flush stdout); 
	    true
	  end;
	  if Dispatch_sig.equal_mgs_value cte (topo#get_value_from_pos cpos)
	  then
	    begin
	      topo#use cpos;
	      self#update_elem_variables topo [cpos];
	      match next#filtre topo with
		| None      -> topo#unuse cpos; None
		| Some path -> Some (cpos::path)
	    end
	  else
	    None
	end
    end

  method required_species =
    begin
      let h = next#required_species in
	try 
	  let cur_nb_occ = Hashtbl.find h cte in
	    Hashtbl.replace h cte (cur_nb_occ + 1) ;
	    h
	with Not_found -> (Hashtbl.replace h cte 1 ; h)
    end

end;;


(* ********************************************************************************
   On est conservatif sur les cte_patternN : on suppose que la topologie puisse
   de pas disposer d'une methode optimale pour determiner rapidement si elle
   a bien $n$ occurences voisines d'une constante. On va donc passer a cette
   classe un object 'alternatif' qui est l'expansion de cte_pattern1 a 'nbocc'. Ainsi
   on a juste a appeller alternatif#filtre si topo#dispose_get_NvalueS_from_pos vaut false.
  ******************************************************************************** *)

class ['position] cte_patternN patvar (cte : mgs_value) nbocc alternatif next =
object(self)
  inherit ['position] linked_pattern patvar true next

  method minimal_path_length = nbocc + next#minimal_path_length

  method to_string = (string_of_int nbocc)^" "^(to_string(prsvalue cte))^(next#to_string)

  method to_full_string = (string_of_int nbocc)^" "^(to_string(prsvalue cte))^(self#string_var)^" "^(next#to_full_string)
			     

  method filtre topo = 
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ cte_patternN: filtre pos: #%d %s\n" nbocc (topo#prpos (topo#cpos)); flush stdout); 
	true 
      end;

      if topo#dispose_get_NvalueS_from_pos
      then 
	begin
	  let cpos   = topo#cpos in
	  match topo#get_NposS_from_pos cpos nbocc cte
	  with
	      Some (chemin, lastpos) -> (List.iter topo#use chemin;
					 self#update_elem_variables topo chemin;
					 topo#goto lastpos;
					 match next#filtre topo
					 with
					   | None      -> (List.iter topo#unuse chemin;
							   topo#goto cpos;
							   None)
					   | Some path -> Some (chemin@path))
			      
	    | None        -> None
	end
      else alternatif#filtre topo
    end

  method required_species =
    begin
      let h = next#required_species in
	try (
	  let cur_nbocc = Hashtbl.find h cte in
	    Hashtbl.replace h cte (cur_nbocc + nbocc) ;
	    h
	) with Not_found -> (Hashtbl.replace h cte nbocc ; h)
    end

  initializer
    assert(nbocc > 1)

end;;




class ['position] record_pattern patvar recspec (varindex : (string*int) list)  next =
object(self)
  inherit ['position] linked_pattern patvar true next

  val specfield = List.map (function (None, ft) -> (None, ft) 
			      | (Some id, ft) -> (Some (List.assoc id varindex), ft)) recspec
  method minimal_path_length = 1 + next#minimal_path_length

  method to_string = (to_string (prspat (RecordPat recspec)))^(next#to_string)
  method to_full_string =
    (to_string (prspat (RecordPat recspec)))^(self#string_var)^" "^(next#to_full_string)

  method private check_and_bind_record (topo : 'position generic_topo) pos = 
    let check_and_bind_field pos s r ft = function
      | None -> Type_rec.check_field s r ft
      | Some idx ->
	  begin
	    match Type_rec.get_field (VAL_symbol (make_symbol_name s)) r ft with
	      | Type_rec.OK(v) -> (topo#set_immediate idx v); true
	      | Type_rec.Good -> true;
	      | Type_rec.RecFailure -> false
	  end
    in 
      match topo#get_value_from_pos pos with
	| VAL_rec(r,_) -> List.for_all (function (var,(s,ft)) -> check_and_bind_field pos s r ft var) specfield 
	| _ -> false


  method filtre topo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ record_pattern: motif %s filtre pos: %s\n" 
	   self#to_full_string (topo#prpos (topo#cpos)); flush stdout); 
	true 
      end;
      let cpos = topo#cpos in
	begin
	  assert begin if !trace_opt then 
	    (printf "    ~~ record_pattern: use OK\n"; flush stdout); 
	    true
	  end;
	  if self#check_and_bind_record topo cpos then
	    begin
	      topo#use cpos;
	      self#update_elem_variables topo [cpos];
	      match next#filtre topo with
		| None      -> topo#unuse cpos; None
		| Some path -> Some (cpos::path)
	    end
	  else None
	end
    end
end;;




(** direction = indication de direction, par exemple x |est> y
 ** dirname = None ou Some X si on a p1 , as X p2
 ** index = index de cette variable dans l'environnement des variables de pattern
 ** pat1 et pat2 correspondent a pat1, pat2
 **)

class ['position] comma_pattern
  (patvar    : int list)
  (direction : direction_specification)
  (dirname   : string option)
  (index     : int)
  (next      : 'position pattern)
  =
object(self)
  inherit ['position] linked_pattern patvar false next

  method minimal_path_length = next#minimal_path_length

  method to_string = 
    (match direction  with
       | AllDirections      -> ", " 
       | SomeDirections tab -> 
	   "<|" ^ (Array.fold_left (fun s x -> s^", "^(to_string (prsvalue x))) "" tab) ^ "|>")
    ^ (if dirname = None then "" else " as "^(from_option dirname)) 
    ^ next#to_string


  method to_full_string = 
    (match direction  with
       | AllDirections      -> " ," 
       | SomeDirections tab -> 
	   " <|" ^ (Array.fold_left (fun s x -> s^", "^(to_string (prsvalue x))) "" tab) ^ "|>")
    ^ (if dirname = None then "" else " as "^(from_option dirname)) 
    ^ self#string_var
    ^ " "
    ^ next#to_full_string


  method filtre topo = 
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ comma_pattern: filtre\n"; flush stdout); 
	true
      end;
      let (n, voisins) = topo#some_neighbors topo#cpos direction
      and found_path = ref [] 
      and cpos = topo#cpos in
	try
	  assert begin if !trace_opt then
	    (printf "    ~~ comma_pattern:  neighbors OK, itere sur %d voisins\n" n;
	     flush stdout);
	    true 
	  end;
	  for i = 0 to n-1 do
	    topo#move voisins.(i);
	    assert begin if !trace_opt then
	      (print_endline ("    ~~ comma_pattern:  move ok to neighbor pos="^(topo#prpos topo#cpos)));
	      true
	    end;
	    (if dirname <> None then topo#set_immediate index (topo#direction_from_to cpos topo#cpos));
	    match next#filtre topo with
	      | None -> None
	      | Some path ->
		  assert begin if !trace_opt then
		    (print_endline "    ~~ comma_pattern:  pat2#filtre ok");
		    true
		  end;
		  found_path := path;
		  raise FoundPath
	  done;
	  None
	with FoundPath -> Some !found_path
    end

  method required_species = next#required_species

end;;


(** filtre pat1 \/ pat2
 **)
class ['position] 
  down_pattern
  patvar
  patvar1
  one_patvar1
  is_elementary_path 
  subtree1 
  subtree2 
  pat1
  pat2
  motifD
  fun_motif_down
  (next  : 'position pattern)  =
object(self)
  inherit ['position] linked_pattern patvar is_elementary_path next

  method minimal_path_length = pat1#minimal_path_length + next#minimal_path_length

  method to_string =  (pat1#to_string ^ " \\/ " ^ pat2#to_string) ^ next#to_string
  method to_full_string =  
    "(" 
    ^ (pat1#to_full_string ^ " \\/ " ^ pat2#to_full_string) 
    ^ ")"
    ^ (self#string_var)
    ^ next#to_full_string
      
  method private update_downTopo_variables =
    match motifD.cm_patvar with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo t ->  topo#set_downTopo idx1 t)
      | idx1::idx2::[] -> (fun topo t -> topo#set_downTopo idx1 t; topo#set_downTopo idx2 t)
      | _ ->  (fun topo t -> List.iter (fun idx -> topo#set_downTopo idx t) motifD.cm_patvar)

  method private update_immediate_variables =
    match patvar1 with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo v ->  topo#set_immediate idx1 v)
      | idx1::idx2::[] -> (fun topo v -> topo#set_immediate idx1 v; topo#set_immediate idx2 v)
      | _ ->  (fun topo v -> List.iter (fun idx -> topo#set_immediate idx v) patvar1)

  method private update_down_variables =
    match patvar with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo v ->  topo#set_env idx1 v)
      | idx1::idx2::[] -> (fun topo v -> topo#set_env idx1 v; topo#set_env idx2 v)
      | _ ->  (fun topo v -> List.iter (fun idx -> topo#set_env idx v) patvar)

  method filtre topo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ down_pattern: filtre\n"; flush stdout); 
	true
      end;
      assert begin if !trace_opt then 
	(printf "    ~~ down_pattern: try pat1 %s\n" pat1#to_string ; flush stdout); 
	true
      end;
      match pat1#filtre topo
      with
	| None     -> None

	| Some [p] -> (
	    let current_pos = topo#cpos 
	    and v           = topo#get_value_from_pos p
	    in
	      if not(is_collection v)
	      then (
		assert begin if !trace_opt then 
		  (printf "    ~~ down_pattern: pat1 failed: can't dive in the value %s (not a collection)\n"
		     (to_string(prsvalue v)); flush stdout); 
		  true
		end;
		topo#unuse p;
		None
	      ) else (
		assert begin if !trace_opt then 
		  (printf "    ~~ down_pattern: pat1 OK: let dive in %s using pat2 %s\n"
		     (to_string(prsvalue v))
		     pat2#to_full_string
		  ; flush stdout); 
		  true
		end;

		let topoD = fun_motif_down v in
		  match (topoD#apply_one_matching motifD) with
		    | Some p' -> (
			assert begin if !trace_opt then 
			  (printf "    ~~ down_pattern: pat2 OK\n" ; flush stdout);
			  true
			end;
			(* On doit rajouter topoD dans l'environnement !!!!                     *)
			(*                                                                      *)
			(* pat1 en filtrant a associer la valeur de l'objet filtr  son patvar *)
			(* Or on veut que pat1#patvar rfre vers la collection priv de pat2   *)
			(* On associe  patvar1 la new_val et  patvar la valeur de onepatvar1  *)
			(*                                                                      *)
			(* On associe galment un lien vers TopoD pour les variables de pat2    *)
			
			let pat1_val = topoD#build_result in
			let down_val = topo#get_env one_patvar1 in

			  assert begin if !trace_opt then 
			    (printf "    ~~ down_pattern: update environment: %s\n" (to_string(prsvalue pat1_val));
			     printf "patvar: ";
			     List.iter (fun i -> printf "%i " i) patvar ; printf "\n" ;
			     printf "patvar1: ";
			     List.iter (fun i -> printf "%i " i) patvar1 ; printf "\n" ;
			     printf "one_patvar: ";
			     printf "%i\nTopo_env:\n" one_patvar1 ;
			     topo#print_env ;
			     printf "TopoD_env:\n" ;
			     topoD#print_env ;
			     flush stdout);
			    true
			  end;
			
			  self#update_down_variables topo down_val ;

			  assert begin if !trace_opt then 
			    (printf "    ~~ down_pattern: update_down_variables\n" ;
			     topo#print_env ;
			     flush stdout);
			    true
			  end;

			  self#update_immediate_variables topo pat1_val ;

			  assert begin if !trace_opt then 
			    (printf "    ~~ down_pattern: update_immediates_variables\n" ;
			     topo#print_env ;
			     flush stdout);
			    true
			  end;

			  self#update_downTopo_variables topo topoD;

			  assert begin if !trace_opt then 
			    (printf "    ~~ down_pattern: environment updated\n" ;
			     topo#print_env ;
			     flush stdout);
			    true
			  end;

			  match next#filtre topo with
			    | None      -> topo#unuse p; None
			    | Some path -> Some (p::path)
		      )
		    | None -> (
			assert begin if !trace_opt then 
			  (printf "    ~~ down_pattern: pat2 failed\n"; flush stdout);
			  true
			end;
			topo#unuse p;
			None
		      )
	      )
	  )

	| Some p   -> (
	    assert begin if !trace_opt then 
	      (printf "    ~~ down_pattern: pat1 failed: can't dive in a path (Nyi)\n"; flush stdout); 
	      true
	    end;
	    topo#unuse_all p;
	    None
	  )
    end

(* ************** A ecrire ***************** *)

  method required_species = next#required_species

end;;



(** Garde: 
 ** pour simplifier la linarisation d'un filtre, une garde ne consomme rien
 ** en terme d'lments filtrs, mais la valeur retourn peut stopper le
 ** filtrage.
 **)
class ['position] guarded_pattern 
  (patvar : int list)    (* la liste de ses noms : mais une guarde n'a jamais de nom *)
  (depend : int list)    (* la liste des index de variables de patterns utilises dans la garde *)
  (exp    : term)        (* la garde sous forme de terme *)
  (guard  : mgs_value)   (* la garde sous forme de fonction mgs attendant la fct environnement *)
  (next   : 'position pattern)
  =
object(self)
  inherit ['position] linked_pattern patvar true next

  method minimal_path_length = next#minimal_path_length

  method to_string = " //" ^  (to_string(Printer_term.prsterm exp)) ^ "//" ^ next#to_string
  method to_full_string = " //" ^  (to_string(Printer_term.prsterm exp)) ^ "//" ^ next#to_full_string
      
  method filtre topo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ guarded_pattern: motif %s filtre from pos %s\n" 
	   self#to_full_string (topo#prpos topo#cpos); 
	 flush stdout);
	true
      end;
      let vguard = (pf1 guard) topo#get_val_of_rule_variable in
	assert begin if !trace_opt then 
	  (printf "    ~~ guarded_pattern: guard = %a\n" Printer.prfvalue2 vguard);
	  true
	end;
	if Auxfct.bool_from_val vguard
	then next#filtre topo
	else None
    end

  method required_species = next#required_species

end;;


(** Update:
 ** pattern "void" insrer  la bonne place pour mettre  jour les
 ** variables qui ne correspondnet pas  un pattern lementaire ou bien
 ** un iterpat. Par exemple, X dans (u, v) as X 
 **)
class ['position] update_pattern 
  (patvar : int list)  
  (v1 : int)
  (v2 : int)
  (next : 'position pattern)
  =
object(self)
  inherit ['position] linked_pattern patvar false next

  method minimal_path_length = next#minimal_path_length

  method to_string = 
    " //" ^  self#string_var 
    ^ " := "^ (string_of_int v1) 
    ^", " ^(string_of_int v2) 
    ^ "//" ^ next#to_string

  method to_full_string = 
    " //" ^  self#string_var 
    ^ " := "^ (string_of_int v1) 
    ^", " ^(string_of_int v2) 
    ^ "//" ^ next#to_full_string
      
  method filtre topo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ update_pattern: motif %s filtre from pos %s\n" 
	   self#to_full_string (topo#prpos topo#cpos); 
	 flush stdout);
	true
      end;
      (* update de la variable *)
      (match topo#get_env v1 with
	 | ImmediateVal _ -> 
	     raise (Erreur "Error: internal: immediate valuein update_pattern (a)")
	 | EltCollection p1 ->
	     (match  topo#get_env v2 with
		| ImmediateVal _ -> raise (Erreur "Error: internal: immediate value in update_pattern (b)")
		| EltCollection p2 -> self#update_path_variables topo [p1; p2]
		| PathCollection p2l -> self#update_path_variables topo (p1::p2l)
		| DownTopo t -> raise (Erreur "Error: internal: dtopo value in update_pattern (b')"))
	 | PathCollection p1l -> 
	     (match  topo#get_env v2 with
		| ImmediateVal _ -> raise (Erreur "Error: internal: immediate valuein update_pattern (c)")
		| EltCollection p2 -> self#update_path_variables topo (p1l@[p2])
		| PathCollection p2l -> self#update_path_variables topo (p1l@p2l)
		| DownTopo t -> raise (Erreur "Error: internal: dtopo value in update_pattern (c')"))
	 | DownTopo t -> raise (Erreur "Error: internal: dtopo value in update_pattern (a')")
      );
      (* on passe  la suite du pattern *)
      next#filtre topo
    end

  method required_species = next#required_species

end;;




(** filtre pat1 | pat2
 **)
class ['position] or_pattern patvar is_elementary_path subtree1 subtree2 pat1 pat2 (next  : 'position pattern)  =
object(self)
  inherit ['position] linked_pattern patvar is_elementary_path next

  method minimal_path_length = (assert (pat1#minimal_path_length = pat2#minimal_path_length); pat1#minimal_path_length + next#minimal_path_length)

  method to_string =  (pat1#to_string ^ " | " ^ pat2#to_string) ^ next#to_string
  method to_full_string =  
    "(" 
    ^ (pat1#to_full_string ^ " | " ^ pat2#to_full_string) 
    ^ ")"
    ^ (self#string_var)
    ^ next#to_full_string
      
  method private update_variables (topo : 'position generic_topo) v = 
    if is_elementary_path
    then self#update_elem_variables topo v
    else self#update_path_variables topo v

  method private essai_pat (topo : 'position generic_topo) (pat : 'position pattern) =
    match (pat#filtre topo) with 
      | None   -> None
      | Some p -> (
	  let current_pos = topo#cpos in
	    self#update_variables topo p;
	    match next#filtre topo with
	      | Some p' -> Some (p@p')
	      | None    -> (topo#unuse_all p; topo#move current_pos; None))

  method filtre topo = 
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ or_pattern: filtre\n"; flush stdout); 
	true
      end;
      let (p1, p2) = if (crandom 2) = 0 then (pat1, pat2) else (pat2, pat1) in
      let ret = self#essai_pat topo p1 in
	if (ret = None)
	then self#essai_pat topo p2
	else ret
    end
end;;







type 'a retour = 
  | Iterpat_failed
  | Nextpat_failed
  | Success of 'a
;;



(** Si on considere  le chemin :  a D x E* F b
 ** alors pour le chemin vide je prends :  a D b
 ** Dans la classe:
 **    a et D viennent de ce qui precede et je n'y accede pas
 **    F est donne par dirname
 **    x c'est iterpat
 **    b c'est contpat (la continuation)
 **    kiter c'est E*
 **)



class ['position] iter_pattern  
  (patvar : int list)
  kiter                          (* type de l'iteration E* ou E+ *)
  (iterpat : 'position pattern)  (* le pattern a iterer *)
  (next : 'position pattern)     (* continuation b *)
  =
object(self)
  inherit ['position] linked_pattern patvar false next

  method private iter_direction = match kiter with
    | PatStar (d, _) -> d
    | PatPlus (d, _) -> d

  method private is_plus_iteration = match  kiter with
    | PatStar _ -> false
    | PatPlus _ -> true

  method minimal_path_length = 
    if self#is_plus_iteration 
    then iterpat#minimal_path_length + next#minimal_path_length
    else next#minimal_path_length

  method private general_to_string pre post = 
    "("^(pre())^")"
    ^(match kiter with
	| PatStar (AllDirections, k)      -> Printer_term.string_of_iter_kind "*" k
	| PatPlus (AllDirections, k)      -> Printer_term.string_of_iter_kind "+" k
	| PatStar ((SomeDirections d), k) -> "<|" ^ (Array.fold_left (fun s x -> s^", "^(to_string (prsvalue x))) "" d) ^ "|>" ^ (Printer_term.string_of_iter_kind "+" k)
	| PatPlus ((SomeDirections d), k) -> "<|" ^ (Array.fold_left (fun s x -> s^", "^(to_string (prsvalue x))) "" d) ^ "|>" ^ (Printer_term.string_of_iter_kind "*" k))
    ^ post()
  
  method to_string      = self#general_to_string (function () -> iterpat#to_string)      (function () -> next#to_string)
  method to_full_string = self#general_to_string (function () -> iterpat#to_full_string) (function () -> self#string_var ^ " " ^ next#to_full_string)

  method private update_vars chemin topo = self#update_path_variables topo chemin

  method filtre topo = 
    begin
      assert(if !trace_opt
	     then 
	       begin
		 Printf.printf "********************************************************************************\nL'iterpat est %s\nLe nextpat est %s\n" iterpat#to_string next#to_string;
		 flush Pervasives.stdout
	       end;
	     true);

      assert (if !trace_opt 
	      then (printf "--- filtre: iter_pat %s: begins with empty path at pos=%s\n" 
		      (if self#is_plus_iteration then "+" else "*")
		      (topo#prpos topo#cpos));
	      true);

      assert(if !trace_opt 
	     then 
	       begin
		 Printf.printf "On va essayer de filtrer l'iterpat suivant %s a partir de la position %s\n" (iterpat#to_string) (topo#prpos topo#cpos);
		 Printf.printf "On demarre en position %s de valeur %s\n" (topo#prpos topo#cpos) (self#string_of_current topo);
		 flush Pervasives.stdout
	       end;
	     true);

      let avant = topo#cpos
      in
	(match (iterpat#filtre topo)
	 with
	   | None   -> (assert(if !trace_opt then Printf.printf "N'a pas filtre..\n"; true);
			topo#move avant; None)

	   | Some p -> (assert(if !trace_opt then Printf.printf "A filtre ; on essaye la suite...\n"; true);
			(match (self#match_fin_et_augmente p topo)
			 with 
			   | Some p' -> (assert(if !trace_opt then Printf.printf "FILTRE : OK     : FOUND\n"; true);
					 (* Printf.printf "Etat du chemin %s\n" (clean_comma_string_of_general_list topo#string_of_usage_position p'); *)
					 topo#goto (List.hd (List.rev p'));
					 (* Printf.printf "L'etat du chemin rendu est %s\n" (clean_comma_string_of_general_list topo#string_of_usage_position p'); *)
					 Some p')

			   | None -> (assert(if !trace_opt then Printf.printf "FILTRE : NON OK : NOT FOUND \n"; true);
				      topo#unuse_all p;
				      None))))
    end

  method private match_fin_et_augmente matched topo =
    let depart = topo#cpos
    in
      self#update_vars matched topo;
      (match next#filtre topo
       with
	 | None   -> (
	     assert(if !trace_opt 
		    then 
		      (Printf.printf "La suite %s n'a pas matchee. On augmente l'iterpat (avant : %s).\n" next#to_string (topo#prpos depart);
		       Printf.printf "Le statut est %s et on est sur la position de depart : %s \n" (topo#string_of_usage_position depart) 
			 (if (depart = topo#cpos)
			  then "oui"
			  else "non")); true);
	     topo#goto depart;
	     self#augmente_iterpat matched topo)
	     
	 | Some p -> (assert(if !trace_opt then Printf.printf "La suite %s a matchee. On augmente l'iterpat.\n" next#to_string; true);
		      Some (matched @ p)))

	     

  method private some_none_of_list l1 l2 = (match l1, l2
					    with
					      | [], [] -> assert(if !trace_opt then Printf.printf "None, None\n"; true);  None, None
					      | [], p  -> assert(if !trace_opt then Printf.printf "None, Some\n"; true); None, (Some p)
					      | p, []  -> assert(if !trace_opt then Printf.printf "Some, None\n"; true); (Some p), None
					      | p, p'  -> assert(if !trace_opt then Printf.printf "Some, Some\n"; true); (Some p), (Some p'))

  method private string_of_current topo = to_string(prsvalue (topo#get_value_from_pos topo#cpos))

  method private augmente_iterpat literpat topo =
    let local_p           = ref []
    and current_pos       = topo#cpos in
    let nb, neighbors     = topo#some_neighbors current_pos self#iter_direction  (* Il faut sans doute remettre cpos, mais quand ? *)
    in
      assert(if !trace_opt then Printf.printf "On enumere sur %d voisins a partir de la position %s de valeur %s\n (" nb (topo#prpos current_pos) (self#string_of_current topo); true);
      flush Pervasives.stdout;
      (try
	 for i = 0 to nb-1
	 do
	   let avant = neighbors.(i)
	   in
	     topo#move avant;
	     
	     assert(if !trace_opt then  Printf.printf ") On se depace sur le voisins No %d de position %s et de valeur %s\n" i (topo#prpos neighbors.(i)) (self#string_of_current topo);
		    true);
	     
	     (match iterpat#filtre topo
	      with
		| None   -> assert(if !trace_opt then  Printf.printf "Le filtrage d'un iterpat de plus : ECHEC, alors que l'on est sur le %d voisin sur %d\n" i (nb-1); true)
		| Some p -> (assert(if !trace_opt then Printf.printf "Le filtrage d'un iterpat de plus : SUCCES\n"; true);
			     let path = (literpat@p)
			     in
			       self#update_vars path topo;
			       assert(if !trace_opt then  Printf.printf "##########On update les variables avec : %s\n" (clean_comma_string_of_general_list topo#prpos p); true);
			       let avant_pos = topo#cpos 
			       in
  				 match next#filtre topo
				 with
				     None    -> (assert(if !trace_opt then  Printf.printf "Filtrage du next : ECHEC ===> On essaye d'augmenter l'ITERPAT\n"; true);
						 topo#goto avant_pos;
						 (match (self#augmente_iterpat path topo)
						  with
						    | Some p' -> (assert(if !trace_opt then Printf.printf "<==== Retour (SOME) : SUCCES.\n"; true);
							         local_p := p';
							         raise FoundPath)
							
						    | None     -> (assert(if !trace_opt then Printf.printf "<==== Retour (NONE) : NONE.\n"; true);
								   topo#unuse_all p;
								   topo#move avant))
						)
				       
				   | Some p' -> (assert(if !trace_opt then Printf.printf "Filtrage du next : SUCCES\n"; true);
						 (* topo#goto (List.hd (List.rev p')); *) (* Il sera fait au FoundPath *)
						 local_p := literpat @ p @ p';
						 raise FoundPath)
			    ))
	 done;
	 None
       with FoundPath -> topo#goto (List.hd (List.rev !local_p)); Some !local_p)
end;;

(** Tag de non consommation:
 **)
class ['position] not_pattern 
  (patvar : int list)          (* la liste de ses noms : mais une noncons n'a jamais de nom *)
  (next   : 'position pattern) (* la continuation                                           *)
  =
object(self)
  inherit ['position] linked_pattern patvar true next

  method minimal_path_length = next#minimal_path_length

  method to_string = " ~"^next#to_string
  method to_full_string = " ~"^next#to_full_string
      
  method filtre topo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ not_pattern: motif %s filtre from pos %s\n" 
	   self#to_full_string (topo#prpos topo#cpos); 
	 flush stdout);
	true
      end;
      match next#filtre topo with
	| None        -> None
	| Some (t::q) -> (
	    assert begin if !trace_opt then 
	      (printf "    ~~ not_pattern: motif %s untag first pos %s\n" 
		 self#to_full_string (topo#prpos t); 
	       flush stdout);
	      true
	    end;
	    topo#force_unuse t ;
	    Some q
	  )
	| Some []     -> raise (Erreur "Pattern: not_pattern: filtre: waiting for a non empty path")
    end

  method required_species = next#required_species

end;;
