(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topology.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2008-06-03 12:59:19 $
 * Version: $Revision: 1.23 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *          
 * Une topo est une structure qui embarque :
 * - une collection
 * - la notion de position dans cette collection
 * - les deplacements dans cette collection
 * - les elements deja utilise dans le pattern-matching (marquage)
 * - la position de l'element courant
 * - la notion de batch qui est l'enregistrement des occurences filtres
 *   de regles et qui permet de construire la collection finale.
 *
 * Un pattern accepte ou non la position et met a jour en accord les
 * elements utilise dans topo. On peut lui demander aussi d'annuler la
 * prise des elements filtres.
 *
 * Le marquage associe a chaque noeud un etat parmi 4:
 *
 *  - PosFree : l'element est libre et n'a jamais ete pris comme depart de chemin
 *
 *  - PosVisited : l'element est libre pour un chemin mais a deja ete considere 
 *                 comme depart de chemin (et cela n'a pas abouti)
 *
 *  - PosUsed_1 : l'element est pris dans un chemin et etait auparavent marque libre
 *
 *  - PosUsed_2 : l'element est pris dans un chemin et etait auparavent marque visite
 *
 * Transition entre marque:
 *
 *                  /<----unuse-------+
 *                 /                  |
 *            posFree  ----use----> PosUsed_1
 *            |     ^
 *       visit|     |
 *            |     |unvisit
 *            v     |
 *            posVisited ---use---> PosUsed_2
 *                  \                 |
 *                   \                v
 *                    \<-----unuse----+
 *
 *=========================================================================
 *)

open Printf;;

open Auxfct;;
open Global;;
open Valeur;;
open Terme;;
open Printer;;
exception DynamicPattern;;
exception StochApplied of bool;;

__register_version "$RCSfile: topology.ml,v $" "$Revision: 1.23 $" "$Date: 2008-06-03 12:59:19 $";;

type ('position, 'a) pattern_variable_value = 
    ImmediateVal of mgs_value
  | EltCollection of 'position
  | PathCollection of 'position list
  | DownTopo of 'a    (** destin  devenir un 'position generic_topo **)

and 'position path = 'position list                           (** un chemin                    **)
and  'position return_path = ('position path) option           (** chemin + env associe         **)
and  ('position, 'a) env = (('position, 'a) pattern_variable_value) array (** liaison d'un nom a un chemin **)

and 'pattern compiled_rule = { rname : string option;
				motif : 'pattern;   (** defini apres l'usage, c'est pourquoi en parametre **)
				rtype : rule_property;
				rhs   : mgs_value;
				nbvar : int 
			      }

and 'pattern compiled_motif = { cm_motif  : 'pattern;
				cm_nbvar  : int;
				cm_patvar : int list;
			      }
;;

let in_place_shuffle_array len a = 
  let nb  = ref (if (len <= 1)
		 then 0
		 else 
		   if (len = 2) 
		   then 1 
		   else len) (* Pour eviter que sur un tableau de longueur 2 on fasse 2 permutations inverses. *)
  in                         (* Le risque est moins grand sur un tableau de taille superieure.                 *)
    while (!nb > 0)
    do
      let src = Random.int len
      and dst = Random.int len
      in
	if (src = dst)
	then ()
	else (let tmp = a.(src)
	      in
		a.(src) <- a.(dst);
		a.(dst) <- tmp;
		decr nb)
    done;
    a
;;

 
(** Acces a la valeur d'une variable de pattern, a partir d'un 
 ** environnement sous forme de tableau: l'attribution d'un index 
 ** (de ce tableau) a une variable du pattern est faite par la methode  
 ** Transformation.bind_pvar. La fonction get_val_of_rule_variable attend 
 ** un VAL_int qui identifie la variable. 
 **
 ** voir dans le fichier transformation pour les indices des variables predefinies 
 **) 
 
let get_val_of_rule_variable =

  (** acces  la valeur d'une variable de filtre **)
  let rec get_value_of_pattern_variable topo = function 
    | 0 -> topo#collection 
    | i ->  
	(try  
           (match (topo#get_env i) with 
              | ImmediateVal v -> v
	      | EltCollection p -> topo#get_value_from_pos p
              | PathCollection l -> VAL_seq((Array.of_list(List.map (topo#get_value_from_pos) l)), Type_seq.seq_desc)
	      | DownTopo t ->
		  let (t' : 'a) = Obj.magic (t) in
		    get_value_of_pattern_variable t' i
	   )
	 with _ ->  
           let err = "Error: internal: do not find pattern variable number "
		     ^(string_of_int i) 
                     ^" in current pattern environment (of size "
		     ^ (string_of_int topo#length_env)
		     ^") (did you use an or-pattern ?)\n" 
           in print_endline err; flush stdout; raise (Erreur err))

  (** acces  la position d'une variable de filtre **)
  and get_pos_of_pattern_variable topo = function 
    | 0 -> raise (Erreur "Error: internal: pos of self")
    | i ->  
	(try  
           (match (topo#get_env i) with 
              | ImmediateVal v -> raise (Erreur "Error: internal: immediate of pos")
	      | EltCollection p -> topo#mgs_value_from_pos p
              | PathCollection l -> VAL_seq((Array.of_list(List.map (topo#mgs_value_from_pos) l)), Type_seq.seq_desc)
	      | DownTopo t ->
		  let (t' : 'a) = Obj.magic (t) in
		    get_pos_of_pattern_variable t' i
	   )
	 with _ ->  
           let err = "Error: internal: do not find pattern variable number "
		     ^(string_of_int i) 
                     ^" in current pattern environment (of size "
		     ^ (string_of_int topo#length_env)
		     ^") (did you use an or-pattern ?)\n" 
           in print_endline err; flush stdout; raise (Erreur err)) 
	
  in (** dispatch entre les 2 types d'accs **)
    fun topo index -> let idx = to_int index in
      begin
	(* Printf.printf "On cherche la variable a l'index %d alors que la taille est %d\n" idx topo#length_env; *)
	if idx < 0 then get_pos_of_pattern_variable topo (0 - idx)
	else get_value_of_pattern_variable topo idx
      end
;; 
 



(* --- gestion du marquage ------------------------------------------*)

type  usage_position = PosFree | PosVisited | PosUsed_1 | PosUsed_2
;;


let pos_visit = function
  | PosFree -> PosVisited
  | _ -> raise (Erreur "Error: internal: cannot visit a non free position")

(*
and pos_use = function
  | PosFree -> PosUsed_1
  | PosVisited -> PosUsed_2
  | _ -> raise (Erreur "Error: internal: cannot use an already used position")
*)
and pos_unvisit = function
  | PosVisited -> PosFree
  | x -> x

(*
and pos_unuse = function
  | PosUsed_1 -> PosFree
  | PosUsed_2 -> PosVisited
  | PosFree -> raise (Erreur "Error: internal: cannot unuse a not-used free position") 
  | PosVisited -> raise (Erreur "Error: internal: cannot unuse a not-used visited position") 
*)
and pos_used = function
  | PosUsed_1 -> true
  | PosUsed_2 -> true
  | _ -> false

and pos_visited = function 
  | PosVisited -> true
  | _ -> false

and pos_usable = function
  | PosFree    -> true
  | PosVisited -> true
  | _ -> false
;;



(** Interface vers une TOPOLOGIE :
 ** c'est une classe abstraite et polymorphe en le type representant
 ** une position.
 **)


exception NoMoreUsable;;
exception NoMoreFound;;
exception AsynchronousFound;;

class virtual ['position] generic_topo 
  (prpos : 'position -> string)
  (coll, couleur)
  (mgsval : mgs_value)
  (init_pos : 'position) =
object (self)

  method collection = mgsval

  method string_of_usage_position p = 
    match self#state p
     with
       | PosFree    -> "PosFree"
       | PosVisited -> "PosVisited"
       | PosUsed_1  -> "PosUsed_1"
       | PosUsed_2  -> "PosUsed_2"

  (** gestion du marquage **)
  method use                 = (let err () = raise (Erreur (Printf.sprintf "Cannot use an already used position : %s with value %s " (prpos self#cpos) (to_string(prsvalue (self#get_value_from_pos self#cpos)))))
				in
				  self#update_state (function
						       | PosFree    -> PosUsed_1
						       | PosVisited -> PosUsed_2
						       | _          -> err ()))
  method visit               = self#update_state pos_visit
  method unvisit             = self#update_state pos_unvisit
  method unuse               = (let err s = raise (Erreur (Printf.sprintf "Cannot unuse an already %s position : %s with value %s " s (prpos self#cpos) (to_string(prsvalue (self#get_value_from_pos self#cpos)))))
				in
				  self#update_state (function
						       | PosUsed_1  -> PosFree
						       | PosUsed_2  -> PosVisited
						       | PosFree    -> err "free"
						       | PosVisited -> err "visited"))

  method force_unuse p       = if (self#used p) then self#unuse p

  method unuse_and_move path p  = (self#unuse_all path;
				   self#move p)

  method used p              = pos_used(self#state p)
  method visited p           = pos_visited(self#state p)
  method usable p            = pos_usable(self#state p)

  method virtual state        : 'position -> usage_position
  method virtual update_state : (usage_position -> usage_position) -> 'position -> unit 

  method virtual next_usable : 'position -> 'position
  method virtual unvisit_all : unit

  method use_all   = List.iter self#use 
  method reuse_all = List.iter (fun x -> if self#usable(x) then self#use(x))
  method unuse_all = List.iter self#unuse
    
  val mutable _cpos = init_pos
  method cpos : 'position = _cpos
  method move p = assert(not(self#used p)); _cpos <- p
  method goto p = _cpos <- p
  method visit_current = self#visit(_cpos)
  method use_current = self#use(_cpos)
  method move_next_usable = try _cpos <- self#next_usable _cpos; true with NoMoreUsable -> false
  method init_pos = init_pos

    
  val mutable _env = ([||] : ('position, 'position generic_topo) env)
  method get_env i = _env.(i)
  method set_env i v = _env.(i) <- v
  method length_env = Array.length _env
  method set_elt i v = _env.(i) <- EltCollection v
  method set_path i v = _env.(i) <- PathCollection v
  method set_immediate i v = _env.(i) <- ImmediateVal v
  method set_downTopo i t = _env.(i) <- DownTopo t
 (* Printf.printf "set_xxx : On cherche la variable a l'index %d alors que la taille est %d\n" i self#length_env; *)
  method print_env =
    Array.iteri
      (fun i a -> match a with
	 | ImmediateVal v    -> printf "%i -> Immediate: %s\n" i (to_string(prsvalue v))
	 | DownTopo t        -> printf "%i -> DownTopo :\n\n" i ; t#print_env ; printf "\n"
	 | EltCollection p   -> printf "%i -> EltColl  : %s @ %s\n" i (to_string(prsvalue (self#get_value_from_pos p))) (prpos p)
	 | PathCollection lp -> printf "%i -> PathColl\n" i
      )
      _env


  method resize_env max_size = 
    let mm = 1 + max_size  (** l'elt 0 est reserv  self **)
    in (if (mm > Array.length _env) 
	then _env <- Array.make mm (ImmediateVal (VAL_undef "uninitialized pattern variable")))
	

  (** affichage d'une position **)
  method prpos = prpos

  (** calcul de la direction qui va de p1 a p2.
   ** On suppose que le resultat est encore une position (?)  **)
  method virtual direction_from_to : 'position -> 'position -> mgs_value

  (** gestion du voisinage :  some_neighbors p dir
   ** les voisins definis suivant dir du point p.
   ** Ca retourne un entier n et le tableau des 0 a (n-1) voisins definis   **)
  method virtual some_neighbors : 'position -> direction_specification -> int * ('position array) 
    (** tous les voisins **)
  method neighbors p = self#some_neighbors p AllDirections

  method virtual get_value_from_pos : 'position -> mgs_value
  method virtual mgs_value_from_pos : 'position -> mgs_value

  method get_val_of_rule_variable = VAL_funct(get_val_of_rule_variable self)

  (** gestion de la reconstruction **)
  method virtual add_match   : rule_property -> 'position path -> mgs_value -> unit
  method virtual build_result : mgs_value
  method iter_batch : ((rule_property * 'position path * mgs_value) -> unit) -> unit = function _ -> raise (Erreur "Topologie: iter_batch: not implemented for this collection")

  (** gestion de la transformation **)
  method transform rules =
    begin
      (try

	 match (Environnement.get_varidx Options.index_var_strategy)
	 with
(* Application maximale //                                                                                                           *)
	     VAL_int (0 | 6) -> (List.iter (fun r -> let _ = self#apply_one_rule true r in ()) rules;
				 self#build_result)

(* Application  la gillespie                                                                                                        *)
	   | VAL_int 1 -> self#apply_gillespie rules

(* Application maximale // sans priorit sur les rgles: type MULTI... A chaque itration une rgle est applique une fois           *)
	   | VAL_int 2 -> self#apply_multiStochastic rules

(* Application maximale // sans priorit sur les rgles: type SINGLE... A chaque itration une rgle est applique de facon maximale *)
	   | VAL_int 3 -> self#apply_singleStochastic rules

(* Application stochastique : a chaque application UNE SEULE REGLE EST APPLIQUE UNE FOIS (donc pas de facon maximale)                *)
	   | VAL_int 4 -> self#apply_true_stochastic rules

(* Application asynchrone : a chaque application une seule regle est applique une fois. Priorit sur les rgles  *)
	   | VAL_int 5 -> self#apply_asynchronous rules

	   | _         -> raise (Erreur "Topology: transform: impossible strategy")

	 with BreakTransformation res -> res)
    end

  method apply_singleStochastic rules =
    try
      Array.iter
	(fun r -> let _ = self#apply_one_rule true r in ())
	(in_place_shuffle_array (List.length rules) (Array.of_list rules)) ;
      self#build_result
    with BreakTransformation res -> res

  method apply_multiStochastic rules =
    (try
       let rr = ref rules
       and nr = ref (List.length rules) in
	 while (!nr <> 0) do
	   let n = Random.int (!nr) in
	   let r = List.nth (!rr) n in
	     if ( (self#apply_one_rule false r) = 0 ) then (
	       decr nr ;
	       rr := List.fold_right ((fun r' acc -> if ((Pervasives.compare r' r) = 0) then acc else r'::acc)) (!rr) ([])
	     )
	 done ;
	 self#build_result
     with BreakTransformation res -> res)

  method apply_asynchronous rules =
    try (
      List.iter (fun r -> if (self#apply_one_rule false r = 1) then raise AsynchronousFound) rules;
      self#build_result
    ) with
      | AsynchronousFound -> self#build_result
      | BreakTransformation res -> res
    
  method apply_true_stochastic rules =
    let epsilon = 0.01 in
    let somme_proba = ref 0.0 in
    let liste = List.map (function
			      (({rtype = {a_stochastic  = (Some v) }}) as rule) -> 
				let proba = (match v
					     with
						 StochCte f -> f
					       | StochFun f -> (match (f mgsval)
								with 
								    VAL_float flo -> flo
								  | VAL_int i     -> float_of_int i
								  | _             -> raise (Erreur "TOPOLOGY: only float/int in the definition of a probability (1)"))
					       | StochExp t -> (
						   let v = (to_funct (Compile.eval t)) (VAL_undef "") in
						     match v with
						       | VAL_float flo -> flo
						       | VAL_int i     -> float_of_int i
						       | VAL_funct f   -> (
							   match (f mgsval) with
							     | VAL_float flo -> flo
							     | VAL_int i     -> float_of_int i
							     | _             ->  raise (Erreur "TOPOLOGY: only float/int in the definition of a probability (3)")
							 )
						       | _             -> raise (Erreur "TOPOLOGY: only float/int in the definition of a probability (2)")
						 )
					       | StochPExp f -> (
						   let v = f (VAL_undef "") in
						     match v with
						       | VAL_float flo -> flo
						       | VAL_int i     -> float_of_int i
						       | VAL_funct f   -> (
							   match (f mgsval) with
							     | VAL_float flo -> flo
							     | VAL_int i     -> float_of_int i
							     | _             ->  raise (Erreur "TOPOLOGY: only float/int in the definition of a probability (3)")
							 )
						       | _             -> raise (Erreur "TOPOLOGY: only float/int in the definition of a probability (2)")
						 )					    )
				in
				  (*
				  if ((proba < 0.0) or (proba > 1.0))
				  then raise (Erreur (Printf.sprintf "TOPOLOGY: probabilities should be between 0 and 1(here : %f)" proba));*)
				  somme_proba := !somme_proba +. proba;
				  proba, rule
				    
			    | {rtype = {a_stochastic  = None }} -> raise (Erreur "TOPOLOGY: ALL rules must have a probability specified") 
			 ) rules 
    in
    let liste = List.map (function (p,r) -> (p /. !somme_proba, r)) liste in
      somme_proba := 1.0 ;
(*      if ((abs_float(1.0 -. !somme_proba)) > epsilon)
      then raise (Erreur (Printf.sprintf "TOPOLOGY: the sum of all probabilities should be around 1.0 (here : %f)" !somme_proba))
      else *)
	begin
	  let rnd = Random.float 1.0 
	  in
	    try
	      List.fold_right (fun (p, r) accu -> (let naccu = accu +. p
						   in
						     if (naccu >= rnd)
						     then (self#apply_one_rule false r;
							   raise (StochApplied true))

						     else naccu)) liste 0.0;
	      raise (StochApplied false)

	    with 
		StochApplied false -> raise (Erreur (Printf.sprintf "TOPOLOGY: internal error on stochastic. No rule applied while (%f, %f)." !somme_proba rnd))
	      | StochApplied true  -> self#build_result
	end


  method apply_gillespie rules = 
    let tbl_occurences = ref (Hashtbl.create 0) in
    let is_static_rule  =(try
			    tbl_occurences := self#all_occurences();
			    (match ((List.hd rules).rtype).a_gillespie_H
			     with

(* On fait une serieuse optimisation ici qui n'est pas toujours correcte :        *)
(* Si on a un motif en l.h.s. qui n'est compose que de constantes (et non pas     *)
(* de valeurs calculees) alors il suffit de calculer *une seule fois* la fonction *)
(* H_i. Par contre, si dans le motif en l.h.s. on a une expression qui change,    *)
(* par exemple par usage d'un motif filtrant un nombre *variable* d'argument tel  *)
(* que (x* as X / size(X) == f())  alors le calcul *statique* des H_i est faux.   *)
(* Il faudrait determiner *statiquement* si le motif est lui meme statique ou     *)
(* dynamique.                                                                     *)

				 Some _ -> ()        (* Ok, on a deja calcule les fonctions _H precedemment *)
			       | None   -> (let fun_h glsp_hash occ_tbl = 
					      (let v_h = Hashtbl.fold (fun spec nb s -> 
(*	Printf.printf "\nRule: (%s, %d) \n" (to_string(prsvalue spec)) nb; flush Pervasives.stdout; *)
									 (let occ_spec = (try
											    Hashtbl.find occ_tbl spec
											  with Not_found -> 0) in
(*      Printf.printf "occurence %d \n" occ_spec; flush Pervasives.stdout; *)
									  let cnp =
									    if (nb > occ_spec) then
									      Big_int.big_int_of_int 0
									    else
									      cnp_big_int 
										(Big_int.big_int_of_int occ_spec) 
										(Big_int.big_int_of_int nb)
									  in
									    Big_int.mult_big_int cnp s)) glsp_hash Big_int.unit_big_int
					       in
						 Big_int.float_of_big_int v_h)
					    in
					      List.iter (fun rule -> (
							   if (rule.rtype.a_gillespie_c <> GAiExp) then
							     let hash = (rule.motif)#required_species
							     in
							       rule.rtype.a_gillespie_hash <- Some hash;
							       rule.rtype.a_gillespie_H    <- Some (fun_h hash)
							   else
							     rule.rtype.a_gillespie_H <- Some (fun _ -> raise (Erreur "Topology: gillespie: trying a_gillespie_H application for a GAiExp rule type"))
							 ))  rules
					      ;
				 ));
			    true
			  with DynamicPattern -> false) in
    let min_tau  = ref infinity
    and min_rule = ref None
    in
      List.iter (function
(* C est un terme *)
		   | (({rtype = {a_gillespie_H  = (Some h);
				 a_gillespie_c  = GCiExp;
				 a_gillespie_t  = t;
				 a_gillespie_m  = f     }}) as rule) when is_static_rule ->
		       (let ci = f (VAL_undef "") in(*Compile.eval t in*)
			  match ci with
			    | VAL_float ci -> (
				let taui = (1.0 /. (ci *. (h !tbl_occurences))) *. (log (1.0 /. Random.float(1.0)))
				in
				  if (taui < !min_tau) then (min_tau  := taui;
							      min_rule := Some rule)
			      )
			    | VAL_int i -> (
				let ci = float_of_int i in
				let taui = (1.0 /. (ci *. (h !tbl_occurences))) *. (log (1.0 /. Random.float(1.0)))
				in
				  if (taui < !min_tau) then (min_tau  := taui;
							      min_rule := Some rule)
			      )
			    | VAL_funct f_c -> (
				let ci = (match (f_c mgsval)
					  with
					      VAL_float x -> x
					    | VAL_int   i -> float_of_int i
					    | _           -> raise (Erreur "Topology:Ci has to be either int or float (1)")) in
				let taui = (1.0 /. (ci *. (h !tbl_occurences))) *. (log (1.0 /. Random.float(1.0)))
				in
				  if (taui < !min_tau) then (min_tau  := taui;
							      min_rule := Some rule)
			      )
			    | _ -> (raise (Erreur "Topology: apply_gillespie: Ci must evaluate in a float, an integer or a function"))
		       )
(* A est un terme *)
		   | (({rtype = {a_gillespie_c  = GAiExp;
				 a_gillespie_t  = t;
				 a_gillespie_m  = f     }}) as rule) -> (
		       
		       let ct = f (VAL_undef "") in(*Compile.eval t in*)
			 match ct with
			   | VAL_funct f_ai -> (
			       let ai = (match (f_ai mgsval)
					 with
					     VAL_float x -> x
					   | VAL_int   i -> float_of_int i
					   | _           -> raise (Erreur "Topology:Ai has to be either int or float")) in
			       let taui = (1.0 /. ai) *. (log (1.0 /. Random.float(1.0)))
			       in
				 if (taui < !min_tau) then (min_tau  := taui;
							     min_rule := Some rule)
			     )
			   | _ -> (raise (Erreur "Topology: apply_gillespie: Ai must evaluate in a function"))
		     )
		   | _ -> raise (Erreur "TOPOLOGY:this kind of rule is not handled yet")
      ) rules;
      (try (
	 (* On fait avancer tau du tau_i correspondant a la regle *)
	 Environnement.set_varidx
	      Options.index_var_tau 
	      (match (Environnement.get_varidx(Options.index_var_tau))  with
		 | VAL_float f when (!min_tau >= 0.0) -> VAL_float (f +. !min_tau)
		 | VAL_float f when (!min_tau < 0.0) -> raise (Erreur (Printf.sprintf "Topology: apply_gillespie: negative tau %f" !min_tau))
		 | _                                  -> raise (Erreur "Topology: apply_gillespie: major faillure")) ;
	 
	 
	 (* On applique la rgle choisie *)
	 (match !min_rule with
	    | Some rule -> (
		if not (self#apply_one_rule false rule > 0)
		then (raise (Erreur ("Topology: apply_gillespie: application of a rule with no available occurence: "^((rule.motif)#to_full_string))))
	      )
	    | None      -> ()
	 );
	 self#build_result
       ) with BreakTransformation res -> res
      )

  (** On itere jusqu'a ce qu'on ne trouve plus d'element de libre **)
  method apply_one_rule one_or_maximum (rule : ('position pattern) compiled_rule) =
    let rule_name = rule.rname
    and motif = rule.motif
    and rule_type = rule.rtype
    and exp = rule.rhs
    and nbvar = rule.nbvar
    and nb_applied = ref 0 (* Une regle a-t-elle ete appliquee *)
    in
      begin
	assert begin if !trace_opt then 
	  (printf "-- topo: apply_one_rule: %s motif: %s\n" 
	     (match rule_name with None -> "(anonymous)" | Some s -> s) motif#to_full_string);
	  flush stdout;
	  true
	end;
	(** Printf.printf "Motif %s a une taille de %d et est resize a %d\n" motif#to_full_string self#length_env nbvar; **)
	self#resize_env nbvar;
	(try
	   while self#move_next_usable do
	     let tmppos = self#cpos 
	     in
	       assert 
		 begin 
		   if !trace_opt
		   then (printf "---- recherche d'une occurence du motif a partir de %s\n" (prpos tmppos));
		   true
		 end;
	       self#visit(tmppos);  (** cette position sert de depart a un chemin **)

	       match motif#filtre (self :> 'position generic_topo) 
	       with
		 | None -> (assert 
			      begin 
				if !trace_opt 
				then (printf "---- topo: apply_one_rule: failed motif %s\n" motif#to_full_string; flush stdout);
				true
			      end;
			    ())
		 | Some path -> (assert
				   begin 
				     if !trace_opt 
				     then (printf "---- topo: apply_one_rule: accept motif %s and add to batch\n" motif#to_full_string;
					   self#print_env ;
					   flush stdout);
				     true
				   end;
				 incr nb_applied;
				 let rhs = (pf1 exp) self#get_val_of_rule_variable
				 in 
				   self#add_match rule_type path rhs;
				   if (not one_or_maximum) then raise NoMoreFound)
	   done
	 with NoMoreFound -> ());
	self#unvisit_all;
	!nb_applied
      end

(* *)
  method apply_one_matching (cmotif : ('position pattern) compiled_motif) =
    let nbvar      = cmotif.cm_nbvar
    and motif      = cmotif.cm_motif
    and nb_applied = ref 0
    and ret        = ref []
    in
      begin
	assert begin if !trace_opt then 
	  (printf "-- topo: apply_one_matching motif: %s\n" motif#to_full_string);
	  flush stdout;
	  true
	end;
	(** Printf.printf "Motif %s a une taille de %d et est resize a %d\n" motif#to_full_string self#length_env nbvar; **)
	self#resize_env nbvar;
	(try
	   while (!nb_applied = 0) && (self#move_next_usable) 
	   do
	     let tmppos = self#cpos 
	     in
	       assert 
		 begin 
		   if !trace_opt
		   then (printf "---- recherche d'une occurence du motif a partir de %s\n" (prpos tmppos));
		   true
		 end;
	       self#visit(tmppos);  (** cette position sert de depart a un chemin **)

	       match motif#filtre (self :> 'position generic_topo)
	       with
		 | None -> (assert 
			      begin 
				if !trace_opt 
				then (printf "---- topo: apply_one_matching: failed motif %s\n" motif#to_full_string; flush stdout);
				true
			      end;
			    ())
		 | Some path -> (assert
				   begin 
				     if !trace_opt 
				     then (printf "---- topo: apply_one_matching: accept motif %s and add to batch\n" motif#to_full_string;
					   flush stdout);
				     true
				   end;
				 incr nb_applied;
				 let rhs = (pf1 (VAL_funct(fun _ -> VAL_seq(Array.make (List.length path) (VAL_undef ""), Type_seq.seq_desc))))
					     self#get_val_of_rule_variable
				 in
				   self#add_match (Terme.emtpy_rule_property ()) path rhs;
				 ret := path)
	   done
	 with NoMoreFound -> ());


	(* On serait tente de faire self#unvisit_all mais c'est inutile   *)
        (* car lors d'un \/, on reconstruit la topologie *a chaque fois*. *)
        (* Elle sera par consequent "fraiche" avec les flags corrects.    *)
	if (!nb_applied = 0) 
	then (
	  assert begin if !trace_opt then 
	    printf "-- topo: apply_one_matching motif: no path found\n";
	    flush stdout;
	    true
	  end;
	  None
	)
	else (
	  assert begin if !trace_opt then 
	    printf "-- topo: apply_one_matching motif: one path found\n";
	    flush stdout;
	    true
	  end;
	  Some (!ret)
	)

      end

(* *)


  method find : mgs_value -> 'position option = function _ ->
    raise (Erreur "Error: internal: cannot use method find on a non-set topology (to be changed !) 1")

  method findN : int -> mgs_value -> ('position list * 'position) option = function _ ->
    raise (Erreur "Error: internal: cannot use method findN on a non-set topology (to be changed !) 2")

  (* Cela a ete introduit pour Gillespie, ou on a besoin de *rapidement* trouver $n$       *)
  (* occurences d'une meme valeur ; il se peut que la topologie ne dispose pas d'une telle *)
  (* fonction, on utilise alors (dans patternN) un motif alternatif qui reecrit un motif   *)
  (* de la forme #3 `X en `X, `X, `X                                                       *)
  method dispose_get_NvalueS_from_pos = false

  method get_NposS_from_pos : 'position -> int -> mgs_value -> ((('position list)  * 'position) option) = fun _ _ _ -> 
    raise (Erreur "TOPOLOGY: method wget_NposS_from_pos not implemented");

  method occurence          : mgs_value -> int = fun _ ->
    raise (Erreur "TOPOLOGY: method occurence not implemented")

  method all_occurences     : unit -> species = fun _ -> 
    raise (Erreur "TOPOLOGY: method all_occurences not implemented")
end



(** patvar = la liste des index des variables qui nomment ce pattern
 ** is_elementary_path : le chemin est-il elementaire -longueur 1 mais different de P*- ou compose 
 **)


and virtual ['position] pattern (patvar : int list) (is_elementary_path : bool) =
object(self)

  method virtual to_string : string
  method print = printf "%s\n" (self#to_string); flush Pervasives.stdout
  method string_var = (List.fold_left (fun acc x -> acc^" "^(string_of_int x)) "^(" patvar)^")"
  method virtual to_full_string : string


  val need_matched_value = (patvar <> [])

  (** mthode qui met  jour *les* variables associes  ce pattern.
   ** On 'prcalcule' cette mthode en renvoyant des fct spcialises 
   ** pour 0 1, 2 et n variables  mettre  jour.
   **)
  method update_elem_variables =
    match patvar with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo v ->  topo#set_elt idx1 (List.hd v))
      | idx1::idx2::[] -> (fun topo v -> topo#set_elt idx1 (List.hd v); topo#set_elt idx2 (List.hd v))
      | _ ->  (fun topo v -> List.iter (fun idx -> topo#set_elt idx (List.hd v)) patvar)

  method update_path_variables =
    match patvar with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo v ->  topo#set_path idx1 v)
      | idx1::idx2::[] -> (fun topo v -> topo#set_path idx1 v; topo#set_path idx2 v)
      | _ ->  (fun topo v -> List.iter (fun idx -> topo#set_path idx v) patvar)

  method update_variables (topo : 'position generic_topo) v = 
    if is_elementary_path
    then self#update_elem_variables topo v
    else self#update_path_variables topo v

  (* lg minimal d'un chemin filtre *) 
  method virtual minimal_path_length : int

  (* la sous-classe end_pattern, cf. matching.ml sert de terminateur dans certaines 
   * expressions de patterns, et j'ai besoin de tester "1 coup en avance" si l'expression 
   * se termine 
   *)
  method is_end_pattern = false
 
  (* file : environnement -> topo -> path option 
   * le resultat : None = fail, Some p = OK with matched path pt and env is updated
   *)
  method virtual filtre : 'position generic_topo -> 'position return_path

  method required_species : ((mgs_value, int) Hashtbl.t) = raise DynamicPattern
(*    raise (Erreur "Topology: pattern: requiredspecies: only defined on cte_pattern1 and cte_patternN") *)

  method get_patvar = patvar
		 
end


and virtual  ['position] linked_pattern 
  (patvar : int list) 
  (is_elementary_path : bool)
  (next   : 'position pattern) =
object(self)
  inherit ['position] pattern patvar is_elementary_path
end
;;




(* === Les instances de topologies 
 * se retrouvent dans topo_monoidale.ml, topo_gbf.ml et topo_grf.ml
 * ================================================================= *)

