(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: transformation.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2008-06-03 12:59:19 $
 * Version: $Revision: 1.13 $
 * -------------------------------------------------------------------
 *
 * Status: Voir les commentaires de topology.ml, topo_monoidale.ml
 *         et matching.ml
 *          
 * Une partie droite de rgle est une expression (un terme) qui contient
 * des variables de pattern. L'ide est de transformer cette expression en 
 * une fonction qui prend en argument l'environnement $rule_env de liaisons des
 * variables de pattern et qui transforme toute occurrence de variable
 * de pattern X en un appel $rule_env(X).
 * L'environnement de liaison est simplement un array. Toute variable de pattern 
 * est associe a un index dans ce tableau. Cette association se fait par la 
 * fonction indexation.
 * Cette transformation de l'expression est ralise par la fonction 
 *     prepare_rhs_rule
 *
 * Une transformation est reprsente par une liste de quadruplets,
 * chaque quadruplet (name, lhs, rtype, rhs) representant une rgle :
 *     name =   lhs  ={rtype}=>  rhs
 * - name est une string option reprsentant un nom optionnel
 * - lhs est le motif code par un pattern_term
 * - rhs est la partie droite de la rgle code par un terme
 *
 * Cette liste de quadruplet est transform en une liste de record contenant 
 * les 4 champs prcdents plus le nombre de variables du pattern 
 * (fonction instancie_one_rule)
 * 
 * Un motif cod par un pattern_term est transform en un 'position
 * pattern qui dpend de la topologie de la collection qu'on veut 
 * transformer. Par exemple si la collection est une seq, on veut des
 * int pattern (car int est le type des positions d'une seq).
 * Cette traduction se fait par la fonction   instancie_pattern
 * (cette fonction est polymorphe par rapport a 'position).
 *
 * La fonction transform prend une transformation (sous la forme d'une
 * liste de quadruplets) et une collection et cre une instance
 * particulire de generic_topo correspondant a la topologie de la collection
 * et appelle la mthode transform avec l'instance positionn du pattern.
 * 
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Dispatch_sig;;
open Topology;;
open Topo_rec;;
open Topo_seq;;
open Topo_set;;
open Topo_ring;;
open Topo_bag;;
open Topo_gbf;;
open Topo_graph;;
open Topo_del;;
open Topo_prox;;
open Topo_qmf;;
open Topo_achain;;
open Pattern;;
open Instancie_pat;;
open Instancie_pat_ac;;

__register_version "$RCSfile: transformation.ml,v $" "$Revision: 1.13 $" "$Date: 2008-06-03 12:59:19 $";;

(** transforme une liste de regles dfinies par des quadruplets (cf. plus haut) contenant des pattern-term
 ** et des termes en une liste de record contenant des pattern-objets et des expressions compiles
 **)

type topoSorte = TopoSet | TopoBag | TopoGen ;;

let instancie_pattern nbvar varindex cont pat = function
	| TopoSet -> List.map (function p -> instancie_pattern_ac nbvar varindex cont p) (remove_OrPat [] pat)
	| TopoBag -> List.map (function p -> instancie_pattern_ac nbvar varindex cont p) (remove_OrPat [] pat)
	| _       -> [instancie_pattern_gen nbvar varindex cont pat]
;;


let instancie_rtype rtype =
  if (rtype.a_gillespie_c = GCiExp) || (rtype.a_gillespie_c = GAiExp)
  then rtype.a_gillespie_m <- to_funct(Compile.eval rtype.a_gillespie_t) ;
  
  (match rtype.a_stochastic with
    | Some (StochExp t) -> rtype.a_stochastic <- Some (StochPExp (to_funct (Compile.eval t)))
    | _                 -> ()
  );
  
  rtype
;;

let rec instancie_rules rules = function topoSort -> (Printf.printf "\n\n" ; List.flatten (List.map (instancie_one_rule topoSort) rules))

and instancie_one_rule topoSort (name, motif, rtype, exp) =
  let motif2 = namepattern motif in
  let pvar = pattern_var motif2 in 
  let varindex = indexation pvar in
  let nb_var = List.length pvar in
    begin
      assert begin if !trace_opt then 
	(Printf.printf "____ indexations des variables:\n";
	 List.iter (fun (id, idx) -> Printf.printf "\t%s -> %d\n" id idx) varindex);
	true
      end;

      List.map 
	(fun m -> 
	   { rname = name;
	     motif = m;
	     rtype = instancie_rtype { rtype with a_flat=rtype.a_flat };
	     rhs = compile_rhs_rule(prepare_rhs varindex exp);
	     nbvar = nb_var
	   }
	)
	(instancie_pattern nb_var varindex (new end_pattern) motif2 topoSort)
    end
(**
**)
;;



(* Dans cette fonction, si on n'utilise pas Obj.magic, l'instanciation des regles se specialise et rend
   les topologies incompatibles ; pourtant, le code des patterns est generique et rr devrait s'instancier
   dans chaque cas du match. C'est important que rr soit factorise, sinon a chaque application ou iteration
   de la transformation on recalcule l'instanciation des regles.

   Autre solution prise ici, remplacer   "let rr  = instancie_rules rules" par autant
   d'instances de rr que de type de 'position differentes (int et int array).
*)

let transform rules =
  let rr_int_gen        = instancie_rules rules TopoGen
  and rr_ncell_gen      = instancie_rules rules TopoGen
  and rr_splx_gen       = instancie_rules rules TopoGen
  and rr_intArray_gen   = instancie_rules rules TopoGen
  and rr_mgsvalue_gen   = instancie_rules rules TopoGen
  and rr_int_set        = instancie_rules rules TopoSet
  and rr_int_bag        = instancie_rules rules TopoBag
  in
    function coll -> 
      match coll with
	| VAL_rec c    -> (new topo_record   c coll)#transform rr_mgsvalue_gen
	| VAL_seq c    -> (new topo_sequence c coll)#transform rr_int_gen
	| VAL_ring c   -> (new topo_ring     c coll)#transform rr_int_gen
	| VAL_set c    -> (
	    if !opt_ac
	    then (new topo_set c coll (Core_set.cardinal (fst c)))#transform rr_int_set
	    else (new topo_set c coll (Core_set.cardinal (fst c)))#transform rr_int_gen
	  )
	| VAL_bag c    -> (
	    if !opt_bag_matching then
	      if !opt_ac
	      then (new optimized_topo_bag c coll (Core_bag.set_cardinal_b (fst c)))#transform rr_int_bag
	      else (new optimized_topo_bag c coll (Core_bag.set_cardinal_b (fst c)))#transform rr_int_gen
	    else
	      if !opt_ac
	      then (new topo_bag c coll (Core_bag.cardinal_b (fst c)))#transform rr_int_bag
	      else (new topo_bag c coll (Core_bag.cardinal_b (fst c)))#transform rr_int_gen
	  )
	| VAL_gbf c    -> (new topo_gbf c coll)#transform rr_intArray_gen
	| VAL_grf c    -> (new topo_grf c coll)#transform rr_int_gen
	| VAL_prox c   -> (new topo_prox c coll)#transform rr_int_gen
	| VAL_del c    -> (new topo_del c coll)#transform rr_int_gen
	| VAL_qmf c    -> (new topo_qmf c coll 0 1)#transform rr_ncell_gen
	| VAL_achain c -> (new topo_achain c coll 0 1)#transform rr_splx_gen
	| v            -> (let mesg = "Error: transformation: bad argument type "^(constructor_string v)
			   in prerr_endline mesg; print_string "\t"; prvalue v; 
			     prerr_endline "\nis not yet handled in a transformation (only aggregate types)"; 
			     raise (Erreur mesg))
;;

let transform2 vdim ndim rules =
  let rr_ncell_gen      = instancie_rules rules TopoGen
  and rr_splx_gen       = instancie_rules rules TopoGen
  in
    function coll -> 
      match coll with
	| VAL_qmf c    -> (new topo_qmf ((Core_qmf.extends_to_dim (fst c) vdim),snd c) coll vdim ndim)#transform rr_ncell_gen
	| VAL_achain c -> (new topo_achain c coll vdim ndim)#transform rr_splx_gen
	| v            -> (let mesg = "Error: transformation2: bad argument type "^(constructor_string v)
			   in prerr_endline mesg; prerr_string "\t"; prfvalue stderr v; 
			     prerr_endline "\nis not yet handled in a transformation2 (only aggregate types)"; 
			     raise (Erreur mesg))
;;
