(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: byte_code.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-09-23 21:24:23 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Le type  byte_code  represente les combinateurs utilises pour l'evaluation.
 * On utilise les mots combinateur ou byte_code indifferement.
 *
 *    Un terme                         (cf. terme.ml) 
 *    est traduit                      (cf. combinatorise.ml) 
 *    en un byte_code qui est optimise (cf. optimise_combinateur.ml) 
 *    puis evalue.                     (cf. compile.ml)
 *
 *    Le code de la fonction qui       (cf. kombinator.ml)
 *    realise l'evaluation d'un 
 *    combinateur est une valeur "mgs". (cf. valeur.ml)
 * 
 * Les fonctions de base (comme par exemple +), sont aussi codee
 * comme des fonctions "mgs" (par exemple cf. arith.ml) et rendu
 * accessible a l'utilisateur en les enregistrant (cf. kernel.ml) dans
 * l'environnement (cf. environement.ml).
 *
 * Les combinateurs defini dans kombinator.ml sont aussi
 * accessibles a l'utilisateur a travers un nom (par exemple 'I)
 *
 * Certains combinateurs, par exemple Cupl, ne correspondent pas a
 * une fonction mgs bien defini. Ils sont traduit au vol dans compile.ml.
 *
 *
 * Lors de la traduction d'un terme, le byte_code contient des termes "hybride". 
 * Le byte_code final ne contient plus d'identificateur correspondant
 * a des parametres de fonctions, ni d'abstraction. La fonction
 * check_byte_code permet de verifier que le byte_code est bien "final".
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;


__register_version "$RCSfile: byte_code.ml,v $" "$Revision: 1.1 $" "$Date: 2004-09-23 21:24:23 $";;



(* Un operateur parametre un Cupl (cf. type byte_code ci-dessous) qui
 * reprsente un type d'operation: conditionelle, sequencement,
 * application d'une fonction dsigne par son nom, affectation de
 * variable.
 *)
type operator =
  | Oif          (** liste-de-term (cf. constructeur Cupl) de lg 3   **)
  | Ofby         (** liste-de-term de lg quelconque                  **)
  | Ofct of int  (** int = adresse memoire, lg quelconque            **)
  | Oset of int  (** int = @memoire, lg = 1,                         **)


type byte_code =

  | C_I    (*  I x = x          *)
  | C_K    (*  K a x = a        *)

  | C_B    (*  B f a x = f(a x)            *)
  | C_C    (*  C f a x = (f x)(a)          *)    
  | C_S    (*  S f a x = (f(x))(a x)       *)

  | C_S4   (*  S4 f a b x = f (a x) (b x)  *)

  | Capp of byte_code * byte_code    (* application *)
  | Cabs of string * byte_code       (* abstraction temporaire ! *)
  | Cp of string        (* nom de parametre de fct  *)
  | Cd of string        (* une constante symbolique *)
  | Cv of mgs_value     (* une constante valeur     *)
 
  (* represente un tuple de chose, avec une indication du nbre
   * d'abstraction englobante et une operation de type operator qui
   * s'applique sur ce tuple. Exemple : 
   *  - une sequence de 3 instructions p1  p3: Cupl(Ofby, n, [p1;p2;p3])
   *  - un if(c,p,q) : Cupl(Oif, n, [c;p;q])
   *  - une variable : Cupl(Ofct(index), n, [])
   *  - une appli. de var  k arguments: Cupl(Ofct(index), n, [p1;...pk])
   *  - une affectation : Cupl(Oset(index), n, [val])
   *)
  | Cupl of operator * int * (byte_code list)
;;




(** On verifie qu'il ne reste plus de lambda 
 ** ni de parametres explicite dans le byte-code 
 **)

let rec check_byte_code = function
  | Cabs _       -> false
  | Capp(p, q)   -> (check_byte_code p) & (check_byte_code q)
  | Cupl(_,_, l) -> List.for_all check_byte_code l
  | Cp _         -> false
  | x            -> true
;;


(****

open Printf;;

let rec genere file name bc =
  if check_byte_code bc then gen (open_out file) name bc
  else raise (Erreur "Error: genere byte code in file: incompletely combinatorized byte code")

and gen out name bc = 
  genere_comb out;
  fprintf out "let %s = " name;
  genfct out bc;
  fprintf out ";;\n\nEnvironnement.addenv \"%s\" Terme.Impur %s;;\n\n" name name

and genere_comb out =  fprintf out "\n\nopen Kombinator;;\n"

and genfct out = function
  | C_I  -> fprintf out "mgs_combI"
  | C_K  -> fprintf out "mgs_combK"
  | C_B  -> fprintf out "mgs_combB"
  | C_C  -> fprintf out "mgs_combC"
  | C_S  -> fprintf out "mgs_combS"
  | C_S4 -> fprintf out "mgs_combS4"

  | Capp(p, q) -> fprintf out "(("; genfct out p; fprintf out ")("; genfct out q; fprintf out "))"

  | Cd name -> fprintf out "(!Environnement.var_env).(index_cte(%s))" name

  | Cv v -> fprintf out "("; Printer.prfvalue out v; fprintf out ")"
 
  | Cupl (Oif, _, [c; p; q]) ->
      fprintf out "(if "; genfct out c;
      fprintf out "then "; genfct out p;
      fprintf out "else "; genfct out q;
      fprintf out ")"

  | Cupl (Ofby, _, l) ->  fprintf out "("; genfby out l;  fprintf out ")"

  | Cupl (Ofct(n), _, l) -> 
      fprintf out "(((!Environnement.var_env).(%d)) " n; 
      List.iter (genfct out) l;  
      fprintf out ")"

  | _ -> raise (Erreur "Error: genfct: incompletely combinatorized byte code")


and genfby out = function
  | [] -> raise (Erreur "Error: internal error: genfby: empty sequence of instruction")
  | a::[] -> genfct out a
  | a::l -> genfct out a; fprintf out "; "; genfby out l
;;


*****)
