(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: printer_term.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-02-24 02:49:39 $
 * Version: $Revision: 1.5 $
 * -------------------------------------------------------------------
 *
 * Les fonctions d'impressions.
 *
 * Les fonctions d'impression disponibles:
 *
 *    prbxxx b v : print dans le buffer b de v: c'est la fonction de base a ecrire
 *    prfxxx f v : print dans le fichier f (f est un out-channel)
 *    prxxx  v   : print dans la sortie standard
 *    prsxxx v   : print dans une string qui est retournee
 *
 * xxx = val   : pour les mgs_value
 *       value : pour les mgs_value  (les 2 formes sont permises)
 *       term  : pour les termes
 *       code  : pour les byte_codes
 *
 * Si xxx = val ou value, 
 * on peut rajouter "unformatted_" ou juste "u" devant, ce qui  donne une impression non-formatte 
 * des valeurs. 
 * Exemple: uprsval ou bien unformatted_prsval ou bien uprsvalue ou bien unformatted_prsvalue
 * (ces 4 formes dsignent la mme fonction).
 *
 * On peut aussi rajouter un 2 derrire pour obtenir la fonction qui fait la mme 
 * chose mais qui retourne unit: prfval2.
 * Attention, il n'y a pas de prsval2 ou de prsvalue2, car pour cette fonction, il n'y a pas
 * d'effet de bord: une string est retourne.
 *
 * Pour rendre disponible les fonction prcdentes, il faut implmenter par type Xxx:
 *
 *  -  _uprbval_'outBuffer_xxx : impression non-formatte de xxx
 *
 *  -  _pfbval_'outBufferF_xxx : impression formatte de xxx
 *
 *=========================================================================
 *)

open Buffer;;
open Printf;;
open Format;;

open Global;;
open Valeur;;

open Dispatch_sig;;
open Environnement;;
open Printer;;
open Terme;;

__register_version "$RCSfile: printer_term.ml,v $" "$Revision: 1.5 $" "$Date: 2006-02-24 02:49:39 $";;


let valundef = Valeur.mgsundef
;;



(** Attention: les noms collisionnent dans Printf et Format.
 ** Aussi on n'ouvre pas ces modules mais on dfinit ici 2 alias
 **)
let uprintf = Printf.bprintf (** impression standard dans un buffer **)
and fprintf = Format.bprintf (** impression formatt dans un buffer **)
;;

let parray = Auxfct.parray
;;


(* --- Pour les termes ----------------------------------------- *)

let rec prbterm b = function
  | Cte v       -> unformatted_prbvalue2 b v
  | Id  s       -> uprintf b "%s" s
  | App(f, a)   -> uprintf b "(%a %a)" prbterm f prbterml a
  | Abs(s, t)   -> uprintf b "\\%s.%a" s prbterm t
  | If(u, v, w) -> uprintf b "if %a then %a else %a fi" prbterm u prbterm v prbterm w
  | Fby(t1, t2) -> uprintf b "%a; %a" prbterm t1 prbterm t2
  | Set(s, t)   -> uprintf b "%s := %a" s prbterm t
  | Nyi         -> uprintf b "<NYI>"
  | Cnil        -> uprintf b "<cNIL>"

and prbterml b = function 
  | [] -> ()
  | [a] -> uprintf b "%a" prbterm a      
  | a::l -> uprintf b "%a %a" prbterm a prbterml l
;;




let prterm x = 
  clear out_buffer;
  prbterm out_buffer x; 
  output_buffer stdout out_buffer; 
  x

(** prfterm out x: sortie de x sur le fichier out **)
and prfterm out x = 
  clear out_buffer;
  prbterm out_buffer x; 
  output_buffer out out_buffer; 
  x

(** prsterm x : retourne une string qui represente la valeur x **)
and prsterm x =
  clear out_buffer;
  prbterm out_buffer x; 
  VAL_string(contents out_buffer)
;;

(** idem que prfterm, mais renvoie unit plutot que l'argument **)
let prfterm2 out x = ignore(prfterm out x)
;;




(* --- Idem pour les pattern  ----------------------------------------- *)

open Type_rec;;

let fieldspec_to_string s = function
  | Present        -> s
  | Absent         -> "~"^s
  | WithType t     -> s^" : "^(to_string (prsvalue t))
  | WithValue v    -> s^" = "^(to_string (prsvalue v))
  | WithoutValue v -> s^" ~= "^(to_string (prsvalue v))
  | WithCstrt c    -> raise (Erreur "Type_rec: fieldspec_to_string: internal error: this case shouldn't happen")
;;

let string_of_iter_kind s = function
    FirstFound   -> s
  | MaximumFound -> s^s

let rec prbpat b = function
  | NotYetPat          -> uprintf b "<not yet implemented pattern>"
  | BlankPat           -> uprintf b "_"
  | CtePat(v, 1)       -> uprintf b "%a" unformatted_prbvalue2 v
  | CtePat(v, n)       -> uprintf b "%d %a" n unformatted_prbvalue2 v
  | AsPat(s, BlankPat) -> uprintf b "%s" s
  | AsPat(s, p)        -> uprintf b "%a as %s" prbpat p s
  | GuardedPat(p, exp) -> uprintf b "%a / %a" prbpat p prbterm exp
  | OrPat(p1, p2)      -> uprintf b "%a | %a" prbpat p1 prbpat p2
  | RecordPat l        -> prbfilt b l
  | IterPat(PatStar (AllDirections, FirstFound), p)   -> uprintf b "(%a)*" prbpat p 
  | IterPat(PatStar (AllDirections, MaximumFound), p) -> uprintf b "(%a)**" prbpat p 
  | IterPat(PatPlus (AllDirections, FirstFound), p)   -> uprintf b "(%a)+" prbpat p 
  | IterPat(PatPlus (AllDirections, MaximumFound), p) -> uprintf b "(%a)++" prbpat p 

  | IterPat(PatStar ((SomeDirections l), FirstFound), p) ->  
      uprintf b "(%a)" prbpat p;
      parray (uprintf b "%s")  (uprintf b "%a" unformatted_prbvalue2) "<|" ", " "|>*" l

  | IterPat(PatStar ((SomeDirections l), MaximumFound), p) ->  
      uprintf b "(%a)" prbpat p;
      parray (uprintf b "%s")  (uprintf b "%a" unformatted_prbvalue2) "<|" ", " "|>**" l

  | IterPat(PatPlus ((SomeDirections l), FirstFound),  p) ->  
      uprintf b "(%a)" prbpat p;
      parray (uprintf b "%s")  (uprintf b "%a" unformatted_prbvalue2) "<|" ", " "|>+" l

  | IterPat(PatPlus ((SomeDirections l), MaximumFound),  p) ->  
      uprintf b "(%a)" prbpat p;
      parray (uprintf b "%s")  (uprintf b "%a" unformatted_prbvalue2) "<|" ", " "|>++" l

  | CommaPat(AllDirections, dirname, p1, p2) -> 
      if dirname = None then uprintf b "%a, %a" prbpat p1 prbpat p2
      else uprintf b "%a, as %s (%a)" prbpat p1 (from_option dirname) prbpat p2

  | CommaPat(SomeDirections d, dirname, p1, p2) -> 
      if dirname = None 
      then (uprintf b "%a <|%a|> %a" 
	      prbpat p1 
	      (fun b l -> parray (uprintf b "%s") (uprintf b "%a" unformatted_prbvalue2) "" ", " "" l) d 
	      prbpat p2)
      else (uprintf b "%a <|%a|> as %s (%a)" 
	      prbpat p1  
	      (fun b l -> parray (uprintf b "%s") (uprintf b "%a" unformatted_prbvalue2) "" ", " "" l) d 
	      (from_option dirname) 
	      prbpat p2)

  | DownPat(p1, p2) -> uprintf b "%a \\/ %a" prbpat p1 prbpat p2
  | NotPat p        -> uprintf b "~ %a" prbpat p

and prbfilt b l = uprintf b "{ "; prbfilt2 b l; uprintf b " }" 
and prbfilt2 b = function
  | [] -> uprintf b " }"
  | [(None, (s, ft))] -> uprintf b "%s" (fieldspec_to_string s ft)
  | [(Some s1, (s2, ft))] ->  uprintf b "%s as %s" (fieldspec_to_string s2 ft) s1
  | (None, (s, ft))::l ->  uprintf b "%s, %a" (fieldspec_to_string s ft) prbfilt2 l 
  | (Some s1, (s2, ft))::l -> uprintf b "%s as %s, %a" (fieldspec_to_string s2 ft) s1 prbfilt2 l 
;;


let prpat x = 
  clear out_buffer;
  prbpat out_buffer x; 
  output_buffer stdout out_buffer; 
  x

(** prfpat out x: sortie de x sur le fichier out **)
and prfpat out x = 
  clear out_buffer;
  prbpat out_buffer x; 
  output_buffer out out_buffer; 
  x

(** prspat x : retourne une string qui represente la valeur x **)
and prspat x =
  clear out_buffer;
  prbpat out_buffer x; 
  VAL_string(contents out_buffer)
;;

(** idem que prfpat, mais renvoie unit plutot que l'argument **)
let prfpat2 out x = ignore(prfpat out x)
;;



(* --- Idem pour le byte_code ----------------------------------------- *)


(***

let rec prbcode b = function

  | C_I  -> uprintf b "I"
  | C_K  -> uprintf b "K"
  | C_B  -> uprintf b "B"
  | C_C  -> uprintf b "C"
  | C_S  -> uprintf b "S"
  | C_S4 -> uprintf b "S4"

  | Cp s -> uprintf b"#%s" s
  | Cd s -> uprintf b "%s" s
  | Cv v -> prbvalue b v; ()

  | Cabs(s, p) ->  uprintf b("(%s)[%a]") s prbcode p

  | Capp(p, q) ->  uprintf b "(%a %a)" prbcode p prbcode q

  | Cupl(Oset(idx), j, [a]) -> uprintf b "%s<%d> := %a"  (name_of_var idx) j prbcode a

  | Cupl(Ofct(idx), j, []) -> uprintf b "%s<%d>" (name_of_var idx) j

  | Cupl(Ofct(idx), j, l) -> uprintf b "%s<%d>@%a" (name_of_var idx) j (prlcode "(" ", " ")") l

  | Cupl(Ofby, level, l) -> prlcode ("FBY<"^(string_of_int level)^"> ")   " ; "   " YBF"  b  l

  | Cupl(Oif, level, [p;q;r]) -> 
      uprintf b "IF<%d> %a Then %a Else %a Fi" level prbcode p prbcode q prbcode r

  | Cupl(_, _, l) as t -> uprintf b "\nERROR: Cupl syntax: %a\n" (prlcode "[" "/" "]\n") l
      

and prlcode deb sep fin b l = 
  uprintf b "%s" deb;
  prlcode2 sep b l;
  uprintf b "%s" fin;

and prlcode2 sep b = function
  | [] -> ()
  | [a] -> uprintf b "%a" prbcode a
  | a::l -> uprintf b "%a%s%a" prbcode a sep (prlcode2 sep) l
;;


let prcode x = 
  clear out_buffer;
  prbcode out_buffer x; 
  output_buffer stdout out_buffer; 
  x

(** prfcode out x: sortie de x sur le fichier out **)
and prfcode out x = 
  clear out_buffer;
  prbcode out_buffer x; 
  output_buffer out out_buffer; 
  x

(** prscode x : retourne une string qui represente la valeur x **)
and prscode x =
  clear out_buffer;
  prbcode out_buffer x; 
  VAL_string(contents out_buffer)
;;

(** idem que prfcode, mais renvoie l'argument plutot que unit **)
let prfcode2 out x = (prfcode out x; ())
and error_c x = (print_string "\nError: "; prcode x; print_string "\n");;


***)





(** === Listing des liaisons de l'environnement ======================================== **)

let skip_variable_in_listing = [
  "'copyright";
  ""
];;


let rec sortvar l = List.sort (fun (x,_) (y,_) -> Pervasives.compare x y) l

and printentry out b (cle, (index, m)) =
  if (b = mode_is_pur(m)) & (cle <> "'copyright") then
    (Printf.fprintf out "\t%s %s "
       cle
       (if mode_is_pur(m) then "=" else ":=");
     prfvalue2 out (!var_env).(index);
     Printf.fprintf out "\n")

and plistvar out b l =  List.iter (printentry out b) l

and listing_var () = plistvar stdout false (sortvar (Hash.to_list var_tbl))
and listing_cte () = plistvar stdout true  (sortvar (Hash.to_list var_tbl))

and lst_var out = plistvar out false (sortvar (Hash.to_list var_tbl))
and lst_cte out = plistvar out true  (sortvar (Hash.to_list var_tbl))



and listing_bindings out =
  let l = sortvar (Hash.to_list var_tbl) in
    begin
      List.iter (print_tab_var out) l;
      Printf.fprintf out "\n\n\n\n(* --- end of listing ---- *)\n\n\n"
    end

and print_tab_var out (cle, (index, m)) = 
  if not (hiden_variable cle) 
  then
    begin
      Printf.fprintf out "\t%s %s %a\n"
	(pad_name 10 cle)
	(if mode_is_pur(m) then "==" else ":=")
	prfvalue2  (!var_env).(index)
    end

and list_var_in_seq () =
  let l = sortvar (Hash.to_list var_tbl) in
  let ll = List.map fst l in
  let lll = List.filter (function id -> not(hiden_variable id)) ll
  in  VAL_seq(Array.of_list (List.map (function id -> VAL_string id) lll), Type_seq.seq_desc)

and hiden_variable id = 
  (List.mem id skip_variable_in_listing) 
  or (id.[0] == '$')
  or (id.[0] == '%')
  or (id.[0] == '')
  or (id.[0] == '')
;;

