(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: terme.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-21 19:25:09 $
 * Version: $Revision: 1.9 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

open Global;;
open Valeur;;

__register_version "$RCSfile: terme.ml,v $" "$Revision: 1.9 $" "$Date: 2006-04-21 19:25:09 $";;


(** Syntaxe abstraite de MGS **)

type term = 
  | Cte of mgs_value           (* une constante (code par un wrapper autour de la valeur mgs *)
  | Id  of string              (* un identificateur *)
  | App of term * term list    (* une application a n argument *)
  | Abs of string * term       (* une abstraction a 1 argument *)
  | If  of term * term * term  (* la conditionelle *)
  | Fby of term * term         (* le sequencement de 2 expressions *)
  | Set of string * term       (* l'affectation a une variable globale *)
  | Nyi                        (* Not Yet Implemented *)
  | Cnil                       (* idem mais genere une valeur undef au lieu d'une erreur *)
;;

(** Quelques macro pour construire des termes.
 ** Bien d'autres macros sont dfinies dans Vm/macro.ml
 **)
let op_apply1 f x1       = App (Id f, [x1])
and op_apply2 f x1 x2    = App (Id f, [x1; x2])
and op_apply3 f x1 x2 x3 = App (Id f, [x1; x2; x3])

and fct_apply1 f x1       = App (f, [x1])
and fct_apply2 f x1 x2    = App (f, [x1; x2])
and fct_apply3 f x1 x2 x3 = App (f, [x1; x2; x3])
;;




module Ss = Set.Make(struct type t = string let compare=compare end);;

let empty_string_set = Ss.empty;;
let string_singleton e = Ss.add e empty_string_set;;

let rec foldterm ((cte, app, abs, id, fby, cond, update, nyi) as toujours) e = 
  let ff = foldterm toujours 
  in
    match e
    with
      | Cte v       -> cte v
      | Id s        -> id s
      | App(u, v)   -> app (ff u) (List.map ff v)
      | Abs(s, u)   -> abs s (ff u)
      | If(u, v, w) -> cond (ff u) (ff v) (ff w)
      |	Fby(p, q)   -> fby (ff p) (ff q)
      |	Set(s, p)   -> update (ff p)
      |	Nyi         -> nyi ()
      |	Cnil        -> nyi ()
;; 

let free_var_term =
  let cte _       = empty_string_set
  and app u v     = (Ss.union u (List.fold_left (fun p q -> Ss.union p q) empty_string_set v))
  and abs s u     = Ss.diff  u (string_singleton s)
  and id s        = string_singleton s
  and fby p q     = Ss.union p q
  and cond u v w  = Ss.union u (Ss.union v w)
  and update p    = p
  and nyi ()      = empty_string_set
  in
    foldterm (cte, app, abs, id, fby, cond, update, nyi)
;;


(** les variables de vars qui apparaissent comme variables libres de t **)
let appear_list vars t = 
  let varset = List.fold_left (fun p q -> Ss.add q p) empty_string_set vars
  in Ss.elements (Ss.inter varset (free_var_term t))
;;




(* --- Les PATTERNS ------------------------------------------------------- *)

type pat_iteration_kind = 
  | PatStar of direction_specification * iter_kind
  | PatPlus of direction_specification * iter_kind

and iter_kind = 
  | FirstFound 
  | MaximumFound

and pattern_term = 
  | NotYetPat
  | BlankPat
  | CtePat of mgs_value * int
  | AsPat of string * pattern_term
  | CommaPat of direction_specification * (string option) * pattern_term * pattern_term 
  | DownPat of pattern_term * pattern_term
  | GuardedPat of pattern_term * term
  | OrPat of pattern_term * pattern_term
  | RecordPat of fields_pattern_specification
  | IterPat of pat_iteration_kind * pattern_term
  | NotPat of pattern_term
;;


let rec fold_pattern ((notyetpat, blankpat, ctepat, aspat, commapat, downpat, guardedpat, orpat, recordpat, iterpat, notpat) as arg) e =
  let app = fold_pattern arg
  in
    match e with
      | NotYetPat               -> notyetpat
      | BlankPat                -> blankpat
      | CtePat (v, n)           -> ctepat v n
      | AsPat(s, p)             -> aspat s (app p)
      | CommaPat(d, dn, p1, p2) -> commapat d dn (app p1) (app p2)
      | DownPat (l, r)          -> downpat (app l) (app r)
      | GuardedPat(p, t)        -> guardedpat (app p) t
      | OrPat(p1, p2)           -> orpat (app p1) (app p2)
      | RecordPat l             -> recordpat l
      | IterPat (ki, p)         -> iterpat ki (app p)
      | NotPat p                -> notpat (app p)
;;	

let pattern_variables = 
  let notyetpat           = []
  and blankpat            = []
  and ctepat _ _          = []
  and aspat s p           = s::p
  and commapat d dn p1 p2 = (let vars = p1 @ p2 in if (dn = None) then vars else (from_option dn)::vars)
  and downpat l r         = l @ r
  and guardedpat p _      = p
  and orpat p1 p2         = p1 @ p2
  and recordpat l         = List.fold_left (fun acc -> function (None, _) -> acc | (Some id, _) -> id::acc) [] l
  and iterpat _ p         = p
  and notpat p            = p
  in
    fold_pattern (notyetpat, blankpat, ctepat, aspat, commapat, downpat, guardedpat, orpat, recordpat, iterpat, notpat)
;;


(** Transformation d'un pattern en structure linaire:
 ** on passe par une forme intermdiaire o on a donn un nom  chaque pattern 
 ** et o on a transform les gardes en actions.
 **)
 
type stringl = string list 
and named_pattern_term = 
  | PA_BlankPat of stringl
  | PA_CtePat of stringl * mgs_value * int * named_pattern_term
  | PA_CommaPat of stringl * direction_specification * (string option) * named_pattern_term * named_pattern_term
  | PA_DownPat of stringl * named_pattern_term * named_pattern_term
  | PA_GuardedPat of stringl * named_pattern_term * term
  | PA_OrPat of stringl * named_pattern_term * named_pattern_term
  | PA_RecordPat of stringl * fields_pattern_specification
  | PA_IterPat of stringl * pat_iteration_kind * named_pattern_term
  | PA_NotPat of stringl * named_pattern_term
;;

let rec fold_named_pattern ((blankpat, ctepat, commapat, downpat, guardedpat, orpat, recordpat, iterpat, notpat) as arg) e =
  let app = fold_named_pattern arg
  in
    match e
    with
      | PA_BlankPat s                 -> blankpat s
      | PA_CtePat(s, v, n, p)         -> ctepat s v n p
      | PA_CommaPat(s, d, dn, p1, p2) -> commapat s d dn (app p1) (app p2)
      | PA_DownPat(s, p1, p2)         -> downpat s (app p1) (app p2)
      | PA_GuardedPat(s, p, t)        -> guardedpat s (app p) t
      | PA_OrPat(s, p1, p2)           -> orpat s (app p1) (app p2)
      | PA_RecordPat(s, l)            -> recordpat s l
      | PA_IterPat (s, ki, p)         -> iterpat s ki (app p)
      | PA_NotPat (s, p)              -> notpat s (app p)
;;	


(* --- Les PATTERNS de PATCH ----------------------------------------------- *)

type patch_pattern  = (string * bool * (guards list)) list
and guards          = Test of term | Pred of string*term
and cguards         = CTest of mgs_value | CPred of string*mgs_value
and connecteur      = NoConnection | FaceOf | CoFaceOf | TfaceOf | TcoFaceOf
;;


(* --- Les REGLES ---------------------------------------------------------- *)


type rule_property = { 
  mutable a_flat           : type_flat;                           (* Insertion ou bien append du resultat de l'applicaton de la regle *)
  mutable a_priority       : int;                                 (* Priorite d'application de la regle                               *)
  mutable a_trace          : int;                                 (* Trace de sortie (et niveau)                                      *)
  mutable a_guard          : mgs_value option;                    (* Garde a verifier avant d'essayer d'appliquer une regle           *)
  mutable a_parall         : bool;                                (* A appliquer de facon non-exclusive d'une autre regle //          *)
  mutable a_abstract       : bool;                                (* Regle abstraite (remontee vers un niveau d'abstraction)          *)

(* ******************************************************************************** *)
(* La suite sert pour Gillespie.                                                    *)

  mutable a_gillespie_ai   : mgs_value -> mgs_value;    (* pour le calcul explicite de (1 / a_i)                            *)
  mutable a_gillespie_c    : a_gil_type;                (* Une fonction ou une constate ou une expression?                  *)
  mutable a_gillespie_ci   : float;                     (* Constante statique Ci de Gillespie sur une regle                 *)
  mutable a_gillespie_f    : mgs_value -> mgs_value;    (* Expression pour calculer dynamiquement une cte                   *)
  mutable a_gillespie_t    : term;                      (* Terme  compiler plus tard                                       *)
  mutable a_gillespie_m    : mgs_value -> mgs_value;    (* Terme pr-compil                                                *)
  mutable a_gillespie_hash : species option;            (* Table de hash contenant les espces filtres et leur occurence   *)
  mutable a_gillespie_H    : (species -> float) option; (* La fonction H pour Gillespie                                     *)

(* ******************************************************************************** *)
(* La suite sert pour le stochastique.                                              *)

  mutable a_stochastic     : stochastic_spec option     (* Une constante qui exprime la probabilite d'application, on une fonction *)
}

and a_gil_type = GNoParam | GAiExp | GCiExp


and species = ((mgs_value, int) Hashtbl.t) 

and path_type = rule_property         

	       
		       
and type_rule_property = 
    RP_flat of type_flat
  | RP_priority of int
  | RP_trace of (int option)
  | RP_on of mgs_value
  | RP_parall
  | RP_abstract
  | RP_gillespie of gillespie_spec
  | RP_probability of stochastic_spec

and gillespie_spec = 
  | GilCExp of term
  | GilAExp of term

and type_flat = FlatPath | NoFlatPath (* for backward compability ;) *)


and stochastic_spec = 
  | StochCte  of float
  | StochFun  of (mgs_value -> mgs_value)
  | StochExp  of term
  | StochPExp of (mgs_value -> mgs_value)

;;

type rule      = string option * pattern_term * rule_property * term;;
type rule_list = rule list;;  
    
let rec compile_one_rp v = function
  | RP_gillespie (GilCExp t) -> (v.a_gillespie_c <- GCiExp;
				 v.a_gillespie_t <- t)

  | RP_gillespie (GilAExp t) -> (v.a_gillespie_c <- GAiExp;
				 v.a_gillespie_t <- t)

  | RP_probability s         -> v.a_stochastic <- Some s

  | RP_on f                  -> v.a_guard <- Some f

  | RP_flat a          -> v.a_flat <- a
  |    _               -> raise (Erreur "Terme: compile_one_rp: Not yet REimplemented")

and emtpy_rule_property () = { a_flat           = FlatPath;
			       a_priority       = 0;
			       a_trace          = 0;
			       a_guard          = None;
			       a_parall         = false; 
			       a_abstract       = false; 
			       a_gillespie_ai   = (fun x -> x);
			       a_gillespie_c    = GNoParam;
			       a_gillespie_ci   = 0.0;
			       a_gillespie_f    = (fun x -> x);
			       a_gillespie_t    = Cte(VAL_funct (fun _ -> raise (Erreur "Terme: empty_rule_property: no term for a gillespie strategy application")));
			       a_gillespie_m    = (fun x -> x);
			       a_gillespie_hash = None;
			       a_gillespie_H    = None;
			       a_stochastic     = None;
			     }
and compile_rule_property l = 
  let v = emtpy_rule_property()
  in
    List.iter (compile_one_rp v) l;
    v
      
(* --- Les COMMANDES ------------------------------------------------------- *)

(* Les interactions avc le parseur. C'est ici car on a besoin du type term. *)

type commande = 
  | Carrowdef of string               (* une definition de fleche (propriete de regle)       *)
  | Ccoldef of mgs_type_def           (* une definition de collection                        *)
  | Ccoldefl of mgs_type_def list     (* une definition d'une liste de collections           *)
  | Cquit                             (* quitter tout                                        *)
  | Cnothing                          (* rien a faire                                        *)
  | Chelp                             (* affichage de l'aide                                 *)
  | Clegal                            (* affichage du copyright                              *)
  | Cliste                            (* listing des fonctions                               *)
  | Ctype of int                      (* listing des types: 1 type, 2 collection, 3 record   *)
  | Ccompil of string * string * term (* compile le byte code dans un fichier                *)
  | Cversion                          (* listing des versions des fichiers d'implementations *)
  | Csetparam of string * term        (* affectation d'un parametre global de l'interprete   *)
  | Cclear of string                  (* reset d'une variable                                *)
  | Cpositionprinter                  (* ordre de formattage pour l'ecrivain ?               *)
  | Ceval of term                     (* evaluation d'un terme                               *)
  | Ctrace of string list             (* mode debug sur la fct string                        *)
  | Cuntrace of string list           (* annule le mode debug                                *)
  | Cinclude of string                (* inclusion d'un fichier                              *)
  | Cgc                               (* affichage et declenchement du gc                    *)
  | Cexp of term                      (* obsolete                                            *)
  | Coutput of bool                   (* ??                                                  *)
;;



(** Une fonction impure utilise dans les macro-expansions: $delay 
 **
 ** delay est la fonction identit, mais elle est dclare impure.
 ** Cette fonction permet de "geler" une expression afin d'tre sur
 ** qu'elle ne sera value qu'au bon moment. En particulier, cela
 ** permet de traiter comme un terme impur, un terme qui ne contient
 ** pas explicitement de fct.  impure. Par exemple dans le terme
 ** App(Ct(VAL_funct( fct-caml )), [...])  on ne peut pas dterminer
 ** que la fct-caml fait un effet de bord ou pas: le terme est pur a
 ** priori. La fonction wrap_delay permet de forcer le caractre
 ** impure.  
 ** Cette fonction n'est pas destin  tre utilis par un programmeur
 ** MGS, c'est pourquoi son nom n'est pas parsable. Cette fonction est
 ** utilise dans les macro-expansions  travers wrap_delay.
 **)

let delay_until_run_time = VAL_funct(function x -> x)
;;
Environnement.addenv_impure "'$delay" delay_until_run_time
;;
let wrap_delay x = App(Id "'$delay", [x])
;;
