/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Les diffrents mouvements que supporte la mthode 'go'.
 */
typedef enum { Forward, Backward, Left, Right } TMovement;

/**
 * *** MTHODE  REVOIR ***
 *
 * Cette mthode permet de marquer ou dmarquer facilement une cellule
 * adjacente  celle donne par le couple (ALastDart, AOrbit).
 * Appele plusieurs fois (voir la mthode 'goAsFarAsPossible'), elle
 * permet de slectionner des brins le long d'un chemin.
 *
 * Par exemple, pour couper une sphre en deux le long d'un de ses
 * parallles, il faut pralablement slectionner les artes qui
 * correspondent  ce parallle.
 * Il s'agt donc de slectionner certaines artes d'un volume. Ces artes
 * correspondent  une orbite 02 (deuxime ligne du tableau). Le tableau
 * indique comment se font les dplacements. Dans ce cas prcis, pour
 * avancer on recherche alpha1210(ALastDart). Si on dsire tourner 
 * droite, c'est alpha120(ALastDart) qui est recherch, etc...
 * Aprs avoir tourn  gauche ou  droite, on voit donc que pour continuer
 * son trajet il faut repartir dans le sens contraire de celui dans lequel
 * on est arriv (si avant de tourner on a fait un go(Forward), il faut
 * continuer le trajet en faisant un go(Backward) et inversement).
 *
 * Si ASelect vaut 'vrai' et si un brin de l'orbite n'est pas marqu, alors
 * tous les brins de l'orbite sont marqus avec la marque AMarkNumber et
 * aucun dplacement n'est effectu (quelle que soit la direction donne).
 * Si ASelect vaut 'vrai' et si tous les brins de l'orbite sont marqus,
 * alors un dplacement est effectu et tous les brins de l'orbite donne
 * correspondant au brin atteint sont marqus.
 * Si ASelect vaut 'faux', tous les brins de l'orbite donne sont dmarqus
 * puis un dplacement est effectu. Le brin atteint est alors marqu.
 *
 * S'il n'y a pas de brin atteignable, la mthode retourne ALastDart.
 * Sinon, qu'il y ait dplacement ou pas, la mthode retourne le brin
 * atteint (s'il n'y a pas de dplacement ce brin est donc ALastDart).
 * 
 * Tableau rcapitulatif:
 *                          DIRECTION
 * ORBITE                   Forward        Backward       Left     Right
 * |                        |              |              |        |
 * | 0--3 (arte /face  )   | 1        0   |   0        1 |    010 | 010
 * | 0-2- (arte /volume)   | 1[2    1]0   |   0[1    2]1 | [2]010 | 010[2]
 * | 0-23 (arte /cc    )   | 1[2[32]1]0   |   0[1[23]2]1 |   [2]0 | 010[2]
 * | 01-- (face  /volume)   | 2      1 01  |  10 1      2 |     01 | 10
 * | 01-3 (face  /cc    )   | 2  [32]1 01  |  10 1[23]  2 |     01 | 10
 * | 012- (volume/cc    )   | 3 2    1 012 | 210 1    2 3 |    012 | 210
 *
 * Les autres orbites ne sont pas supportes (appeler la mthode avec une
 * d'entre elles provoque un rsultat imprvisible).
 * Les crochets [] indiquent les orbites qui peuvent tre ignores.
 * Par exemple, dans le cas de slection de faces dans une composante
 * connexe (orbite 013), si on effectue un go(Forward) alors que
 * alpha2(ALastDart) est 3-libre, c'est alpha2101(ALastDart) qui sera
 * atteint.
 * Par contre, si ALastDart est 2-libre, ou si alpha2(ALastDart) est
 * 1-libre, ou alpha21(ALastDart) est 0-libre, ou etc... la mthode
 * retourne ALastDart.
 *
 * @param ADirection La direction dans laquelle se fait le dplacement
 * @param ALastDart Le brin de dpart
 * @param AOrbit L'orbite indiquant quel type de dplacement effectuer
 * @param AMarkNumber Un numro de marque
 * @param ASelect Indique s'il faut slectionner ou dslectionner les brins
 * @return Le brin d'arrive
 *
 * @precondition L'orbite donne est mentionne dans le tableau rcapitulatif
 */
CDart* go(TMovement ADirection, CDart* ALastDart, TOrbit AOrbit,
	  int AMarkNumber, bool ASelect = true);

/**
 * Cette mthode permet de slectionner facilement plusieurs cellules
 * adjacentes de la carte.
 * Elle appelle la mthode 'go' autant de fois que possible, c'est--dire
 * jusqu' atteindre un brin dj rencontr ou un brin  partir duquel il
 * est impossible d'avancer.
 *
 * @param ADirection La direction dans laquelle se fait le dplacement
 * @param ALastDart Le brin de dpart
 * @param AOrbit L'orbite indiquant quel type de dplacement effectuer
 * @param AMarkNumber Un numro de marque
 * @param ASelect Indique s'il faut slectionner ou dslectionner les brins
 * @return Le brin d'arrive
 *
 * @precondition L'orbite donne est mentionne dans le tableau
 *   rcapitulatif de la mthode 'go'
 */
CDart* goAsFarAsPossible(TMovement ADirection,
			 CDart* ALastDart, TOrbit AOrbit,
			 int AMarkNumber, bool ASelect = true);

//******************************************************************************
