/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Teste si la demi-cellule de dimension ADim incidente  ADart
 * (ORBIT_INF[ADim]) peut tre extrude.
 * C'est le cas si tous les brins qui la constituent sont ADim-libres.
 *
 * Si AMarkToExtrude est positif ou nul, il indique les cellules qui
 * doivent tre extrudes en parallle. Ce paramtre est utilis par la
 * mthode 'canExtrudeByPath'.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AMarkToExtrude Un numro de marque
 * @return Un boolen indiquant si l'extrusion est possible
 */
bool canExtrudeHalfCell(CDart* ADart, int ADim, int AMarkToExtrude = -1);

/**
 * Teste si la cellule de dimension ADim incidente  ADart
 * (ORBIT_CELL[ADim]) peut tre extrude.
 * C'est le cas si tous les brins qui la constituent sont ADim+1-libres ou
 * si tous les brins qui la constituent sont ADim-libres.
 *
 * Si l'extrusion est possible et si au moins un brin de la cellule est
 * ADim-cousu, cela signifie que la cellule va tre duplique avant
 * l'extrusion. Par exemple pour extruder une face d'un volume, les brins
 * de la face tant 2-cousus il faut crer de nouveaux brins et les
 * 3-coudre  la face avant d'appliquer l'extrusion sur la demi-face
 * nouvellement cre.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @return Un boolen indiquant si l'extrusion est possible
 */
bool canExtrudeCell(CDart* ADart, int ADim);

/**
 * Donne la plus petite dimension D telle que la cellule dimension D
 * incidente  ADart puisse tre extrude.
 * Si l'extrusion est impossible, la mthode retourne la valeur 3.
 *
 * @param ADart Un brin de la carte
 * @return Une dimension d'extrusion (0, 1, 2 ou 3)
 */
int getExtrusionDimension(CDart* ADart);

/**
 * Extrude la cellule de dimension ADim incidente  ADart.
 *
 * Si AMarkExtruded est positif ou nul et si la mthode
 * 'getExtrusionWithAdacentSews' renvoie la valeur 'vrai', la cellule
 * extrude est cousue aux cellules adjacentes dj extrudes.
 * Si ADirectInfoIndex est positif ou nul, les cellules adjacentes sont
 * dtermines  partir du champ directInfo correspondant (ce champ
 * mmorise simplement l'image par alpha-ADim de chaque brin).
 * Ceci est particulirement utile pour la mthode 'extrudeByPath'
 * lorsque le chemin est ferm ou lorsque le brin initial est au milieu
 * du chemin. En effet, dans ce cas l les cellules sont isoles avant
 * extrusion d'o l'impossibilit de retrouver les cellules adjacentes
 * qui ont dj t extrudes (voir la mthode 'extrudeByPath' pour plus
 * d'informations).
 *
 * Ces cellules de dimension ADim+1 peuvent facilement tre repres
 * car les cellules de dimension ADim  partir desquelles on les alpha
 * obtenues sont marques avec la marque AMarkExtruded.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AExtrusionWithAdjacentSews Un boolen pour indiquer comment
 *   doit s'effectuer l'extrusion. Si vrai', les cellules adjacentes
 *   extrudes en parallle sont cousues entre elles, sinon elles ne
 *   sont pas cousues entre elles
 * @param AMarkExtruded Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un brin de la cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeCell(ADart, ADim)
 */
CDart* topoExtrude(CDart* ADart, int ADim, bool AExtrusionWithAdjacentSews,
		   int AMarkExtruded = -1, int ADirectInfoIndex = -1);

/**
 * Extrude la cellule de dimension ADim incidente  ADart en s'appuyant sur
 * la mthode 'topoExtrude'.
 *
 * Le paramtre AMarkToExtrude est un numro de marque indiquant quelles
 * cellules doivent tre extrudes. Ce paramtre est utile lors du
 * calcul des plongements.
 *
 * Pour la signification des paramtre AExtrusionWithAdjacentSews et
 * AMarkExtruded, se reporter  la mthode 'topoExtrude'.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param AExtrusionCoef Un TCoordinate fixant le coefficient d'extrusion.
 * @param AMarkToExtrude Un numro de marque
 * @param AMarkExtruded Un numro de marque
 * @return Un brin de la cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeCell(ADart, ADim)
 */
virtual CDart* extrudeByNormal(CDart* ADart, int ADim,
			       bool AExtrusionWithAdjacentSews,
			       TCoordinate AExtrusionCoef,
			       int AMarkToExtrude = -1,
			       int AMarkExtruded = -1);

/**
 * Extrude lorsque c'est possible les cellules de dimension ADim qui sont
 * marques avec la marque AMarkNumber.
 *
 * Cette mthode s'appuie sur la mthode protge
 * 'extrudeByNormalMarkedCells(int, int, int, bool)'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param AExtrusionCoef Un TCoordinate fixant le coefficient d'extrusion.
 * @return Le nombre de cellules extrudes
 */
int extrudeByNormalMarkedCells(int AMarkNumber, int ADim,
			       bool AExtrusionWithAdjacentSews,
			       TCoordinate AExtrusionCoef);

/**
 * Extrude lorsque c'est possible les cellules marques avec la marque
 * AMarkNumber. La mthode essaie d'abord d'extruder les sommets marqus,
 * puis les artes et en enfin les faces.
 *
 * @param AMarkNumber Un numro de marque
 * @param AExtrusionWithAdjacentSews Un boolen
 * @return Le nombre de cellules extrudes
 */
int intuitiveExtrudeByNormalMarkedCells(int AMarkNumber,
					bool AExtrusionWithAdjacentSews,
					TCoordinate AExtrusionCoef);

/**
 * Teste si l'extrusion de la cellule de dimension ADim incidente  ADart
 * selon le chemin donn par APath est possible.
 *
 * @param ADart Un brin de la carte incident  la cellule  extruder
 * @param ADim Une dimension (1 ou 2)
 * @param APath Un brin de la carte incident au chemin d'extrusion
 * @return Un boolen indiquant si l'extrusion est possible
 */
bool canExtrudeByPath(CDart* ADart, int ADim, CDart* APath);

/**
 * Extrude la cellule de dimension ADim incidente  ADart selon le chemin
 * donn par APath (orbite O1)  en s'appuyant sur la mthode 'topoExtrude'
 * appele plusieurs fois.
 *
 * Le paramtre ADirectInfoIndex est un indice de champ directInfo dans
 * lequel sont mmorises les ADim-coutures des brins avant extrusion.
 * Ce paramtre est utile pour la couture des cellules extrudes adjacentes
 * si le chemin est ferm ou si le brin initial est au milieu du chemin.
 * En effet, dans ce cas l la cellule extrude se retrouve isole et
 * lorsque ses ex-voisines sont  leur tour extrudes, on ne peut pas
 * dterminer  quelle cellules les cellules issues de l'extrusion doivent
 * tre cousues.
 *
 * Pour la signification des paramtre AExtrusionWithAdjacentSews et
 * AMarkExtruded, se reporter  la mthode 'topoExtrude'.
 *
 * @param ADart Un brin de la carte incident  la cellule  extruder
 * @param ADim Une dimension (1 ou 2)
 * @param APath Un brin de la carte incident au chemin d'extrusion
 * @param AExtrusionWithAdjacentSews Un boolen 
 * @param AMarkExtruded Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un brin de la dernire cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeByPath(ADart, ADim, APath)
 */
virtual CDart* extrudeByPath(CDart* ADart, int ADim, CDart* APath,
			     bool AExtrusionWithAdjacentSews,
			     int AMarkExtruded = -1,
			     int ADirectInfoIndex = -1);

/**
 * Extrude lorsque c'est possible selon le chemin donn les cellules de
 * dimension ADim qui sont marques avec la marque AMarkNumber.
 *
 * Le rsultat dpend obtenu de la valeur que donne la mthode
 * 'getExtrusionWithAdjacentSews'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (1 ou 2)
 * @param APath Un brin de la carte incident au chemin d'extrusion
 * @param AMarkExtruded Un numro de marque
 * @return Le nombre de cellules extrudes selon le chemin donn
 */
virtual int extrudeByPathMarkedCells(int AMarkNumber, int ADim,
				     CDart* APath,
				     bool AExtrusionWithAdjacentSews);

/**
 * Teste si l'extrusion de la cellule de dimension ADim incidente  ADart
 * autour d'un axe est possible.
 *
 * @param ADart Un brin de la carte incident  la cellule  extruder
 * @param ADim Une dimension (1 ou 2)
 * @param AClosed Un boolen indiquant si la rvolution est complte
 *   (angle de 360)
 * @return Un boolen indiquant si l'extrusion est possible
 */
bool canExtrudeByRevolution(CDart* ADart, int ADim, bool AClosed);

/**
 * Cre un chemin d'extrusion compos de ANbEdges artes, ouvert ou ferm
 * selon la valeur de AClosed. Ce chemin est utilis pour simuler
 * l'extrusion par rvolution autour d'un axe.
 *
 * @param AClosed Un boolen indiquant si la rvolution est complte
 *   (angle de 360)
 * @param ANbEdges Le nombre d'artes du chemin
 * @return Un brin extrmit du chemin
 */
virtual CDart* createRevolutionPath(bool AClosed, int ANbEdges);

/**
 * Dtruit un chemin de rvolution cr avec la mthode 'createRevolutionPath'.
 *
 * @param ADart Un brin du chemin de rvolution
 *
 * @precondition isIsolatedPolyline(ADart)
 */
void destroyRevolutionPath(CDart* ADart);

/**
 * Extrude la cellule de dimension ADim incidente  ADart selon autour d'un
 * axe. La mthode cre un chemin constitu de ANbEdges artes qui est
 * ferm ou ouvert selon la valeur de AClosed puis appelle la mthode
 * 'extrudeByPath'.
 *
 * Pour la signification des paramtres AMarkExtruded et ADirectInfoIndex,
 * se reporter  la mthode 'extrudeByPath'.
 *
 * @param ADart Un brin de la carte incident  la cellule  extruder
 * @param ADim Une dimension (1 ou 2)
 * @param AClosed Un boolen indiquant si la rvolution est complte
 *   (angle de 360)
 * @param ANbEdges Le nombre de sections  crer
 * @param AExtrusionWithAdjacentSews Un boolen 
 * @param AMarkExtruded Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un brin de la dernire cellule cre, situ  l'oppos de ADart
 *
 * @precondition canExtrudeByRevolution(ADart, ADim, AClosed)
 * @precondition ANbEdges >= 3
 */
virtual CDart* extrudeByRevolution(CDart* ADart, int ADim,
				   bool AClosed, int ANbEdges,
				   bool AExtrusionWithAdjacentSews,
				   int AMarkExtruded = -1,
				   int ADirectInfoIndex = -1);

/**
 * Extrude par rvolution lorsque c'est possible les cellules de
 * dimension ADim qui sont marques avec la marque AMarkNumber.
 *
 * Le rsultat dpend obtenu de la valeur que donne la mthode
 * 'getExtrusionWithAdjacentSews'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (1 ou 2)
 * @param AClosed Un boolen indiquant si la rvolution est complte
 *   (angle de 360)
 * @param ANbEdges Le nombre de sections  crer
 * @param AMarkExtruded Un numro de marque
 * @return Le nombre de cellules extrudes
 */
virtual int extrudeByRevolutionMarkedCells(int AMarkNumber, int ADim,
					   bool AClosed, int ANbEdges,
					   bool AExtrusionWithAdjacentSews);

protected:

/**
 * Marque entirement les cellules de dimension ADim qui doivent tre
 * extrudes.
 * Cette mthode ne fait rien d'autre qu'appeler 'markIncidentCells', mais
 * elle peut tre surcharge par les classes qui drivent de CGMapGeneric
 * pour ne marquer que la moiti des cellules  extruder, ceci afin de
 * rsoudre les problmes d'orientation de l'extrusion dans le cadre de
 * l'extrusion selon un vecteur normal.
 *
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AMarkNumberSrce Un numro de marque source
 * @param AMarkNumberDest Un numro de marque destination
 */
virtual void markCellsToBeExtrudedByNormal(int ADim,
					   int AMarkNumberSrce,
					   int AMarkNumberDest);

/**
 * Extrude lorsque c'est possible les cellules de dimension ADim qui sont
 * marques avec la marque AMarkNumber.
 *
 * Le rsultat obtenu dpend de la valeur que donne la mthode
 * 'getExtrusionWithAdjacentSews'.
 *
 * Si AMarkTreated est positif ou nul, les cellules extrudes sont marques
 * comme telles avec la marque AMarkTreated. Ceci permet de ne pas extruder
 * plusieurs fois la mme cellule dans des dimensions diffrentes (voir
 * mthode 'intuitiveExtrudeMarkedCells').
 *
 * Remarque: Comme CGMapGeneric ne traite pas les plongements, aucune
 * vrification sur les brins slectionns n'est effectue.
 * Par exemple, en dimension 2 si un brin d'une face est slectionn et si
 * sont alpha0 l'est aussi, on ne peut pas effectuer le calcul des
 * plongements car il y alpha ambigut sur le sens de la normale de la
 * face  extruder. Cependant, topologiquement il n'y alpha aucun problme.
 * Cette mthode devra donc tre surcharge dans les classes drivant de
 * CGMapGeneric.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AMarkTreated Un numro de marque
 * @param AExtrusionWithAdjacentSews Un boolen
 * @param AExtrusionCoef Un TCoordinate fixant le coefficient d'extrusion.
 * @return Le nombre de cellules extrudes
 */
int extrudeByNormalMarkedCells(int AMarkNumber, int ADim, int AMarkTreated,
			       bool AExtrusionWithAdjacentSews,
			       TCoordinate AExtrusionCoef);

//******************************************************************************
