/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Applique la matrice  tous les sommets appartenant  l'orbite AOrbit
 * incidente au brin ADart.
 * Un sommet est modifi si au moins un brin de son orbite sommet appartient
 *  l'orbite AOrbit incidente  ADart.
 * Si ADirectInfoVertex est positif ou nul, la matrice de transformation est
 * applique sur le sommet stock dans le champ directInfo correspondant du
 * brin, et le rsultat est plac dans l'attribut sommet (stock dans la
 * liste d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, la matrice de transformation
 * modifie directement l'attribut sommet.
 *
 * @param AMatrix Une matrice de transformation
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 */
void applyMatrix(const CTransformationMatrix & AMatrix, CDart* ADart,
		 TOrbit AOrbit, int ADirectInfoVertex = -1);

/**
 * Applique la matrice  tous les sommets marqus avec la marque AMarkNumber.
 * Un sommet est modifi si au moins un brin de son orbite sommet est marqu.
 * Si ADirectInfoVertex est positif ou nul, la matrice de transformation est
 * applique sur le sommet stock dans le champ directInfo correspondant du
 * brin, et le rsultat est plac dans l'attribut sommet (stock dans la
 * liste d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, la matrice de transformation
 * modifie directement l'attribut sommet.
 *
 * @param AMatrix Une matrice de transformation
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer la transformation
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 */
void applyMatrix(const CTransformationMatrix & AMatrix, int AMarkNumber,
		 int ADirectInfoVertex = -1);

/**
 * Pour chaque sommet dont un brin au moins est marqu avec la marque
 * AMarkNumber, calcule la distance euclidienne entre le plongement de ce
 * sommet et le sommet AVertex puis place le rsultat dans le champ
 * directInfo[ADirectInfoIndex] du brin possdant le plongement sommet.
 *
 * Avant d'appeler cette mthode, il convient de rserver de la mmoire
 * pour stocker les rsultats. Cette rservation se fait avec la mthode
 * 'CGMapGeneric::allocMemoryToDirectInfo' avec ABytes = sizeof(TCoordinate) et
 * AOrbitUsed = ORBIT_VERTEX.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoIndex Un indice indiquant quel directInfo utiliser
 * @param AVertex Un sommet quelconque de l'espace
 */
void computeDistanceToVertex(int AMarkNumber, int ADirectInfoIndex,
			     const CVertex & AVertex);

/**
 * Pour chaque sommet dont un brin au moins est marqu avec la marque
 * AMarkNumber, calcule la distance euclidienne entre le plongement de ce
 * sommet et la droite donne par le sommet ALineVertex et le vecteur
 * directeur ALineDirection puis place le rsultat dans le champ
 * directInfo[ADirectInfoIndex] du brin possdant le plongement sommet.
 *
 * Avant d'appeler cette mthode, il convient de rserver de la mmoire
 * pour stocker les rsultats. Cette rservation se fait avec la mthode
 * 'CGMapGeneric::allocMemoryToDirectInfo' avec ABytes = sizeof(TCoordinate) et
 * AOrbitUsed = ORBIT_VERTEX.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoIndex Un indice indiquant quel directInfo utiliser
 * @param ALineVertex Un sommet quelconque de la droite
 * @param ALineDirection Un vecteur directeur de la droite
 *
 * @precondition ! ALineDirection.isNull()
 */
void computeDistanceToLine(int AMarkNumber, int ADirectInfoIndex,
			   const CVertex & ALineVertex,
			   const CVertex & ALineDirection);

/**
 * Pour chaque sommet dont un brin au moins est marqu avec la marque
 * AMarkNumber, calcule la distance euclidienne entre le plongement de ce
 * sommet et le plan donn par le sommet APlaneVertex et le vecteur
 * normal APlaneNormal puis place le rsultat dans le champ
 * directInfo[ADirectInfoIndex] du brin possdant le plongement sommet.
 *
 * Avant d'appeler cette mthode, il convient de rserver de la mmoire
 * pour stocker les rsultats. Cette rservation se fait avec la mthode
 * 'CGMapGeneric::allocMemoryToDirectInfo' avec ABytes = sizeof(TCoordinate) et
 * AOrbitUsed = ORBIT_VERTEX.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoIndex Un indice indiquant quel directInfo utiliser
 * @param APlaneVertex Un sommet quelconque du plan
 * @param APlaneNormal Un vecteur normal du plan
 *
 * @precondition ! APlaneNormal.isNull()
 */
void computeDistanceToPlane(int AMarkNumber, int ADirectInfoIndex,
			    const CVertex & APlaneVertex,
			    const CVertex & APlaneNormal);

/**
 * Rpartit sur le segment [AMin,AMax] par changement de repre 1d les
 * valeurs (de type TCoordinate) pointes par le champ
 * directInfo[ADirectInfoIndex] des brins porteurs d'un attribut sommet.
 * Les sommets pris en compte sont ceux dont un brin au moins est marqu avec
 * la marque AMarkNumber.
 * Remarque: AMin peut tre suprieur  AMax.
 *
 * @param AMarkNumber Un numro de marque indiquant les sommets slectionns
 * @param ADirectInfoIndex Un indice indiquant o sont stockes les valeurs
 * @param AMin La borne infrieure du segment de normalisation
 * @param AMax La borne suprieure du segment de normalisation
 */
void normalizeParameter(int AMarkNumber, int ADirectInfoIndex,
			TCoordinate AMin, TCoordinate AMax);

/**
 * Applique une fonction sur les valeurs (de type TCoordinate) pointes par le
 * champ directInfo[ADirectInfoIndex] des brins porteurs d'un attribut sommet.
 * Les sommets pris en compte sont ceux dont un brin au moins est marqu avec
 * la marque AMarkNumber.
 * Les diffrents types de fonctions sont rpertoris dans le fichier
 * "math-extension.hh".
 *
 * @param AMarkNumber Un numro de marque indiquant les sommets slectionns
 * @param ADirectInfoIndex Un indice indiquant o sont stockes les valeurs
 * @param AFunctionType Le type de fonction  appliquer
 */
void applyFunctionOnParameter(int AMarkNumber, int ADirectInfoIndex,
			      TFunctionType AFunctionType);

/**
 * Applique une translation sur tous les sommets marqus avec la marque
 * AMarkNumber.
 * Un sommet est modifi si au moins un brin de son orbite sommet est marqu.
 *
 * Si ADirectInfoVertex est positif ou nul, la translation est applique sur
 * le sommet stock dans le champ directInfo correspondant du brin, et le
 * rsultat est plac dans l'attribut sommet (stock dans la liste
 * d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, l'attribut sommet est
 * directement modifi.
 *
 * Si ADirectInfoParam est positif ou nul, la translation est pondre avec
 * la valeur stocke dans le champ directInfo[ADirectInfoParam].
 *
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer la translation
 * @param AVector Le vecteur de translation
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 * @param ADirectInfoParam Indique o rcuprer le paramtre de pondration
 */
void translate(int AMarkNumber, const CVertex & AVector,
	       int ADirectInfoVertex = -1, int ADirectInfoParam = -1);

/**
 * Applique une rotation sur tous les sommets marqus avec la marque
 * AMarkNumber.
 * Un sommet est modifi si au moins un brin de son orbite sommet est marqu.
 *
 * Si ADirectInfoVertex est positif ou nul, la rotation est applique sur
 * le sommet stock dans le champ directInfo correspondant du brin, et le
 * rsultat est plac dans l'attribut sommet (stock dans la liste
 * d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, l'attribut sommet est
 * directement modifi.
 *
 * Si ADirectInfoParam est positif ou nul, l'angle de rotation est pondr
 * avec la valeur stocke dans le champ directInfo[ADirectInfoParam].
 *
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer la rotation
 * @param AAxeVertex Un sommet de l'axe de rotation
 * @param AAxeDirection Un vecteur directeur de l'axe de rotation
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 * @param ADirectInfoParam Indique o rcuprer le paramtre de pondration
 *
 * @precondition !ALineDirection.isNull()
 */
void rotate(int AMarkNumber,
	    const CVertex & AAxeVertex, const CVertex & AAxeDirection,
	    TCoordinate AAngle,
	    int ADirectInfoVertex = -1, int ADirectInfoParam = -1);

/**
 * Applique un tirement sur tous les sommets marqus avec la marque
 * AMarkNumber.
 * Un sommet est modifi si au moins un brin de son orbite sommet est marqu.
 *
 * Si ADirectInfoVertex est positif ou nul, l'tirement est appliqu sur
 * le sommet stock dans le champ directInfo correspondant du brin, et le
 * rsultat est plac dans l'attribut sommet (stock dans la liste
 * d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, l'attribut sommet est
 * directement modifi.
 *
 * Si ADirectInfoParam est positif ou nul, l'tirement est pondr avec
 * la valeur stocke dans le champ directInfo[ADirectInfoParam].
 *
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer l'tirement
 * @param ACenter Le centre d'tirement
 * @param ACoef Les coefficients d'tirement
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 * @param ADirectInfoParam Indique o rcuprer le paramtre de pondration
 */
void scale(int AMarkNumber,
	   const CVertex & ACenter, const CVertex & ACoef,
	   int ADirectInfoVertex = -1, int ADirectInfoParam = -1);

/**
 * Applique un tirement sur tous les sommets marqus avec la marque
 * AMarkNumber.
 * Pour plus d'informations, voir la mthode 'scale' ci-dessus.
 *
 * Remarque: La diffrence avec la mthode prcdente rside dans le
 * coefficient d'homothtie. Ici ce coefficient est un nombre rel, ce qui
 * correspond  une homothtie avec des coefficients identiques dans les
 * trois dimensions.
 *
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer l'tirement
 * @param ACenter Le centre d'tirement
 * @param ACoef Le coefficient d'tirement
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 * @param ADirectInfoParam Indique o rcuprer le paramtre de pondration
 */
void scale(int AMarkNumber,
	   const CVertex & ACenter, TCoordinate ACoef,
	   int ADirectInfoVertex = -1, int ADirectInfoParam = -1);

/**
 * Applique une homothtie de coefficient ACoef selon l'axe dtermin par
 * le sommet AAxeVertex et le vecteur AAxeDirection sur tous les sommets
 * marqus avec la marque AMarkNumber.
 *
 * Si ADirectInfoVertex est positif ou nul, l'homothtie est applique sur
 * le sommet stock dans le champ directInfo correspondant du brin, et le
 * rsultat est plac dans l'attribut sommet (stock dans la liste
 * d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, l'attribut sommet est
 * directement modifi.
 *
 * Si ADirectInfoParam est positif ou nul, l'tirement est pondr avec
 * la valeur stocke dans le champ directInfo[ADirectInfoParam].
 *
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer l'homothtie
 * @param AAxeVertex Un sommet de l'axe d'homothtie
 * @param AAxeDirection Un vecteur direction de l'axe d'homothtie
 * @param ACoef Le coefficient d'homothtie
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 * @param ADirectInfoParam Indique o rcuprer le paramtre de pondration
 *
 * @precondition !AAxeDirection.isNull()
 */
void axialScale(int AMarkNumber,
		const CVertex & AAxeVertex, const CVertex & AAxeDirection,
		TCoordinate ACoef,
		int ADirectInfoVertex = -1, int ADirectInfoParam = -1);

/**
 * Applique une homothtie de coefficient ACoef selon le plan dtermin par
 * le sommet APlaneVertex et le vecteur APlaneNormal sur tous les sommets
 * marqus avec la marque AMarkNumber.
 *
 * Si ADirectInfoVertex est positif ou nul, l'homothtie est applique sur
 * le sommet stock dans le champ directInfo correspondant du brin, et le
 * rsultat est plac dans l'attribut sommet (stock dans la liste
 * d'embeddings).
 * Si ADirectInfoVertex est strictement ngatif, l'attribut sommet est
 * directement modifi.
 *
 * Si ADirectInfoParam est positif ou nul, l'tirement est pondr avec
 * la valeur stocke dans le champ directInfo[ADirectInfoParam].
 *
 * @param AMarkNumber Un numro de marque permettant de reconnatre les
 *   sommets sur lesquels il faut appliquer l'homothtie
 * @param APlaneVertex Un sommet du plan d'homothtie
 * @param APlaneNormal Un vecteur normal du plan d'homothtie
 * @param ACoef Le coefficient d'homothtie
 * @param ADirectInfoVertex Indique o rcuprer les coordonnes initiales
 * @param ADirectInfoParam Indique o rcuprer le paramtre de pondration
 *
 * @precondition ! APlaneNormal.isNull()
 */
void planarScale(int AMarkNumber,
		 const CVertex & APlaneVertex, const CVertex & APlaneNormal,
		 TCoordinate ACoef,
		 int ADirectInfoVertex = -1, int ADirectInfoParam = -1);

/**
 * Projette orthogonalement les sommets dont un brin au moins est marqu
 * sur le plan d'quation AA * x + AB * y + AC * z + AD = 0.
 *
 * @param AMarkNumber Un numro de marque
 * @param A? Les paramtres de l'quation cartsienne du plan
 */
void orthoProjectOnPlane(int AMarkNumber,
			 TCoordinate AA,
			 TCoordinate AB,
			 TCoordinate AC,
			 TCoordinate AD);

//******************************************************************************
