/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Calcule le barycentre d'un ensemble de sommets dtermin par un brin ADart
 * et une orbite AOrbit.
 * Les brins marqus ne sont pas dmarqus  la fin du parcours.
 *
 * La mthode utilise un itrateur qui marque les brins avec les marques
 * ATreatedMark et AOrbitMark et qui n'effectue pas de dmarquage  la fin.
 *
 * Si ADirectInfoVertex est positif ou nul, le plongement des sommets n'est
 * pas recherch  l'aide de la mthode 'findVertex' mais est accd
 * directement  l'aide du champ directInfo[ADirectInfoVertex].
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param AOrbitMark Le numro de marque utilis par l'itrateur pour le
 *   marquage des brins atteints
 * @param ATreatedMark Le numro de marque utilis pour le marquage des brins
 *   dj traits
 * @param ADirectInfoVertex Un indice correspondant  un champ directInfo
 *
 * @precondition AOrbit==ORBIT012 || AOrbit==ORBIT123 || AOrbit==ORBIT0123
 */
CVertex basicBarycenter(CDart* ADart, TOrbit AOrbit,
			int AOrbitMark, int ATreatedMark,
			int ADirectInfoVertex = -1);

/**
 * Calcule le barycentre d'un ensemble de sommets dtermin par un brin ADart
 * et une orbite AOrbit.
 *
 * Si ADirectInfoVertex est positif ou nul, le plongement des sommets n'est
 * pas recherch  l'aide de la mthode 'findVertex' mais est accd
 * directement  l'aide du champ directInfo[ADirectInfoVertex].
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param ADirectInfoVertex Un indice correspondant  un champ directInfo
 */
CVertex barycenter(CDart* ADart, TOrbit AOrbit, int ADirectInfoVertex = -1);

/**
 * Calcule le barycentre d'un ensemble de sommets marqus avec la marque
 * AMarkNumber.
 *
 * @param AMarkNumber Un numro de marque
 */
CVertex barycenter(int AMarkNumber);

/**
 * Calcule le barycentre d'un ensemble de sommets se trouvant dans le
 * champ directInfo d'indice ADirectInfoVertex.
 * Les champs directinfo doivent tre initialiss avec la mthode
 * duplicateVertexToDirectInfo.
 *
 * @param ADirectInfoVertex Un indice correspondant  un champ directInfo
 */
CVertex directInfoBarycenter(int ADirectInfoVertex);

/**
 * Calcule la bote englobante d'un ensemble de sommets dont un brin au
 * moins est marqu avec la marque AMarkNumber.
 *
 * @param AMarkNumber Un numro de marque
 * @param AMin Le point minimum de la bote englobante
 * @param AMax Le point maximum de la bote englobante
 */
void boundingBox(int AMarkNumber, CVertex & AMin, CVertex & AMax);

/**
 * Calcule la bote englobante d'un ensemble de sommets dont un brin au
 * moins est marqu avec la marque AMarkNumber puis retourne le centre de
 * cette bote englobante.
 *
 * @param AMarkNumber Un numro de marque
 * @return Le centre de la bote englobante
 */
CVertex centerOfBoundingBox(int AMarkNumber);

/**
 * Calcule la bote englobante d'un ensemble de sommets situs sur l'orbite
 * (ADart, AOrbit).
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param AMin Le point minimum de la bote englobante
 * @param AMax Le point maximum de la bote englobante
 */
void boundingBox(CDart* ADart, TOrbit AOrbit, CVertex & AMin, CVertex & AMax);

/**
 * Calcule la bote englobante d'un ensemble de sommets situs sur l'orbite
 * (ADart, AOrbit) puis retourne le centre de cette bote englobante.
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @return Le centre de la bote englobante
 */
CVertex centerOfBoundingBox(CDart* ADart, TOrbit AOrbit);

/**
 * Calcule la longueur d'une arte.
 *
 * @param ADart Un brin de la carte
 * @return La longueur de l'arte incidente  ADart
 */
TCoordinate edgeLength(CDart* ADart);

/**
 * Calcule le primtre d'une face.
 *
 * @param ADart Un brin de la carte
 * @return Le primtre de la face incidente  ADart
 */
TCoordinate facePerimeter(CDart* ADart);

/**
 * Calcule la longueur d'un 1-bord, c'est--dire la longueur de l'arte qui
 * permettrait de boucher ce 1-bord.
 *
 * @param ADart Un brin de la carte
 * @return La longueur du 1-bord incident  ADart
 */
TCoordinate border1Length(CDart* ADart);

/**
 * Calcule la longueur d'une 2-bord, c'est--dire la somme de la longueur de
 * toutes les artes qui le constituent.
 *
 * @param ADart Un brin de la carte
 * @return La longueur du 2-bord incident  ADart
 */
TCoordinate border2Length(CDart* ADart);

/**
 * Calcule * findVertex(alpha0(ADart)) - * findVertex(ADart).
 *
 * @param ADart Un brin de la carte
 * @return Le vecteur correpondant  l'arte incidente  ADart
 */
CVertex edgeVector(CDart* ADart);

/**
 * Calcule le vecteur normal d'une arte. Ce vecteur n'est jamais nul.
 * Si l'arte est de longueur nulle, la mthode retourne OX, sinon
 * elle appelle 'CGeometry::getNormalVector'.
 *
 * @param ADart Un brin de la carte
 * @return Le vecteur normal de l'arte incidente  ADart
 */
CVertex edgeNormalVector(CDart* ADart);

/**
 * Calcule le vecteur normal d'une face. Ce vecteur n'est jamais nul.
 * La face n'est pas forcment plane. Elle peut tre ouverte.
 * Elle peut comporter moins de 3 sommets (le rsultat est alors OZ).
 *
 * @param ADart Un brin de la carte
 * @return Le vecteur normal de la face incidente  ADart
 */
CVertex faceNormalVector(CDart* ADart);

/**
 * Calcule le vecteur normal d'une cellule en s'appuyant sur les mthodes
 * 'edgeNormalVector' et 'faceNormalVector'.
 *
 * @param ADim La dimension de la cellule (0, 1 ou 2)
 * @param ADart Un brin incident  la cellule
 * @return Le vecteur normal de la cellule
 */
CVertex cellNormalVector(int ADim, CDart* ADart);

/**
 * Calcule le vecteur normal d'un 2-bord. Ce vecteur n'est jamais nul.
 * Le 2-bord n'est pas forcment plan. Il peut tre ouvert.
 * Il peut comporter moins de 3 sommets (le rsultat est alors OZ).
 *
 * @param ADart Un brin de la carte
 * @return Le vecteur normal du 2-bord incident  ADart
 */
CVertex border2NormalVector(CDart* ADart);

/**
 * Calcule la moyenne des vecteurs normaux des faces (restreintes au volume
 * auquel appartient ADart) incidentes  la cellule de dimension ADim 
 * laquelle ADart appartient.
 * Les faces 3-cousues ne sont pas prises en compte dans le calcul.
 * Le vecteur calcul peut tre nul.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @return La moyenne des vecteurs normaux des faces de la rgion donne
 */
CVertex regionNormalVector(CDart* ADart, int ADim);

/**
 * Calcule la moyenne des vecteurs normaux des faces incidentes  la cellule
 * de dimension ADim  laquelle ADart appartient.
 * Seules les faces dont un brin commun avec l'orbite
 * (ADart, ORBIT_INF[ADart]) est marqu sont prises en compte.
 * Le vecteur calcul peut tre nul.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (0, 1 ou 2)
 * @param AMarkNumber Un numro de marque
 * @return La moyenne des vecteurs normaux des faces de la rgion donne
 */
CVertex regionNormalVector(CDart* ADart, int ADim, int AMarkNumber);

/**
 * Calcule la moyenne des vecteurs normaux des faces de la cellule de
 * dimension ADim incidente  ADart.
 * Si la cellule n'est pas orientable, le vecteur retourn est nul.
 *
 * @param ADart Un brin de la carte
 * @param ADim Une dimension (3 ou 4)
 * @return Un vecteur (ventuellement nul)
 */
CVertex cellDimensionNormalVector(CDart* ADart, int ADim);

protected:

/**
 * Calcule la longueur de la circonfrence d'une orbite,
 * c'est--dire la somme de la longueur de toutes les artes qui la
 * constituent.
 * Ce calcul n'alpha de sens que si AOrbit==ORBIT01 ou si AOrbit==ORBIT_BORDER1
 * ou si AOrbit==ORBIT_BORDER2.
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite
 * @return La longueur de la circonfrence de l'orbite
 */
TCoordinate orbitLength(CDart* ADart, TOrbit AOrbit);

/**
 * Calcule le vecteur normal d'une orbite.
 * Ce calcul n'alpha de sens que si AOrbit==ORBIT01 ou si AOrbit==ORBIT_BORDER1
 * ou si AOrbit==ORBIT_BORDER2.
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite
 * @return Le vecteur normal de l'orbite
 */
CVertex orbitNormalVector(CDart* ADart, TOrbit AOrbit);

//******************************************************************************
