/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef DART_HH
#define DART_HH
//******************************************************************************
#include "kernel-types.hh"
#include "inline-macro.hh"

#include <bitset>
//******************************************************************************
class CAttribute;
class CEmbedding;

namespace GMap3d
{
  /**
   * Brin d'une 3-G-Carte.
   *
   * Cette classe contient entre autres les fonctions de manipulation bas
   * niveau des alphas, ainsi que celles de gestion des attributs (classe
   * CEmbedding).
   * Ses mthodes ne doivent pas tre directement appeles par un utilisateur
   * lambda : la manipulation des brins se fait toujours  l'aide de la classe
   * de la carte gnralise (CGMapBasic / CGMap).
   *
   * @author MOKA TEAM
   */
  class CDart
  {
  public:
    // @name Constructeurs et destructeur
    // @{

    /**
     * Constructeur par dfaut
     *
     * Construit un nouveau brin non marqu dont les alphas sont initialiss 
     * l'identit. Ce constructeur ne doit tre appel que par des utilisateurs
     * avertis (prfrer la mthode d'ajout de brin de la classe CGMap).
     *
     * @return Une nouvelle instance de la classe CDart
     */
    CDart();

    /**
     * Constructeur par recopie
     *
     * Construit un nouveau brin non marqu dont les alphas sont initialiss 
     * l'identit. Seules les marques du brin pass en paramtre sont
     * recopies.
     *
     * @return Une nouvelle instance de la classe CDart
     */
    CDart(const CDart& ADart);

    /**
     * Constructeur avec marquage
     *
     * Construit un nouveau brin marqu selon la suite de marques AMarks, et
     * dont les alphas sont initialiss  l'identit. Ce constructeur ne doit
     * tre appel que par des utilisateurs avertis (prfrer la mthode
     * d'ajout de brin de la classe CGMap).
     *
     * @param AMarks Un tableau de marques  appliquer au brin cr
     *
     * @return Une nouvelle instance de la classe CDart
     */
    CDart(const std::bitset<NB_MARKS>& AMarks);

    /**
     * Destructeur
     *
     * Dtruit une instance de la classe CDart.
     * Cette mthode est  surcharger lors de la cration d'une classe drive
     * de CDart.
     */
    ~CDart();

    // @}
    /**
     * @name Utilisation des champs direct-info
     *
     * Une information directe est un pointeur vers une structure de donnes
     * quelconque. Le nombre d'informations par brin est fix par la macro
     * NB_DIRECT_INFO (voir fichier kernel-types.hh). Ceci permet un accs aux
     * informations plus rapide qu'en passant par les attributs et les
     * embeddings mais plus coteux en place mmoire.
     */
    // @{

    /**
     * @param AIndice Le numro de l'information directe  rcuprer.
     *   L'information directe permet de simuler une orbite vide
     *   (ORBIT_SELF) de manire plus rapide mais il n'y a aucune
     *   vrification.
     * @return Un pointeur sur l'information recherche.
     *
     * @precondition 0 <= AIndice < NB_DIRECT_INFO
     */
    void* getDirectInfo(int ADirectIndex) const;

    /**
     * @param AIndice Le numro de champ  modifier
     * @param APointer La valeur  affecter
     *
     * @precondition 0 <= AIndice < NB_DIRECT_INFO
     */
    void setDirectInfo(int ADirectIndex, void* AValue);

    // @}
    // @name Gestion des plongements
    // @{

    /**
     * Rcupration du premier lment de la liste des "CEmbedding"
     *
     * Un "CEmbedding" est une liste d'attributs appartenant  une mme orbite,
     * dont le brin est propritaire (voir de plus amples explications avec la
     * clase CEmbedding).
     * Chaque liste d'attributs d'une mme orbite est elle-mme chane avec
     * les autres listes d'un mme brin. Cette mthode retourne le premier
     * lment (un "CEmbedding") de la liste chane des listes d'attributs.
     *
     * @return La premire liste d'attributs
     */
    CEmbedding* getFirstEmbedding() const;

    /**
     * Dfinition du premier lment de la liste des "CEmbedding"
     *
     * Cette mthode pose un "CEmbedding" comme tant le premier lment de la
     * liste des "CEmbedding", sans aucune vrification d'existence du premier
     * "CEmbedding". Cette mthode ne doit tre utilise que par la carte
     * gnralise.
     *
     * @param AEmbedding la liste d'attributs  placer en premire position.
     */
    void setFirstEmbedding(CEmbedding* AEmbedding);

    /**
     * Rcupration de la liste d'attributs plongs sur une orbite donne
     *
     * Cette mthode recherche dans la liste des "CEmbedding" celui qui
     * correspond  l'orbite orbit.
     *
     * @param AOrbit l'orbite correspondant  la liste d'attributs recherchs
     * @return la liste d'attributs si elle est trouve
     */
    CEmbedding* getEmbedding(TOrbit AOrbit) const;

    /**
     * Ajout d'une liste d'attributs  l'instance
     *
     * Cette mthode ajoute la liste AEmbedding  la liste des "CEmbedding".
     *
     * @param AEmbedding une liste d'attributs
     * @precondition La liste AEmbedding n'est pas vide et aucune liste de mme
     *   orbite n'existe sur le brin instanci.
     * @precondition les champs prev et next de AEmbedding sont a NULL
     */
    void addEmbedding(CEmbedding* AEmbedding);

    /**
     * Suppression d'une liste d'attributs donne dans la liste des
     * "CEmbedding" de l'instance
     *
     * Cette mthode ote la liste d'attributs E de la liste des "CEmbedding".
     *
     * @param AEmbedding la liste d'attributs  enlever
     * @return la liste AEmbedding
     *
     * @precondition AEmbedding n'est pas vide
     */
    CEmbedding* removeEmbedding(CEmbedding* AEmbedding);

    /**
     * Suppression de la liste d'attributs correspondant  une orbite donne
     * dans les "CEmbedding"
     *
     * Cette mthode recherche la liste d'attributs correspondant  l'orbite
     * orbit, et la supprime de la liste des "CEmbedding".
     *
     * @param AOrbit l'orbite de la liste  enlever
     * @return la liste correspondante
     */
    CEmbedding* removeEmbedding(TOrbit AOrbit);

    /**
     * Suppression puis destruction d'une liste d'attributs donne dans la
     * liste des "CEmbedding"
     *
     * Cette mthode ote la liste d'attributs E de la liste des "CEmbedding"
     * avant de la dtruire.
     *
     * @param AEmbedding la liste d'attributs  dtruire
     * @precondition E n'est pas vide
     */
    void deleteEmbedding(CEmbedding* AEmbedding);

    /**
     * Suppression puis destruction de la liste d'attributs correspondant  une
     * orbite donne
     *
     * Cette mthode recherche la liste d'attributs correspondant  l'orbite
     * AOrbit, la retire de la liste des "CEmbedding", et enfin la dtruit.
     *
     * @param AOrbit l'orbite de la liste  dtruire
     */
    void deleteEmbedding(TOrbit AOrbit);

    /**
     * Ajout, avec fusion le cas chant, d'une liste d'attributs  l'instance
     *
     * Cette mthode vrifie si une liste d'attributs de mme orbite existe
     * dj sur le brin instanci. Si ce n'est pas le cas, la liste E est
     * simplement ajoute. Sinon les attributs de E sont ajouts  la liste
     * dj prsente.
     *
     * @precondition AEmbedding n'est pas vide
     */
    void mergeEmbedding(CEmbedding* AEmbedding);

    /**
     * Rcupration d'un attribut plong sur une orbite donne
     *
     * Cette mthode trouve dans la liste des "CEmbedding" la liste d'attributs
     * qui correspond  l'orbite orbit, puis y recherche l'attribut de type
     * AAttribType.
     *
     * @param AOrbit l'orbite de plongement de l'attribut recherch
     * @param AAttribType le numro d'identifiant de son type
     * @return l'attribut si il est trouv
     */
    CAttribute* getAttribute (TOrbit AOrbit, TAttributeId AAttribType) const;

    /**
     * Ajout d'un attribut sur une orbite donne
     *
     * Cette mthode recherche dans la liste des "CEmbedding" la liste
     * d'attributs qui correspond  l'orbite orbit. Si elle est trouve,
     * l'attribut data y est ajout. Sinon une liste d'attributs contenant
     * data est cr pour cette orbite.
     *
     * @param orbit l'orbite de plongement de l'attribut  ajouter
     * @param data l'attribut
     * @precondition l'attribut data ne peut tre nul
     */
    void addAttribute(TOrbit AOrbit, CAttribute* AAttribute);

    /**
     * Suppression d'un attribut de la liste des attributs d'une orbite donne
     *
     * Cette mthode recherche la liste d'attributs correspondant  l'orbite
     * orbit, et si elle existe, enlve l'attribut de type typeAttr qu'elle
     * contient. Si la liste d'attributs est alors vide, elle est supprime de
     * la liste des "CEmbedding".
     *
     * @param orbit l'orbite de plongement de l'attribut  supprimer
     * @param attrType le numro d'identifiant de son type
     * @return l'attribut enlev
     */
    CAttribute* removeAttribute(TOrbit AOrbit, TAttributeId AAttribType);

    /**
     * Suppression et destruction d'un attribut sur une orbite donne
     *
     * Cette mthode supprime l'attribut, de type AAttrType, de la liste des
     * attributs de l'orbite AOrbit, puis le dtruit.
     *
     * @param AOrbit l'orbite de plongement de l'attribut  supprimer
     * @param AAttrType le numro d'identifiant de son type
     */
    void deleteAttribute(TOrbit AOrbit, TAttributeId AAttribType);

    /**
     * Mthode prcisant que la prsence d'une orbite donne dans la liste des
     * "CEmbedding" de l'instance.
     *
     * Cette mthode dfinit comme utilise l'orbite orbit sur le brin. Aucune
     * vrification supplmentaire n'est effectu sur la cohrence des
     * plongements, cette mthode tant une mthode de premier niveau.
     *
     * @param AOrbit l'orbite  utiliser
     * @precondition ORBIT_SELF <= AOrbit <= ORBIT0123
     */
    void setOrbitUsed(TOrbit AOrbit);

    /**
     * Mthode testant si une orbite est utilise sur le brin
     *
     * @return le boolen rpondant au test
     * @precondition ORBIT_SELF <= AOrbit <= ORBIT0123
     */
    bool isOrbitUsed(TOrbit AOrbit) const;

    /**
     * Mthode librant l'utilisation d'une orbite donne sur le brin instanci
     *
     * Aucune vrification supplmentaire n'est effectu sur la cohrence des
     * plongements.
     *
     * @precondition ORBIT_SELF <= AOrbit && AOrbit <= ORBIT0123
     */
    void unsetOrbitUsed(TOrbit AOrbit);

    // @}
    // @name Chanage des brins dans la carte gnralise
    // @{

    /**
     * Rcupration du brin suivant dans la liste
     *
     * La liste chane de brins permet  la carte gnralise de passer en
     * revue l'ensemble des brins qu'elle contient. Cette mthode permet de se
     * dplacer au prochain brin. Elle ne doit pas tre utilis par un
     * utilisateur non averti (prfrer un parcours : dynamic ou static
     * CCoverageAll).
     *
     * @return le brin de la liste chane qui suit le brin instanci.
     */
    CDart* getNext() const;

    /**
     * Rcupration du brin prcdent dans la liste
     *
     * Mme remarque que prcdemment.
     *
     * @return le brin de la liste chane qui prcde le brin instanci.
     */
    CDart* getPrev() const;

    /**
     * Dfinition du brin suivant dans la liste
     *
     * Cette mthode permet de dfinir quel est le brin suivant l'instance dans
     * la liste. Cette mthode ne doit pas tre directement utilise, elle
     * permet  la carte gnralise d'ajouter un brin dans la liste (prfrer
     * addMapDart de la classe CGMapBasic).
     *
     * @param le brin de la liste chane qui suit l'instance.
     */
    void setNext(CDart* ADart);

    /**
     * Dfinition du brin prcdent dans la liste
     *
     * Mme remarque que prcdemment.
     *
     * @param le brin de la liste chane qui prcde l'instance.
     */
    void setPrev(CDart* ADart);

    // @}
    // @name Accs aux involutions
    // @{

    /**
     * Accesseur de l'involution alpha0
     *
     * Cette mthode renvoie le brin qui est cousu par alpha0  l'instance.
     * Si cette instance n'est pas cousue  un autre brin, la mthode rentourne
     * l'instance elle-mme. Prfrer tout de mme la mthode de mme nom de la
     * classe CGMapBasic/CGMap.
     *
     * @return le brin cousu par alpha0.
     */
    CDart* getAlpha0() const;

    /**
     * Accesseur de l'involution alpha1
     *
     * Mme remarque qu' la mthode prcdente.
     *
     * @return le brin cousu par alpha1.
     */
    CDart* getAlpha1() const;

    /**
     * Accesseur de l'involution alpha2
     *
     * Mme remarque qu' la mthode prcdente.
     *
     * @return le brin cousu par alpha2.
     */
    CDart* getAlpha2() const;

    /**
     * Accesseur de l'involution alpha3
     *
     * Mme remarque qu' la mthode prcdente.
     *
     * @return le brin cousu par alpha3.
     */
    CDart* getAlpha3() const;

    /**
     * Accesseur de l'involution alphaI
     *
     * Cette mthode renvoie le brin qui est cousu par alphaI  l'instance,
     * avec AIndex un numro donn par l'utilisateur.
     *
     * @param ADimension le numro de la liaison  regarder (entre 0 et 3).
     * @return le brin cousu par alphaI.
     */
    CDart* getAlpha(int ADimension) const;

    /**
     * Dfinition de l'involution alpha0
     *
     * Cette mthode permet de lier un brin  l'instance par alpha0. Seule la
     * classe CGMapBasic/CGMap peut l'utiliser (prfrer pour les coutures les
     * mthodes *  de la classe CGMapBasic/CGMap telles que sew0, voire
     * dartSew0 ou topoSew0, et en dernire solution linkAlpha0).
     *
     * @param le brin  lier par alpha0  l'instance.
     */
    void setAlpha0(CDart* ADart);

    /**
     * Dfinition de l'involution alpha1
     *
     * Mme remarque que prcdemment.
     *
     * @param D le brin  lier par alpha1  l'instance.
     */
    void setAlpha1(CDart* ADart);

    /**
     * Dfinition de l'involution alpha2
     *
     * Mme remarque que prcdemment.
     *
     * @param D le brin  lier par alpha2  l'instance.
     */
    void setAlpha2(CDart* ADart);

    /**
     * Dfinition de l'involution alpha3
     *
     * Mme remarque que prcdemment.
     *
     * @param D le brin  lier par alpha3  l'instance.
     */
    void setAlpha3(CDart* ADart);

    /**
     * Dfinition de l'involution alphaI
     *
     * Mme remarque que prcdemment (prfrer sew, voire dartSew ou topoSew,
     * et en dernier recours linkAlpha).
     *
     * @param ADart le brin  lier par alphaI  l'instance.
     * @param ADimension le dimension de la liaison alpha  considrer.
     */
    void setAlpha (CDart* ADart, int ADimension);

    /**
     * Test permettant de savoir si un brin est libre (non cousu) par alpha0.
     *
     * Cette mthode teste si l'instance est cousu  lui-mme par alpha0. Elle
     * renvoie VRAI si c'est le cas, FAUX si le brin instanci est cousu  un
     * autre brin.
     *
     * @return un boolen signifiant si l'instance est libre par alpha0.
     */
    bool isFree0() const;

    /**
     * Test permettant de savoir si un brin est libre (non cousu) par alpha1.
     *
     * Mme remarque que prcdemment.
     *
     * @return un boolen signifiant si le brin instanci est libre par alpha1.
     */
    bool isFree1() const;

    /**
     * Test permettant de savoir si un brin est libre (non cousu) par alpha2.
     *
     * Mme remarque que prcdemment.
     *
     * @return un boolen signifiant si le brin instanci est libre par alpha2.
     */
    bool isFree2() const;

    /**
     * Test permettant de savoir si un brin est libre (non cousu) par alpha3.
     *
     * Mme remarque que prcdemment.
     *
     * @return un boolen signifiant si le brin instanci est libre par alpha3.
     */
    bool isFree3() const;

    /**
     * Test permettant de savoir si un brin est libre (non cousu) par alphaI.
     *
     * Cette mthode teste si l'instance est cousu  lui-mme par alphaI, avec
     * i la dimension de la liaison  regarder. Elle renvoie VRAI si c'est le
     * cas, FAUX si l'instance est cousu  un autre brin.
     *
     * @param ADimension la dimension de la liaison  tester.
     * @return un boolen signifiant si l'instance est libre par alpha3.
     */
    bool isFree(int ADimension) const;

    /**
     * Libration de l'involution alpha0 de l'instance.
     *
     * Cette mthode coud pour alpha0 l'instance  elle-mme. Le brin
     * devient libre par alpha0, mais aucune vrification de la cohrence
     * de la carte gnralise n'est faite. Prfrer donc les mthodes de
     * la classe CGMapBasic/CGMap telles que unsew0, voire dartUnsew0 ou
     * topoUnsew0, et en dernier ressort unlinkAlpha0.
     */
    void setFree0();

    /**
     * Libration de l'involution alpha1 de l'instance.
     *
     * Mme remarque que prcdemment.
     */
    void setFree1();

    /**
     * Libration de l'involution alpha2 de l'instance.
     *
     * Mme remarque que prcdemment.
     */
    void setFree2();

    /**
     * Libration de l'involution alpha3 de l'instance.
     *
     * Mme remarque que prcdemment.
     */
    void setFree3();

    /**
     * Libration pour le brin instanci de son involution alphaI
     *
     * Cette mthode coud pour alphaI l'instance  elle-mme. Le brin
     * devient libre par alphaI, mais aucune vrification de la cohrence
     * de la carte gnralise n'est faite. Prfrer donc les mthodes de
     * la classe CGMapBasic/CGMap telles que unsew, voire dartUnsew ou
     * topoUnsew, et en dernier ressort unlinkAlpha.
     *
     * @param ADimension le dimension de la liaison alphaI  considrer.
     */
    void setFree(int ADimension);

    // @}
    // @name Gestion des marques
    // @{

    /**
     * Test permettant de savoir si un brin est marqu.
     *
     * Cette mthode renvoie VRAI si le brin possde la marque i.
     * Le nombre de marques par brin est fix par la variable
     * NB_MARKS du fichier dart.hpp.
     *
     * @param i Le numro de la marque  tester (entre 0 et NB_MARKS-1).
     * @return boolen signifiant si le brin est marqu.
     * @precondition 0 <= i < NB_MARKS
     */
    bool getMark(int AMarkNumber) const;

    /**
     * Mthode dfinissant si un brin est marqu ou non.
     *
     * Cette mthode dfinit la i-me marque selon le boolen value.
     * Si value est  VRAI, le brin sera marqu par la marque i.
     *
     * @param i Le numro de la marque  changer (entre 0 et NB_MARKS-1).
     * @param value La valeur boolenne de la marque.
     * @precondition 0 <= i < NB_MARKS
     */
    void setMark(int AMarkNumber, bool AValue);

    /**
     * Mthode d'accs  l'ensemble des marques d'un brin.
     *
     * Cette mthode rcupre toutes les marques d'un brin.
     *
     * @return l'ensemble de marques (de type bitset)
     */
    std::bitset<NB_MARKS> getMarks() const;

    /**
     * Mthode dfinissant en une seule fois l'ensemble des marques d'un brin.
     *
     * Cette mthode dfinit chacune des marques d'un brin suivant la suite de
     * valeurs donnes par marks.
     *
     * @param marks l'ensemble des valeurs  prendre par les marques du brin
     */
    void setMarks(const std::bitset<NB_MARKS> & AMarks);

    // @}

  private:
    // @name Champs privs
    // @{

    /// Tableau des liaisons alpha (suite de pointeurs sur un brin)
    CDart* FAlpha[4];

    /// Dbut de la liste chane des plongements
    CEmbedding* FFirstEmbedding;

    /// Valeur des marques boolennes du brin
    std::bitset<NB_MARKS> FMarks;

    /**
     * Tableau des orbites plonges sur le brin, seules les plongements dont
     * il est propritaire sont signals dans cette suite (voir la notion de
     * proprit d'un plongement dans la classe CGMap).
     */
    std::bitset<16> FUsedOrbits;

    /// Tableau des informations directes
    void* FDirectInfo[NB_DIRECT_INFO];

    /**
     * Variables permettant de dfinir les brins suivant et prcdent dans la
     * liste chane des brins de la carte gnralise.
     */
    CDart* FPrev;
    CDart* FNext;

    // @}
  };

} // namespace GMap3d
//******************************************************************************
#include INCLUDE_INLINE("dart.icc")
//******************************************************************************
#endif // DART_HH
//******************************************************************************
