/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef EMBEDDING_HH
#define EMBEDDING_HH
//******************************************************************************
#include "kernel-types.hh"
#include "inline-macro.hh"
//******************************************************************************
class CAttribute;

/**
 * Cette classe reprsente une liste d'attributs associs  la mme orbite.
 * Les mthodes de cette classe permettent d'ajouter, supprimer, chercher...
 * un attribut particulier.
 * La plupart de ces mthodes ne font pas de contrle car elles sont appeles
 * de CDart.
 *
 * @author MOKA TEAM
 */
class CEmbedding
{
public:
  // @name Constructeurs et destructeur
  // @{

  /**
   * Construction d'une nouvelle instance de la classe.
   * Le champ id est initialis.
   *   La liste d'attribut est initialise  vide.
   *
   * @param AOrbit Une orbite
   * @return Une nouvelle instance de la classe
   */
  CEmbedding(TOrbit AOrbit);

  /**
   * Construction par recopie d'une nouvelle instance de la classe.
   *
   * @param AEmbedding L'instance  recopier
   * @return Une nouvelle instance de la classe
   */
  CEmbedding(const CEmbedding& AEmbedding);

  /**
   * Destruction de l'instance.
   * Tous les attributs de la liste sont ventuellement dtruits
   * par l'appel de la mthode destroy de la classe attribut.
   */
  ~CEmbedding();

  /**
   * Mthode pour recopier un CEmbedding.
   *
   * Construction d'une nouvelle instance de la classe.
   * La liste des attributs de cette instance est initialise par l'appel
   * de la mthode copy sur chaque attribut de AEmbedding.
   * Le champ id est initialis avec la mme orbite que celle de AEmbedding
   *
   * @return Une copie de l'instance. Cette copie peut etre l'instance
   *   elle-mme. (par exemple si l'on veut des embedding partages.)
   */
  CEmbedding* copy() const;

  /**
   * Mthode pour suprimer un CEmbedding.
   *
   * Mthode pendante de la mthode copy. Si la mthode copy cr une nouvelle
   * instance, la methode destroy doit la dtruire. Dans le cas d'embedding
   * partags, on peut par exemple compter le nombre de rferences.
   */
  void destroy();

  // @}

  // @name Accesseurs
  // @{

  /**
   * Accesseur en lecture de l'orbite de l'instance
   */
  TOrbit getOrbit() const;

  /**
   * Accesseur en lecture du prochain CEmbedding dans la liste
   *
   * Le chanage des CEmbedding est effectu dans cette classe.
   * Cet accesseur permet de rcuperer le prochain CEmbedding dans cette
   * liste.
   */
  CEmbedding* getNext() const;

  /**
   * Modifie le chainage des CEmbedding. L'embedding suivant de l'instance
   * est dsormais AEmbedding.
   *
   * @param AEmbedding Un pointeur vers un CEmbedding
   */
  void setNext(CEmbedding* AEmbedding);

  /**
   * Accesseur en lecture du prcdent CEmbedding dans la liste
   *
   * @return Un pointeur vers un CEmbedding
   */
  CEmbedding* getPrev() const;

  /**
   * Modifie le chainage des CEmbedding. L'CEmbedding prcdent de
   * l'instance est dsormais AEmbedding.
   *
   * @param AEmbedding Un pointeur vers un CEmbedding
   */
  void setPrev(CEmbedding* AEmbedding);

  /**
   * Accesseur en lecture du premier attribut de la liste
   */
  CAttribute* getFirstAttribute() const;

  /**
   * Accesseur en criture du premier attribut de la liste
   * Affecte le premier attribut de la liste qui est dsormais AAttribute
   *
   * @param AAttribute Un pointeur vers un CAttribute
   */
  void setFirstAttribute(CAttribute* AAttribute);

  /**
   * Accesseur pour tester si la liste des attributs est vide
   *
   * @return Un boolen indiquant si la liste est vide
   */
  bool isEmpty() const;

  // @}
  // @name Mthodes pour travailler sur les attributs
  // @{

  /**
   * Cherche s'il existe un attribut dans la liste ayant comme identifiant
   * attribType. Si oui retourne un pointeur vers cet attribut, sinon
   * retourne NULL.
   *
   * @param attribType Un TAttributeId
   * @return Un pointeur vers un CAttribute
   */
  CAttribute* getAttribute(TAttributeId AAttribType) const;

  /**
   * Ajoute l'attribut dans la liste
   *
   * @param AAttribute un pointeur vers un CAttribute
   *
   * @precondition getAttribute(AAttribute->getType()) == NULL
   * @precondition AAttribute->getPrev() == AAttribute->getNext() == NULL
   */
  void addAttribute(CAttribute* AAttribute);

  /**
   * Enlve (sans le dtruire) l'attribut de la liste.
   *
   * @param AAttribute un pointeur vers un CAttribute
   * @return Le pointeur vers l'attribut enlev de la liste
   *
   * @precondition AAttribute fait partie de la liste des attributs de
   *   l'instance.
   */
  CAttribute* removeAttribute(CAttribute* AAttribute);

  /**
   * Enlve l'attribut ayant l'identifiant attribType de la liste. Si aucun
   * attribut n'est de ce type, alors retourne NULL.
   *
   * @param attribType un TAttributeId
   * @return Le pointeur vers l'attribut enlev de la liste, ou NULL
   */
  CAttribute* removeAttribute(TAttributeId attribType);

  /**
   * Enlve l'attribut de la liste et le dtruit eventuellement en appelant
   * la mthode destroy sur cet attribut.
   *
   * @param AAttribute un pointeur vers un CAttribute
   *
   * @precondition AAttribute fait partie de la liste des atributs de
   *   l'instance
   */
  void deleteAttribute(CAttribute* AAttribute);

  /**
   * Enlve l'attribut ayant l'identifiant attribType de la liste et le
   * dtruit eventuellement en appelant la mthode destroy sur cet attribut.
   * Si aucun attribut n'est de ce type, ne fait rien.
   *
   * @param attribType un TAttributeId
   */
  void deleteAttribute(TAttributeId attribType);

  /**
   * Ajoute tout les attributs de AEmbedding non prsent dans la liste des
   * attributs de l'instance dans cette liste. Ces attributs sont enlevs
   * de la liste des attributs de AEmbedding.
   *
   * @param AEmbedding un pointeur vers un CEmbedding
   *
   * @precondition l'instance et AEmbedding concernent la mme orbite
   */
  void mergeAttribute(CEmbedding* AEmbedding);

  // @}

private:
  // @name Champs privs
  // @{

  /// Pointeur vers le premier CAttribute de la liste
  CAttribute* FFirstAttribute;

  /// Identifiant de l'orbite de l'instance
  TOrbit FId;

  /// Pour le double-chainage des CEmbedding, utilis par la classe CDart
  CEmbedding* FPrev;
  CEmbedding* FNext;

  // @}
};
//******************************************************************************
#include INCLUDE_INLINE("embedding.icc")
//******************************************************************************
#endif // EMBEDDING_HH
//******************************************************************************
