/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef G_MAP_HH
#define G_MAP_HH
//******************************************************************************
#include "g-map-basic.hh"
#include "inline-macro.hh"
// Includes facultatifs :
#include "dart.hh"
#include "attribute.hh"
#include "embedding.hh"
#include "coverage.hh"
#include "dynamic-coverage.hh"
#include "static-coverage.hh"
//******************************************************************************
namespace GMap3d
{
  /**
   * Cette classe fournit tous les outils de base pour manipuler une G-carte,
   * et notamment :
   *  - les coutures et dcousures (topologiques et gomtrique),
   *  - les mthodes pour grer les attributs (ajouter, enlever...),
   *  - des mthodes utilises en interne pour grer les attributs.
   *
   *  @author MOKA TEAM
   */
  class CGMap : public CGMapBasic
  {
  public:
    // @name Constructeur et destructeur
    // @{

    /// Constructeur
    CGMap();
  
    /// Destructeur
    virtual ~CGMap();

    // @}
  
    // @name Mthodes "sres" pour la couture et dcousure.
  
    /**
     * @param ADart1 et ADart2 Deux brins
     * @return vrai si la [i]-couture de ces deux brins est possible,
     *   faux sinon.
     */
    // @{
    bool canSew0(CDart* ADart1, CDart* ADart2);
    bool canSew1(CDart* ADart1, CDart* ADart2);
    bool canSew2(CDart* ADart1, CDart* ADart2);
    bool canSew3(CDart* ADart1, CDart* ADart2);
    // @}

    /**
     * @param ADart1 et ADart2 Deux brins
     * @param ADimension Une dimension
     * @return si la [ADimension]-couture de ces deux brins est possible.
     */  
    bool canSew(CDart* ADart1, CDart* ADart2, int ADimension);

    /**
     * [i]-couture gomtrique des 2 brins. Les attributs sont mis  jour
     * lorsque c'est ncessaire (quand deux orbites sont fusionnes et que
     * chacune avait un attribut). De plus, la couture est effectue sur
     * l'orbite concerne afin de garantir la validit de la G-carte obtenue.
     *
     * @param ADart1 et ADart2 Deux brins
     * @precondition canSew[i](ADart1, ADart2)
     */
    // @{
    void sew0(CDart* ADart1, CDart* ADart2);
    void sew1(CDart* ADart1, CDart* ADart2);
    void sew2(CDart* ADart1, CDart* ADart2);
    void sew3(CDart* ADart1, CDart* ADart2);
    // @}
  
    /**
     * Couture gomtrique des deux brins pour la dimension donne (cf. sew[i])
     *
     * @param ADart1 et ADart2 Deux brins
     * @param ADimension Une dimension
     * @precondition canSew(ADart1, ADart2, ADimension)
     */
    void sew(CDart* ADart1, CDart* ADart2, int ADimension);

    /**
     * [i]-dcousure gomtrique. Les attributs sont mis  jour lorsque c'est
     * ncessaire (quand on coupe une orbites en deux). De plus, la dcousure
     * est effectue sur l'orbite concerne afin de garantir la validit
     * de la G-carte obtenue.
     *
     * @param ADart Un brin
     */
    // @{
    void unsew0(CDart* ADart);
    void unsew1(CDart* ADart);
    void unsew2(CDart* ADart);
    void unsew3(CDart* ADart);
    // @}

    /**
     * Dcousure gomtrique du brin pour la dimension donne. (cf. unsew[i])
     *
     * @param ADart Un brin
     * @param ADimension Une dimension
     */
    void unsew(CDart* ADart,int ADimension);

    /**
     * @name Mthodes pour rcuprer des parcours tant donn un brin et une
     *   orbite. Voir les diffrents types de parcours dans
     *   dynamic-coverage.hh et static-coverage.hh.
     */
    // @{

    CCoverage* getDynamicCoverage     (const CDart* ADart, TOrbit AOrbit);
    CCoverage* getStaticCoverage      (const CDart* ADart, TOrbit AOrbit);
    CCoverage* getMultiDynamicCoverage(const CDart* ADart, TOrbit AOrbit);
    CCoverage* getBasicDynamicCoverage(const CDart* ADart, TOrbit AOrbit,
				       int AMarkNumber);

    // @}
    // @name Mthodes de gestion des attributs.
    // @{

    /**
     * @return Un pointeur vers l'attribut de type donn appartenant  l'orbite
     *   donn du brin ADart. NULL si un tel attribut n'existe pas.
     */
    CAttribute* getAttribute(const CDart* ADart, TOrbit, TAttributeId);

    /**
     * Ajoute l'attribut donn  l'orbite donn du brin ADart.
     * ATTENTION : cette mthode ne doit pas tre appel si un tel attribut
     * existe dj. Cette prcondition doit tre assure par l'appelant.
     *  UTILISER AVEC ATTENTION.
     */  
    void addAttribute(CDart* ADart, TOrbit, CAttribute*);

    /**
     * @return Un pointeur vers l'attribut de type donn appartenant  l'orbite
     *   donn du brin ADart et enlve cet attribut de cette orbite.
     *   Retourne NULL si un tel attribut n'existe pas.
     */  
    CAttribute* removeAttribute(CDart* ADart, TOrbit, TAttributeId);

    /**
     * Dtruit l'attribut de type donn appartenant  l'orbite donn du brin
     * ADart et enlve cet attribut de cette orbite. 
     * Rien ne se passe si un tel attribut n'existe pas.
     */  
    void deleteAttribute(CDart* ADart, TOrbit, TAttributeId);
  
    /**
     * @param ADart1 et ADart2 Deux brins
     * @param AOrbit Une orbite 
     *
     * @return vrai si ADart1 et ADart2 appartiennent  la mme orbite <AOrbit>
     *   faux sinon
     */
    bool isSameOrbit(CDart* ADart1, CDart* ADart2, TOrbit AOrbit);
  
    /**
     * @param ADart  Un brin
     * @param AOrbit Une orbite
     *
     * @return un pointeur vers le brin appartenant  l'orbite <AOrbit> de
     *   ADart contenant l'embedding pour cette orbite. 
     *   Retourne NULL si un tel brin n'existe pas (pas d'embedding).
     */
    CDart* getEmbeddingOwner(const CDart* ADart, TOrbit AOrbit);

    // @}
    /**
     * @name Mthodes "non-sres" pour la coutures. Le rsultat de ces
     *   mthodes entrane l'obtention d'un objet ventuellement non valide
     *   (une 3-G-carte ne vrifiant pas les contraintes sur les alpha, ou
     *   plusieurs attributs sur une mme orbite...). La responsabilit de la
     *   validit est laiss  l'appelant. Ces mthodes servent malgr tout pour
     *   optimiser des traitements particulier. A UTILISER AVEC ATTENTION.
     */

    /**
     * [i]-couture gomtrique des 2 brins uniquement (et pas de toute l'orbite
     * contrairement  sew[i]). Les attributs sont mis  jour lorsque c'est
     * ncessaire (quand deux orbites sont fusionnes et que
     * chacune avait un attribut). 
     *
     * @param ADart1 et ADart2 Deux brins
     */
    // @{
    void dartSew0(CDart*, CDart* ADart2);
    void dartSew1(CDart*, CDart* ADart2);
    void dartSew2(CDart*, CDart* ADart2);
    void dartSew3(CDart*, CDart* ADart2);
    void dartSew (CDart*, CDart* ADart2, int ADimension);
    // @}

    /**
     * [i]-dcousure gomtrique uniquement du brin (et pas de toute l'orbite
     * contrairement  unsew[i]). Les attributs sont mis  jour lorsque c'est
     * ncessaire (quand on coupe une orbites en deux). 
     *
     * @param ADart Un brin
     */
    // @{
    void dartUnsew0(CDart* ADart);
    void dartUnsew1(CDart* ADart);
    void dartUnsew2(CDart* ADart);
    void dartUnsew3(CDart* ADart);
    void dartUnsew (CDart* ADart, int ADimension);
    // @}

    /**
     * [i]-couture topologique des 2 brins (les attributs ne sont pas mis 
     * jour contrairement  sew[i]). La couture est effectue sur l'orbite
     * concerne afin de garantir la validit de la G-carte obtenue.
     *
     * @param ADart1 et ADart2 Deux brins
     * @precondition canSew[i](ADart1, ADart2)
     */
    // @{
    void topoSew0(CDart* ADart1, CDart* ADart2);
    void topoSew1(CDart* ADart1, CDart* ADart2);
    void topoSew2(CDart* ADart1, CDart* ADart2);
    void topoSew3(CDart* ADart1, CDart* ADart2);
    void topoSew (CDart* ADart1, CDart* ADart2, int ADimension);
    // @}

    /**
     * [i]-dcousure topologique (les attributs ne sont pas mis  jour
     * contrairement  unsew[i]). La dcousure est effectue sur l'orbite
     * concerne afin de garantir la validit de la G-carte obtenue.
     *
     * @param ADart Un brin
     */
    // @{
    void topoUnsew0(CDart* ADart);
    void topoUnsew1(CDart* ADart);
    void topoUnsew2(CDart* ADart);
    void topoUnsew3(CDart* ADart);
    void topoUnsew (CDart* ADart, int ADimension);
    // @}

  private:
    // @name Mthodes utilises par sew et unsew pour grer les attributs.
    // @{
  
    /**
     * Met  jour tout les attributs des deux brins lorsque c'est ncessaire.
     * Cette mthode est appele avant une opration de couture. Elle teste
     * pour chaque orbite qui vont tre regroup par la couture, s'il existe
     * un embedding  la fois pour l'orbite contenant ADart1 et celle contenant
     * ADart2. Si c'est le cas, l'embedding associ  ADart1 est supprime, et
     * l'embedding de ADart2 rcupre les ventuels attributs qu'il n'avait
     * pas dj. Les autres attributs sont dtruits.
     *
     * @param ADart1 et ADart2 Deux brins
     * @param AMask un masque de 16 boolens.
     */
    void groupAttributes(CDart* ADart1, CDart* ADart2, TMask16);

    /**
     * Met  jour tout les attributs des deux brins lorsque c'est ncessaire.
     * Cette mthode est appele aprs une opration de dcousure. Elle teste
     * pour chaque orbite qui ont t concerns par la dcousure, s'il existe
     * un embedding d'un cot et pas de l'autre. Si c'est le cas, l'embedding
     * existant est dupliqu (mthode copy de la classe CEmbedding) et ajout
     * du cot qui n'en avait pas.
     *
     * @param ADart1 et ADart2 Deux brins
     * @param AMask un masque de 16 boolens.
     */
    void duplicateAttributes(CDart* ADart1, CDart* ADart2, TMask16);
  
    /**
     * Optimisation pour les mthodes {duplicate/group}Attributes
     *
     * Mthode quivalente  isSameOrbit.
     * Le paramtre ADartEmbeddingOwner1 contient le propritaire de l'embedding
     * si celui-ci est trouv lors du parcours, sinon NULL.
     * Mthode quivalente   :
     *      if (isSameOrbit(ADart1,ADart2))
     *             ADartEmbeddingOwner1=getEmbeddingOwner(ADart1,AOrbit);
     * mais en un seul parcours de l'orbite.
     */
    bool isSameOrbit(CDart* ADart1, CDart* ADart2, TOrbit AOrbit,
		     CDart*& ADartEmbeddingOwner1);
  };

} // namespace GMap3d
//******************************************************************************
#include INCLUDE_INLINE("g-map.icc")
//******************************************************************************
#endif // G_MAP_HH
//******************************************************************************
