/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef TRANSFORMATION_MATRIX_HH
#define TRANSFORMATION_MATRIX_HH
//******************************************************************************
#include "math-extension.hh"
#include "vertex.hh"
#include "inline-macro.hh"

#include <iostream>
//******************************************************************************

class CVector;

/**
 * Dfinit la classe permettant d'effectuer des transformations gomtriques
 * sur les objets en utilisant des matrices 4x4.
 *
 * @author MOKA TEAM
 */
class CTransformationMatrix
{
private:
  TCoordinate FElements[4][4];

public:
  // @name Constantes
  // @{

  /**
   * Types de matrices particulires.
   */
  typedef enum { NullMatrix, IdentityMatrix, AnyMatrix } TMatrixType;

  // @}
  // @name Constructeurs et destructeur
  // @{

  /**
   * Constructeur par dfaut
   * Construction d'une nouvelle instance de la classe, initialise  la
   * matrice identit (voir la mthode setToIdentity).
   *
   * @param AType Le type d'initialisation  effectuer (Identity par dfaut)
   * @return Une nouvelle instance de la classe
   */
  CTransformationMatrix(TMatrixType AType = IdentityMatrix);

  /**
   * Constructeur par copie
   * Construction d'une nouvelle instance de la classe, en copiant la matrice
   * passe en paramtre.
   *
   * @param AMatrix la matrice  copier
   * @return Une nouvelle instance de la classe
   */
  CTransformationMatrix(const CTransformationMatrix& AMatrix);

  /**
   * Constructeur avec initialisation
   * Construction d'une nouvelle instance de la classe, avec initialisation
   * explicite.
   *
   * @param A-- Les valeurs des cellules de la matrice.
   * @return Une nouvelle instance de la classe
   */
  CTransformationMatrix(TCoordinate A00, TCoordinate A01,
			TCoordinate A02, TCoordinate A03,

			TCoordinate A10, TCoordinate A11,
			TCoordinate A12, TCoordinate A13,

			TCoordinate A20, TCoordinate A21,
			TCoordinate A22, TCoordinate A23,

			TCoordinate A30, TCoordinate A31,
			TCoordinate A32, TCoordinate A33);

  /**
   * Constructeur avec initialisation  partir de sommets
   * Construction d'une nouvelle instance de la classe, avec initialisation.
   * Les coordonnes du sommet ALine1 sont places sur la premire ligne,
   * celles de ALine2 sur la deuxime ligne et celles de ALine3 sur la
   * troisime ligne.
   * La dernire ligne et la dernire colonne de la matrice sont initialises
   * avec la valeur 0.
   *
   * @param ALine- Des sommets
   * @return Une nouvelle instance de la classe
   */
  CTransformationMatrix(const CVertex& ALine1,
			const CVertex& ALine2,
			const CVertex& ALine3);

  /**
   * Destructeur
   */
  ~CTransformationMatrix();

  // @}
  // @name Initialisations
  // @{

  /**
   * Rinitialise l'instance  la matrice nulle.
   * La matrice nulle est compose uniquement de 0:
   *                   [0 0 0 0]
   *                   [0 0 0 0]
   *                   [0 0 0 0]
   *                   [0 0 0 0]
   */
  void setToNull();

  /**
   * Rinitialise l'instance  la matrice identit.
   * La matrice identit est compose de 1 sur sa premire diagonale et de 0
   * partout ailleurs: [1 0 0 0]
   *                   [0 1 0 0]
   *                   [0 0 1 0]
   *                   [0 0 0 1]
   */
  void setToIdentity();

  /**
   * Oprateur d'affectation
   *
   * @param AMatrix La matrice initiale
   * @return La matrice courante obtenue aprs affectation
   */
  CTransformationMatrix& operator=(const CTransformationMatrix& AMatrix);

  // @}
  // @name Accesseur aux lments de la matrice
  // @{

  /**
   * Permet d'accder (en lecture et en criture)  une ligne de la matrice.
   *
   * @param ARow La ligne de la matrice  atteindre
   *
   * @precondition ARow est compris entre 0 et 3 inclus.
   */
  TCoordinate* operator[](int ARow) const;

  // @}
  // @name Fonctions mathmatiques (produit, calcul de dterminant, ...)
  // @{

  /**
   * Calcule le produit de la matrice courante par la matrice AMatrix passe
   * en paramtre.
   *
   * @param AMatrix Une matrice quelconque
   * @return Le rsultat du produit de la matrice courante par la matrice
   *   AMatrix
   */
  CTransformationMatrix operator*(const CTransformationMatrix& AMatrix) const;

  /**
   * Calcule le produit de la matrice courante par le vecteur AVector pass
   * en paramtre.
   *
   * @param AVector Un vecteur quelconque
   * @return Le rsultat du produit de la matrice courante par le vecteur
   *   AVector
   */
  CVector operator*(const CVector& AVector) const;

  // @}
  // @name Mthodes correspondant aux transformations gomtriques
  // @{

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une translation de vecteur AVector.
   *
   * @param AVector Le vecteur de translation
   */
  void translate(const CVertex& AVector);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une rotation de AAngle degrs autour de l'axe passant par l'origine et
   * dont le vecteur directeur est donn par le paramtre AAxeDirection.
   *
   * @param AAxeDirection Le vecteur directeur de l'axe de rotation
   * @param AAngle L'angle (en degrs) de la rotation
   *
   * @precondition ! AAxeDirection.isNull()
   */
  void rotate(const CVertex& AAxeDirection, TCoordinate AAngle);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une rotation de AAngle degrs autour de l'axe passant par le sommet
   * AAxeVertex et dont le vecteur directeur est donn par le paramtre
   * AAxeDirection.
   *
   * @param AAxeVertex Un sommet quelconque de l'axe de rotation
   * @param AAxeDirection Le vecteur directeur de l'axe de rotation
   * @param AAngle L'angle (en degrs) de la rotation
   *
   * @precondition AAxe n'est pas un vecteur nul.
   */
  void rotate(const CVertex& AAxeVertex,
	      const CVertex& AAxeDirection, TCoordinate AAngle);

  /**
   * Effectue trois rotations sur le repre de base (OX,OY,OZ):
   *  - la premire  autour de l'axe OX et d'angle AAngles.getX(),
   *  - la deuxime  autour de l'axe OY et d'angle AAngles.getY(),
   *  - la troisime autour de l'axe OZ et d'angle AAngles.getZ().
   *
   * Remarque: les rotations sont faites autour des axes du repre
   * local. La deuxime rotation se fait donc autour de l'axe OY
   * qui alpha t modifi par la premire rotation. La troisime rotation
   * se fait autour de l'axe OZ qui alpha t modifi par les deux
   * premires rotations.
   *
   * @param AAngles Les angles de rotation
   */
  void rotate(const CVertex& AAngles);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une rotation de 180 degrs autour de l'axe passant par l'origine et
   * dont le vecteur directeur est donn par le paramtre AAxeDirection.
   *
   * @param AVertex Le vecteur directeur de l'axe de rotation
   *
   * @precondition !AAxeDirection.isNull()
   */
  void rotate180(const CVertex& AAxeDirection);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une homothtie centre sur l'origine dont le coefficient est donn par
   * le paramtre ACoef.
   *
   * @param ACoef Le coefficient de l'homothtie
   */
  void scale(TCoordinate ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une opration d'tirement dont les coefficients sont donns par le
   * paramtre ACoef.
   *
   * @param ACoef Les coefficients dans les trois dimensions de l'tirement
   */
  void scale(const CVertex& ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une homothtie de centre ACenter et de coefficient ACoef.
   *
   * @param ACenter Le centre d'homothtie
   * @param ACoef Le coefficient d'homothtie
   */
  void scale(const CVertex& ACenter, TCoordinate ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une opration d'tirement de center ACenter et dont les coefficients
   * sont donns par le paramtre ACoef.
   *
   * @param ACenter Le centre de l'tirement
   * @param ACoef Les coefficients dans les trois dimensions de l'tirement
   */
  void scale(const CVertex& ACenter, const CVertex& ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une homothtie d'axe la droite de vecteur directeur AAxeDirection
   * passant par l'origine et de coefficient ACoef.
   *
   * @param AAxeDirection Un vecteur directeur de l'axe
   * @param ACoef Le coefficient d'homothtie
   *
   * @precondition ! AAxeDirection.isNull()
   */
  void axialScale(const CVertex& AAxeDirection, TCoordinate ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une homothtie d'axe la droite de vecteur directeur AAxeDirection
   * passant par le sommet AAxeVertex et de coefficient ACoef.
   *
   * @param AAxeVertex Un sommet de l'axe
   * @param AAxeDirection Le vecteur directeur de l'axe
   * @param ACoef Le coefficient d'homothtie
   *
   * @precondition ! AAxeDirection.isNull()
   */
  void axialScale(const CVertex& AAxeVertex,
		  const CVertex& AAxeDirection, TCoordinate ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une homothtie selon le plan de vecteur normal APlaneNormal passant par
   * l'origine et de coefficient ACoef.
   *
   * @param APlaneNormal Un vecteur normal du plan
   * @param ACoef Le coefficient d'homothtie
   *
   * @precondition ! APlaneNormal.isNull()
   */
  void planarScale(const CVertex& APlaneNormal, TCoordinate ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une homothtie selon le plan de vecteur normal APlaneNormal passant par
   * le sommet APlaneVertex et de coefficient ACoef.
   *
   * @param APlaneVertex Un sommet du plan
   * @param APlaneNormal Un vecteur normal du plan
   * @param ACoef Le coefficient d'homothtie
   *
   * @precondition ! APlaneNormal.isNull()
   */
  void planarScale(const CVertex& APlaneVertex,
		   const CVertex& APlaneNormal, TCoordinate ACoef);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une rotation qui projette AVector1 sur AVector2.
   * L'axe de la rotation est donn par le vecteur AVector1*AVector2.
   *
   * La taille des vecteurs n'est pas prise en compte (pas d'tirement).
   *
   * @param AVector1 Un premier vecteur
   * @param AVector2 Un second vecteur
   *
   * @precondition Aucun des vecteurs passs en paramtre n'est le vecteur nul.
   */
  void orientate(const CVertex& AVector1, const CVertex& AVector2);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une rotation autour du sommet ACenter qui projette AVector1 sur
   * AVector2.
   * L'axe de la rotation est donn par le vecteur AVector1*AVector2.
   *
   * La taille des vecteurs n'est pas prise en compte (pas d'tirement).
   *
   * @param ACenter  Le sommet origine
   * @param AVector1 Un premier vecteur
   * @param AVector2 Un second vecteur
   *
   * @precondition Aucun des vecteurs passs en paramtre n'est le vecteur nul.
   */
  void orientate(const CVertex& ACenter,
		 const CVertex& AVector1, const CVertex& AVector2);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  une ou plusieurs rotations qui correspondent au changement de repre
   * (AVectorU1, AVectorV1) vers (AVectorU2, AVectorV2).
   *
   * Si les vecteurs AVectorUi et AVectorVi ne sont pas orthogonaux, AVectorVi
   * est projet sur le plan orthogonal  AVectorUi et passant par l'origine.
   * Le troisime vecteur de chaque repre est dtermin par produit vectoriel
   * des deux premiers.
   *
   * La taille des vecteurs n'est pas prise en compte (pas d'tirement).
   *
   * @param AVectorU1 Le premier vecteur du premier repre
   * @param AVectorV1 Le second vecteur du premier repre
   * @param AVectorU2 Le premier vecteur du deuxime repre
   * @param AVectorV2 Le second vecteur du deuxime repre
   *
   * @precondition Aucun des vecteurs passs en paramtre n'est le vecteur nul.
   */
  void orientate(const CVertex& AVectorU1, const CVertex& AVectorV1,
		 const CVertex& AVectorU2, const CVertex& AVectorV2);

  /**
   * Modifie la matrice courante en la combinant avec la matrice correspondant
   *  un changement de repre.
   *
   * @param AOrigin1  L'origine du premier repre
   * @param AVectorU1 Le premier vecteur du premier repre
   * @param AVectorV1 Le second vecteur du premier repre
   * @param AOrigin2  L'origine du second  repre
   * @param AVectorU2 Le premier vecteur du deuxime repre
   * @param AVectorV2 Le second vecteur du deuxime repre
   * @param ACoef     Les coefficients d'tirement dans les 3 dimensions
   *
   * @precondition Aucun des vecteurs passs en paramtre n'est le vecteur nul.
   */
  void transform(const CVertex& AOrigin1 ,
		 const CVertex& AVectorU1, const CVertex& AVectorV1,
		 const CVertex& AOrigin2 ,
		 const CVertex& AVectorU2, const CVertex& AVectorV2,
		 const CVertex& ACoef = CVertex(1,1,1));

  // @}
  // @name Application de la matrice sur un sommet
  // @{

  /**
   * @param AVertex Le sommet sur lequel appliquer la transformation
   */
  void applyOn(CVertex& AVertex) const;

  /**
   * @param AVertex Le sommet sur lequel appliquer la transformation
   * @param AResult Le sommet obtenu aprs transformation
   */
  void applyOn(const CVertex& AVertex, CVertex& AResult) const;

  // @}
};

// @name Matrices de lissage
// @{

extern const CTransformationMatrix COONS     ;
extern const CTransformationMatrix HERMITE   ;
extern const CTransformationMatrix BEZIER    ;
extern const CTransformationMatrix OVERHAUSER;
extern const CTransformationMatrix SPLINE    ;

// @}
// @name Affichage
// @{

/**
 * @param AStream Le flot dans lequel afficher la matrice
 * @param AMatrix La matrice  afficher
 * @return Le flot utilis pour l'affichage
 */
std::ostream& operator<<(std::ostream& AStream,
			 const CTransformationMatrix& AMatrix);

// @}
//******************************************************************************
#include INCLUDE_INLINE("transformation-matrix.icc")
//******************************************************************************
#endif // TRANSFORMATION_MATRIX_HH
//******************************************************************************
