/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef STREAMS_HH
#define STREAMS_HH
//******************************************************************************
#include "math-extension.hh"  // pour la dfinition du type tcoordinate
#include "alt-stdint.hh"
#include "inline-macro.hh"

#include <iostream>
//******************************************************************************

/**
 * Bibliothque de fonctions facilitant l'utilisation des flots.
 *
 * @author MOKA TEAM
 */

// @name Conversions tableau de 8 bits <-> octet
// @{

/**
 * Convertit un tableau de 8 boolens en un octet non sign.
 * Les premiers lments du tableau (0, 1, ...) constituent les bits de poids
 * faibles du rsultat.
 * Les derniers lments du tableau (7, 6, ...) constituent les bits de poids
 * forts du rsultat.
 *
 * @param ABoolArray Un tableau de 8 boolens
 * @return L'octet correspondant aux 8 boolens
 */
uint8_t bool2char(bool ABoolArray[/*8*/]);

/**
 * Convertit un octet non sign en un tableau de 8 boolens.
 * Les bits de poids faibles de AChar sont stocks dans les premiers lments
 * du tableau (0, 1, ...).
 * Les bits de poids forts de AChar sont stocks dans les derniers lments
 * du tableau (7, 6, ...).
 *
 * @param AChar Un octet (non sign)
 * @param ABoolArray Un tableay de 8 boolens
 */
void char2bool(uint8_t AChar, bool ABoolArray[/*8*/]);

// @}
// @name Lecture et criture d'lments en mode binaire ou ascii
// @{

/**
 * Passe en mode mode binaire.
 */
void setBinaryMode();

/**
 * Passe en mode ascii.
 */
void setAsciiMode();

/**
 * crit la valeur de ABool dans le flot AStream, en tenant compte du mode de
 * sauvegarde courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @param ABool Un boolen
 */
void writeBool(std::ostream& AStream, bool ABool);

/**
 * crit la valeur de AChar dans le flot AStream, en tenant compte du mode de
 * sauvegarde courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @param AChar Un entier non sign cod sur 1 octet
 */
void writeChar(std::ostream& AStream, uint8_t AChar);

/**
 * crit la valeur de AInt dans le flot AStream, en tenant compte du mode de
 * sauvegarde courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @param AInt Un entier non sign cod sur 4 octets
 */
void writeInt(std::ostream& AStream, uint32_t AInt);

/**
 * crit la valeur de ACoord dans le flot AStream, en tenant compte du mode de
 * sauvegarde courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @param ACoord Un flottant
 */
void writeCoord(std::ostream& AStream, const TCoordinate& ACoord);

/**
 * Lit un boolen dans le flot AStream, en tenant compte du mode de chargement
 * courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @return Un boolen
 */
bool readBool(std::istream& AStream);

/**
 * Lit un entier cod sur 1 octet dans le flot AStream, en tenant compte du
 * mode de chargement courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @return Un entier non sign cod sur 1 octet
 */
uint8_t readChar(std::istream& AStream);

/**
 * Lit un entier cod sur 4 octets dans le flot AStream, en tenant compte du
 * mode de chargement courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @return Un entier non sign cod sur 4 octets
 */
uint32_t readInt(std::istream& AStream);

/**
 * Lit un flottant dans le flot AStream, en tenant compte du mode de chargement
 * courant (binaire ou ascii).
 *
 * @param AStream Un flot
 * @return Un flottant
 */
TCoordinate readCoord(std::istream& AStream);

/**
 * crit un caractre corresondant  une tabulation dans le flot si le mode
 * ascii est le mode courant.
 * En mode binaire, cette fonction n'alpha aucun effet.
 *
 * @param AStream Un flot
 */
void writeTab(std::ostream& AStream);

/**
 * crit un caractre correspondant  un espace dans le flot si le mode ascii
 * est le mode courant.
 * En mode binaire, cette fonction n'alpha aucun effet.
 *
 * @param AStream Un flot
 */
void writeSpc(std::ostream& AStream);

/**
 * crit un caractre correspondant  un retour  la ligne dans le flot si le
 * mode ascii est le mode courant.
 * En mode binaire, cette fonction n'alpha aucun effet.
 *
 * @param AStream Un flot
 */
void writeRet(std::ostream& AStream);

// @}

//******************************************************************************
#include INCLUDE_INLINE("streams.icc")
//******************************************************************************
#endif // STREAMS_HH
//******************************************************************************
