(*
  Minimal separators of a graph.
  
  Based on the article:
  Generating all the minimal separators of a graph.
  by A. Berry, J.-P. Bordat and O.Cogis
  @see http://www.isima.fr/berry/generating.html

  @author Matthieu Sozeau
  @author Pierre-Loic Garoche
*)

let list_change_assoc default f l k =
  let rec aux accu = function
      (x, y) as hd :: tl ->
	if x = k then
	  (k, f y) :: (tl @ accu)
	else
	  aux (hd :: accu) tl
    | [] -> 
	(k, f default) :: accu
  in aux [] l
       
module Components(G : Sig.G) =
struct

  module VertexSet = Set.Make(G.V)
  
  module StdComponents = Components.Make(G)

  let ccs g = 
    let sccs = StdComponents.scc g in
    let replace = list_change_assoc VertexSet.empty in
    let ccs =
      G.fold_vertex 
	(fun x ccs ->
	   let cc = sccs x in
	     replace (fun c -> VertexSet.add x c) ccs cc)
	g []
    in ccs

end

module MinimalSeparators =
struct

  module P(G : Sig.P) =
  struct
        
    open G
      
    let check_graph () = 
      if is_directed then
	raise (Failure "Cannot apply algorithm to directed graph")
      else ()
	
    module Neighbourhood = Neighborhood.Graph(G)
    module Components = Components(G)
    module Diff = Oper.Difference.P(G)
      
    module VertexSet = Diff.VertexSet
    module MinSepSet = Set.Make(struct
				  type t = VertexSet.t
				  let compare = VertexSet.compare
				end)
      
    (*
      AllMinSep: O(n^3 |Seps(G)|)
    *)
      
    let allminsep g =
      check_graph ();
      let minseps = 
	(* Initialization *)
	fold_vertex 
	  (fun v minseps ->
	     let nh = Neighbourhood.of_vertex g v in
	     let vs = VertexSet.add v nh in
	     let g' = Diff.minus_vertices g vs in
	     let components = Components.ccs g' in
	       List.fold_left
		 (fun minseps (_, cc) ->
		    assert (cc <> VertexSet.empty);
		    MinSepSet.add (Neighbourhood.of_vertices g cc) minseps)
		 minseps components)
	  g MinSepSet.empty
      in
	(* Generation *)
      let minsepsr = ref minseps in
      let accu = ref MinSepSet.empty in
      let diff = ref minseps in
	while !diff <> MinSepSet.empty do
	  let s = MinSepSet.choose !diff in
	  let minseps =
	    VertexSet.fold 
	      (fun x minseps ->
		 let nx = Neighbourhood.of_vertex g x in
		 let vs = VertexSet.union nx s in
		 let g' = Diff.minus_vertices g vs in
		 let ccs = Components.ccs g' in
		   List.fold_left
		     (fun minseps (_, n) ->
			assert (n <> VertexSet.empty);
			let ncc = Neighbourhood.of_vertices g n in
			  MinSepSet.add ncc minseps)
		     minseps ccs)
	      s !minsepsr
	  in 
	    accu := MinSepSet.add s !accu;
	    minsepsr := minseps;
	    diff := MinSepSet.diff minseps !accu
	done;
	MinSepSet.elements !minsepsr
	  
  end

  module I(Gr : Sig.I) =
  struct
    module G = Persistent.Graph.Concrete(Gr.V)

    module Copy = Copy.Copy.P(Gr)(G)(struct 
					     let vertex g v = 
					       v, G.add_vertex g v
					     let edge g x y =
					       G.add_edge g x y
					   end)

    module MinSep = P(G)
    module VertexSet = MinSep.VertexSet

    let allminsep g =
      let g' = Copy.copy g in
	MinSep.allminsep g'
  end

end
