/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Cre un triangle topologique (6 brins 0 et 1-cousus).
 *
 * @return Un brin du triangle
 */
CDart* createTopoTriangle();

/**
 * Cre un quadrilatre topologique (8 brins 0 et 1-cousus).
 *
 * @return Un brin du quadrilatre
 */
CDart* createTopoSquare();

/**
 * Cre un polygone  An cts.
 *
 * @param An Le nombre d'artes du polygone
 * @return Un brin du polygone
 */
CDart* createTopoPolygon(int An);

/**
 * Cre un ple ouvert compos de An triangles 2-cousus entre eux et avec
 * un sommet central.
 * Le ple est ouvert, c'est--dire que 2 des artes incidentes au sommet
 * central ne sont pas 2-cousues.
 *
 * @param An Le nombre de triangles qui composent le ple
 * @return Un brin 2-libre du sommet central du ple
 */
CDart* createTopoOpenedPole(int An);

/**
 * Cre un ple compos de An triangles 2-cousus entre eux et avec un
 * sommet central.
 *
 * @param An Le nombre de triangles qui composent le ple
 * @return Un brin du sommet central du ple
 */
CDart* createTopoPole(int An);

/**
 * Cre une grille de dimension AMeridians*AParallels en s'appuyant sur la
 * mthode 'createTopoMesh2', et 2-coud deux cts de cette grille pour
 * obtenir un tube.
 * Les deux 2-bords du tube obtenu sont composs de AMeridians artes chacun.
 *
 * ABorder1 s'il n'est pas NULL est positionn sur un brin 2-libre du
 * premier bord du tube.
 * ABorder2 s'il n'est pas NULL est positionn sur un brin 2-libre du
 * second bord du tube.
 * ABorder1 et ABorder2 sont situs en face l'un de l'autre.
 *
 * Remarque :
 * Les paramtres ABorder- passs peuvent tre gaux au pointeur NULL.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du tube
 * @param AParallels Le nombre (non nul) de parallles du tube
 * @param ABorder1 L'adresse d'un pointeur sur brin
 * @param ABorder2 L'adresse d'un pointeur sur brin
 */
void createTopoTube(int AMeridians, int AParallels,
		    CDart* * ABorder1, CDart* * ABorder2);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoTube' ci-dessus et
 * retourne le brin correspondant  ABorder1.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du tube
 * @param AParallels Le nombre (non nul) de parallles du tube
 * @return Un brin du tube
 */
CDart* createTopoTube(int AMeridians, int AParallels);

/**
 * Cre un tube de dimension AMeridians*AParallels en s'appuyant sur la
 * mthode 'createTopoTube', et 2-coud un ple de chaque ct de ce tube
 * afin de fermer ses bords.
 *
 * Si ASouthPole n'est pas NULL, il est positionn sur un brin du sommet
 * situ au ple sud de la sphre cre.
 * Si ASouthPole n'est pas NULL, il est positionn sur un brin du sommet
 * situ au ple nord de la sphre cre.
 *
 * Remarque :
 * Les adresses des pointeurs sur brins passs en paramtre peuvent tre
 * gales  NULL.
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la sphre
 * @param AParallels Le nombre (non nul) de parallles de la sphre
 * @param ASouthPole L'adresse d'un pointeur sur brin
 * @param ANorthPole L'adresse d'un pointeur sur brin
 */
void createTopoSphere(int AMeridians, int AParallels,
		      CDart* * ASouthPole, CDart* * ANorthPole);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoSphere' ci-dessus
 * et retourne le brin correspondant  ASouthPole.
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la sphre
 * @param AParallels Le nombre (non nul) de parallles de la sphre
 * @return Un brin de la sphre
 */
CDart* createTopoSphere(int AMeridians, int AParallels);

/**
 * Cre un tube de dimension AMeridians*AParallels en s'appuyant sur la
 * mthode 'createTopoTube', et 2-coud une face de AMeridians sommets 
 * chaque bout du tube de manire  obtenir un cylindre (selon les valeurs
 * de AClose1 et AClose2, les bords sont ferms ou non).
 *
 * Remarque :
 * Les paramtres ABorder- passs peuvent tre gaux au pointeur NULL.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du cylindre
 * @param AParallels Le nombre (non nul) de parallles du cylindre
 * @param ABorder1 L'adresse d'un pointeur sur brin
 * @param ABorder2 L'adresse d'un pointeur sur brin
 * @param AClose1 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder1
 * @param AClose2 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder2
 */
void createTopoCylinder(int AMeridians, int AParallels,
			CDart* * ABorder1, CDart* * ABorder2,
			bool AClose1 = true, bool AClose2 = true);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoCylinder' ci-dessus
 * et retourne le brin correspondant  ABorder1.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du cylindre
 * @param AParallels Le nombre (non nul) de parallles du cylindre
 * @param ABorder1 L'adresse d'un pointeur sur brin
 * @param ABorder2 L'adresse d'un pointeur sur brin
 * @param AClose1 Un boolen indiquant s'il faut fermer le bord incident
 *  ABorder1 (voir mthode 'createTopoCylinder' ci-dessus)
 * @param AClose2 Un boolen indiquant s'il faut fermer le bord incident
 *    ABorder2 (voir mthode 'createTopoCylinder' ci-dessus)
 * @return Un brin du cylindre
 */
CDart* createTopoCylinder(int AMeridians, int AParallels,
			  bool AClose1 = true, bool AClose2 = true);

/**
 * Cre une pyramide  AMeridians mridiens et AParallels parallles.
 *
 * ABaseDart, s'il n'est pas NULL est positionn sur un brin 2-libre de la
 * base (ou 2-cousu  un brin de la base si cette base est cre).
 * APoleDart, s'il n'est pas NULL est positionn sur un brin situ sur la
 * pointe de la pyramide.
 *
 * ABaseDart et APoleDart sont situs sur le mme mridien.
 *
 * Remarque :
 * Les paramtres ABaseDart et APoleDart passs peuvent tre gaux au
 * pointeur NULL.
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la pyramide
 * @param AParallels Le nombre (non nul) de parallles de la pyramide
 * @param ABaseDart  L'adresse d'un pointeur sur brin
 * @param APoleDart  L'adresse d'un pointeur sur brin
 * @param ACloseBase Un boolen indiquant si la base de la pyramide doit tre
 *   ferme
 */
void createTopoPyramid(int AMeridians, int AParallels,
		       CDart* * ABaseDart, CDart* * APoleDart,
		       bool ACloseBase = true);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoPyramid' ci-dessus
 * et retourne le brin correspondant  ABaseDart.
 *
 * @param AMeridians Le nombre (non nul) de mridiens de la pyramide
 * @param AParallels Le nombre (non nul) de parallles de la pyramide
 * @param ACloseBase Un boolen indiquant si la base de la pyramide doit
 *   tre ferme
 */
CDart* createTopoPyramid(int AMeridians, int AParallels,
			 bool ACloseBase = true);

/**
 * Cre un tube de dimension AMeridians*AParallels en s'appuyant sur la
 * mthode 'createTopoTube', et 2-coud les deux 2-bords de ce tube de
 * manire  obtenir un tore.
 *
 * AEquator, s'il n'est pas NULL est positionn sur un brin situ sur
 * l'quateur du tore.
 *
 * Remarque :
 * Le paramtre AEquator pass peut tre gal au pointeur NULL.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du tore
 * @param AParallels Le nombre (non nul) de parallles du tore
 * @param AEquator L'adresse d'un pointeur sur brin
 */
void createTopoTorus(int AMeridians, int AParallels, CDart* * AEquator);

/**
 * Cette mthode est fournie dans le but de faciliter l'exploitation du
 * noyau. Elle appelle simplement la mthode 'createTopoTorus' ci-dessus
 * et retourne le brin correspondant  AEquator.
 *
 * @param AMeridians Le nombre (non nul) de mridiens du tore
 * @param AParallels Le nombre (non nul) de parallles du tore
 * @return Un brin du tore
 */
CDart* createTopoTorus(int AMeridians, int AParallels);

/**
 * Cre un carr, ventuellement maill.
 * Le paramtre ASquareCorners permet de rcuprer des brins stratgiques
 * du carr, comme indiqu sur la figure.
 *
 * ASx=3, ASy=2
 *
 * ASquareCorners[0][1]             ASquareCorners[1][1]
 *                  +== --+-- --+-- ==+
 *        	    |                 |
 *        	    |                 |
 *        	    +                 +
 *        	    |                 |
 *        	    |                 |
 *        	    +== --+-- --+-- ==+
 * ASquareCorners[0][0]             ASquareCorners[1][0]
 *
 * Si AMeshDimension==0, les valeurs de ASx et ASy ne sont pas prises en compte.
 *
 * @param ASx Le nombre de subdivisions selon la premire dimension
 * @param ASy Le nombre de subdivisions selon la deuxime dimension
 * @param AMeshDimension La dimension du maillage (0 ou 2)
 * @param ASquareCorners Un tableau de dimension 2*2 de pointeurs sur brin
 */

void createTopoSquareIMeshed(int ASx, int ASy,
			     int AMeshDimension,
			     CDart* ASquareCorners[2][2]);

/**
 * Cre un cube, ventuellement maill.
 * Le paramtre AFacesCorners permet de rcuprer des brins stratgiques du
 * cube, comme indiqu sur la figure.
 *
 * ACreatedFaces[0][0] correspond  la face X0.
 * ACreatedFaces[0][1] correspond  la face X1.
 * ACreatedFaces[1][0] correspond  la face Y0.
 * ACreatedFaces[1][1] correspond  la face Y1.
 * ACreatedFaces[2][0] correspond  la face Z0.
 * ACreatedFaces[2][1] correspond  la face Z1.
 *
 * AFacesCorners[0][0] correspond aux coins de la face X0.
 * AFacesCorners[0][1] correspond aux coins de la face X1.
 * AFacesCorners[1][0] correspond aux coins de la face Y0.
 * AFacesCorners[1][1] correspond aux coins de la face Y1.
 * AFacesCorners[2][0] correspond aux coins de la face Z0.
 * AFacesCorners[2][1] correspond aux coins de la face Z1.
 *
 * Les lments des tableaux AFacesCorners[dim][i] sont des pointeurs qui
 * sont positionns sur certains brins du cube.
 * Ces brins sont numrots sur le schma.
 *
 * Les six boolens donns par ACreatedFaces indiquent quelles faces du
 * cube doivent tre cres.
 *
 *            +----  ----+           +01--  --11+
 *        +   10        11          /          /    +
 *      11|   |          |         /          /   11|
 *      / |      ARRIRE              HAUT        / |
 *            |   (Y1)   |       /   (Z1)   /
 *    /   |   00        01      /          /    /   |
 *  01    |   +----  ----+     +00--  --10+   01    |
 *  +  G. +                                   +  D. +
 *  |(X0)10   +01--  --11+     +----  ----+   |(X1)10
 *  |   /    /          /      10        11   |   /
 *          /          /       |          |
 *  | /        BAS                 AVANT      | /
 *  |00   /   (Z0)   /         |   (Y0)   |   |00
 *  +    /          /          00        01   +
 *      +00--  --10+           +----  ----+
 *
 * AMeshDimension indique dans quelle dimension le cube cr doit tre maill.
 * Pour crer un cube 3-maill, utilisez la mthode createTopoMesh3.
 *
 * Si AMeshDimension==0, les valeurs de ASx, ASy et ASz ne sont pas prises
 * en compte.
 *
 * @param ASx Le nombre de subdivisions selon la premire dimension
 * @param ASy Le nombre de subdivisions selon la deuxime dimension
 * @param ASz Le nombre de subdivisions selon la troisime dimension
 * @param AMeshDimension La dimension du maillage (0, 1 ou 2)
 * @param ACreatedFaces Un tableau de dimension 3*2 de boolens
 * @param AFacesCorners Un tableau de dimension 3*2*2*2 de pointeurs sur
 *   brin
 */
void createTopoCubeIMeshed(int ASx, int ASy, int ASz,
			   int AMeshDimension,
			   bool ACreatedFaces[3][2],
			   CDart* AFacesCorners[3][2][2][2]);

//******************************************************************************
